import java.net.*;
import java.io.*;
import java.util.*;

public class Connection
    extends Thread
{
    private DatagramSocket socket;
    private InetAddress clientIP;
    private int clientPort;
    private int maxMsgSize;
    private Server server;
    private Player player;

    public Connection( Server server, 
		       InetAddress clientIP,
		       int clientPort,
		       int maxMsgSize )
    {
	this.clientIP = clientIP;
	this.clientPort = clientPort;
	this.maxMsgSize = maxMsgSize;
	this.server = server;
	this.player = null;

	try {
	    socket = new DatagramSocket();
	}
	catch ( SocketException e ) {
            System.err.println( "Could not create socket on new port." );
            System.exit( -1 );
        }
    }
    
    public void setPlayer( Player player )
    {
	this.player = player;
    }

    public boolean matchesClient( InetAddress clientIP, 
				  int clientPort )
    {
	return ( clientIP.equals( this.clientIP ) &&
		 clientPort == this.clientPort );
    }
    
    public void run() 
    {
	byte[] inBuffer = new byte[ maxMsgSize ];
        DatagramPacket inPacket = 
	    new DatagramPacket( inBuffer, maxMsgSize );
	boolean listening = true;

        while( listening ) {
	    try {
		socket.receive( inPacket );
	    } 
	    catch ( IOException e ) {
		e.printStackTrace();
	    }

	    if ( matchesClient( inPacket.getAddress(), inPacket.getPort() ) )
		server.processMsg( new String( inPacket.getData(),
					       inPacket.getOffset(),
					       inPacket.getLength() ),
				   player );
	}

	socket.close();
    }

    public void send( String msg )
    {
	DatagramPacket outPacket = 
	    new DatagramPacket( msg.getBytes(), msg.length(),
				clientIP, clientPort );
	try {
	    socket.send( outPacket );
        } 
	catch ( IOException e ) {
	    System.err.println( "send error: " + e.getMessage() );
	    e.printStackTrace();
        }
    }
}
