import java.net.*;
import java.io.*;
import java.util.*;

public class ConnectionListener
    extends Thread
{
    private DatagramSocket listenSocket;
    private int listenPort;
    private int maxMsgSize;
    private Server server;

    public ConnectionListener( Server server,
			       int listenPort, 
			       int maxMsgSize )
    {
	this.listenSocket = null;
	this.listenPort = listenPort;
	this.server = server;
	this.maxMsgSize = maxMsgSize;

        try {
            listenSocket = new DatagramSocket( listenPort );
	} 
	catch ( SocketException e ) {
            System.err.println( "Could not create socket on port: " + 
				listenPort );
            System.exit( -1 );
        }
    }

    public void run() 
    {
	byte[] inBuffer = new byte[ maxMsgSize ];
	DatagramPacket inPacket = 
	    new DatagramPacket( inBuffer, maxMsgSize );
	boolean listening = true;

        while( listening ) {
	    try {
		listenSocket.receive( inPacket );
	    } 
	    catch ( IOException e ) {
		e.printStackTrace();
	    }

	    if ( inPacket.getLength() > 0 )
		processRequest( inPacket.getAddress(),
				inPacket.getPort(),
				new String( inPacket.getData(), 
					    inPacket.getOffset(), 
					    inPacket.getLength() ) );
	}
	listenSocket.close();
    }

    public void send( String msg, InetAddress clientIP, int clientPort )
    {
	DatagramPacket outPacket = 
	    new DatagramPacket( msg.getBytes(), msg.length(),
				clientIP, clientPort );
	try {
	    listenSocket.send( outPacket );
        } 
	catch ( IOException e ) {
	    System.err.println( "send error: " + e.getMessage() );
	    e.printStackTrace();
        }
    }

    private void processRequest( InetAddress clientIP,
				 int clientPort,
				 String msg )
    {
	Connection c = server.findConnection( clientIP, clientPort );

	if ( c == null ) { // New connection
	    if ( msg.compareTo( "(init)" ) == 0 ) {
		c = new Connection( server, clientIP, 
				    clientPort, maxMsgSize );
		server.createPlayer( c );
		c.start();
		c.send( "(connect)" );
	    }
	    else {
		send( "(error bad-init-request)", clientIP, clientPort );
	    }
	}
	else { // Existing connection - shouldn't happen
	    if ( msg.equals( "(init)" ) ) {
		c.send( "(error already-connected)" );
	    }
	    else {
		c.send( "(error only-init-on-init-port)" );
	    }
	}
    }
}
