import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import java.awt.image.*;
import java.io.*;
import javax.imageio.*;
import java.util.Vector;

public class FieldCanvas extends JPanel {


    private static final String imageFilename = "field.png";
    private static final double CANVAS_Y = 528;
    private static final double CANVAS_X = 368;
    private static final double CANVAS_OFFSET = 0;
    private static final double FIELD_X = 4600;
    private static final double FIELD_Y = 6600;
    private static final double GOAL_WIDTH = 800;
    private static final int IMAGE_VERT_OFFSET = 0;
    private static final int NUM_PLAYERS = 4;
    private static final int PARTICLE_SIZE = 2;
    private static final int FORCE_TILE_SIZE = 30;

    /*
     * These just control what the IR display looks like.  The min and
     * max are values that the actual sensor can take on, though.
     */

    private static final double HEAD_LENGTH = -60;
    private static final double IR_MIN = 100;
    private static final double IR_MAX = 900;
    private static final double IR_RANGE = IR_MAX - IR_MIN;
    private static final double IR_WIDTH = 100;
    
    /**
     * The following help define what our players look like.  These
     * are the two parameters for an icosceles triangle.
     */

    private static final int PLAYER_TRIANGLE_CENTER = 0;
    private static final int PLAYER_TRIANGLE_TIP = 1;
 
    private static final double TOP_ANGLE = (Math.PI/5.0);     
    private static final double BASE_LENGTH = 20.0;             // 20 pixels

    private static final double PLAYER_BODY_SIZE = BASE_LENGTH;
    private static final double PLAYER_HEAD_SIZE = (BASE_LENGTH / 2.0);

    /**
     * The following lets us change the ball size;
     */

    private static final int BALL_RADIUS = 6;  // 6 pixels, baby!

    /**
     * How big we draw the opponents
     */
    private static final int OPPONENT_RADIUS = 10;

    /**
     * The following lets us change how sensitive we are in
     * determining if a user is mousing over a player or a ball.
     */
    private static final double HALO_RATIO = 1; 


    private MediaTracker mediaTracker;
    private JFrame frame;

    private Image backgroundImage;
    private Image displayImage;
    private Graphics2D displayBuffer;
    private Image wholeImage;
    private Graphics2D wholeBuffer;

    /*
     * Now we're going to keep track of what we actually want to draw
     * on our field.
     */

    private boolean drawBall;
    private boolean[] drawP;
    private boolean[] drawPText;
    private boolean drawDeltas;
    private boolean[] drawParticles;
    private boolean[] drawVisibleBeacons;
    private boolean[] drawForceFields;
    private boolean[] drawOpponents;
    private boolean[] drawIRSensorData;

    /*
     * Here we keep track of where the mouse is.
     */

    private int mouseX;
    private int mouseY;

    // housekeeping variable for FieldCanvas
    private boolean debug = true;

    /**
     * Construct a FieldCanvas 
     */
    public FieldCanvas( JFrame frame ) {
	super();

	this.frame = frame;

	init();
        setVisible( true );

	drawBall = false;
	drawDeltas = false;
	
	drawP = new boolean[NUM_PLAYERS];
	drawPText = new boolean[NUM_PLAYERS];
	drawParticles = new boolean[NUM_PLAYERS];
	drawVisibleBeacons = new boolean[NUM_PLAYERS];
	drawForceFields = new boolean[NUM_PLAYERS];
	drawOpponents = new boolean[NUM_PLAYERS];
	drawIRSensorData = new boolean[NUM_PLAYERS];
	for( int i = 0; i < NUM_PLAYERS; i++ ) {
	    drawP[i] = false;
	    drawPText[i] = false;
	    drawParticles[i] = false;
	    drawVisibleBeacons[i] = false;
	    drawForceFields[i] = false;
	    drawOpponents[i] = false;
	    drawIRSensorData[i] = false;
	}
    }


    //////////////////////////////////////////////////////////


    /**
     * This will be the function that takes care of updating the
     * moving things on the field.  Calls subfunctions responsible for
     * the individual things.
     */
    public void updateField( Vector players, Ball ball ) {
	if( displayBuffer != null) {
	    
	    // display the background image
	    displayBuffer.drawImage( backgroundImage, 0, 0, this );

	    drawBall( (int) ball.getX(),
		      (int) ball.getY() );
	    drawBallParticles( ball );

	    for ( int i = 0; i < players.size(); i++ ) {
		Player player = (Player) players.elementAt( i );
		drawParticles( player );

		drawPlayer( (int) player.getX(),
			    (int) player.getY(),
			    player.getOrientation(),
			    player.getHeadPan(), 1.0 );

		drawPlayer( (int) player.getEstimateX(),
			    (int) player.getEstimateY(),
			    player.getEstimateT(),
			    player.getEstimatePan(),
			    player.getEstimateP() );
	    }
	    repaint();
	}
    }
	    
// 	    // run through the players and draw each one
// 	    for( int player = 0; player < NUM_PLAYERS; player++ ) {
// 		if ( drawP[player] ) {
// 		    if( drawForceFields[player] ) {
// 			drawBallForceField( f.ballForceField[player] );
// 		    }

// 		    if( drawOpponents[player] ) {
// 			drawOpponents( f.opponentX[player], f.opponentY[player] );
// 		    }

// 		    if ( drawBall ) {
// 			drawBall( f.ballX[player], f.ballY[player], 
// 				  ballHaloed(f,player), player );
// 		    }

// 		    if ( drawParticles[player] ) {
// 			drawParticles( f.particles[player], player );
// 		    }

// 		    if ( drawVisibleBeacons[player] ) {
// 			drawVisibleBeacons( f.visibleBeacons[player], 
// 					    f.pX[player], f.pY[player] );
// 		    }

// 		    if (drawIRSensorData[player] && 
// 			(f.pIR[player] <= 1.0) &&
// 			(f.pIR[player] >= 0.0)) {
// 			drawIRSensorData( f.pX[player], 
// 					  f.pY[player],
// 					  f.pTheta[player] + f.pPhi[player],
// 					  f.pIR[player] );
// 		    }

// 		    drawPlayer( f.pX[player],
// 				f.pY[player],
// 				f.pTheta[player],
// 				f.pPhi[player],
// 				f.pDeltaX[player],
// 				f.pDeltaY[player],
// 				String.valueOf(player),
// 				drawPText[player],
// 				haloed(player,f) );

// 		}
// 	    }

// 	    repaint();
// 	}
//     }

//     /**
//      * Here we want to update where the mouse is.
//      */
    
//     public void updateMouse( int x, int y ) {
// 	mouseX = x;
// 	mouseY = y;
//     }

//     /**
//      * Here we let the outside world change what we draw
//      */

//     public void updateBallDraw( boolean draw ) {
// 	drawBall = draw;
//     }

//     public void updatePlayerDraw( int player, boolean a ) {
// 	if( (player >= NUM_PLAYERS) || (player < 0) ) {
// 	    debug("Not a valid player.");
// 	    return;
// 	}
// 	drawP[player] = a;
//     }

//     public void updatePlayerDrawText( int player, boolean a ) {
// 	if( (player >= NUM_PLAYERS) || (player < 0) ) {
// 	    debug("Not a valid player.");
// 	    return;
// 	}
// 	drawPText[player] = a;
//     }

//     public void updateDeltaDraw( boolean a ) {
// 	drawDeltas = a;
//     }

//     public void updateParticlesDraw( int player, boolean a) {
// 	if( (player >= NUM_PLAYERS) || (player < 0) ) {
// 	    debug("Not a valid player.");
// 	    return;
// 	}
// 	drawParticles[player] = a;
//     }

//     public void updateVisibleBeaconsDraw( int player, boolean a ) {
// 	if( (player >= NUM_PLAYERS) || (player < 0) ) {
// 	    debug("Not a valid player.");
// 	    return;
// 	}
// 	drawVisibleBeacons[player] = a;
//     }

//     public void updateForceFieldsDraw( int player, boolean a ) {
// 	if( (player >= NUM_PLAYERS) || (player < 0) ) {
// 	    debug("Not a valid player.");
// 	    return;
// 	}
// 	drawForceFields[player] = a;
//     }

//     public void updateOpponentsDraw( int player, boolean a ) {
// 	if( (player >= NUM_PLAYERS) || (player < 0) ) {
// 	    debug("Not a valid player.");
// 	    return;
// 	}
// 	drawOpponents[player] = a;
//     }

//     public void updateIRSensorDataDraw( int player, boolean a ) {
// 	if( (player >= NUM_PLAYERS) || (player < 0) ) {
// 	    debug("Not a valid player.");
// 	    return;
// 	}
// 	drawIRSensorData[player] = a;
//     }


//     /**
//      * Draw the position as text
//      */
//     public void drawPosition( int x, int y ) {
// 	displayBuffer.setPaint( Color.BLACK );
// 	displayBuffer.drawString( "(" + x + "," + y + ")", Xx(x), Yy(y) );
// 	repaint();
//     }

//     /**
//      * Draw the mouse!
//      */
//     public void drawMouse( int x, int y ) {
// 	//debug( x + "," + y + " mapped to " + Xx(x) + "," + Yy(y) );
// 	displayBuffer.setPaint( Color.RED );
// 	displayBuffer.fillOval( Xx(x), Yy(y), 10, 10 );
//     }
    
//     /**
//      * Draw the opponents
//      */
//     public void drawOpponents( double[] opponentX, double[] opponentY ) {
// 	if( opponentX != null && opponentY != null ) {
// 	    for( int i = 0; i < opponentX.length; i++ ) {
// 		displayBuffer.setPaint( Color.RED );
// 		displayBuffer.fillOval( Xx( (int) opponentX[i] ) - (OPPONENT_RADIUS - 1),
// 					Yy( (int) opponentY[i] ) - (OPPONENT_RADIUS - 1),
// 					2 * OPPONENT_RADIUS,
// 					2 * OPPONENT_RADIUS );
// 	    }
// 	}
//     }

//     /**
//      * Draw the ball
//      */
//     public void drawBall( int x, int y, boolean halo, int player ) {
// 	displayBuffer.setPaint( Color.ORANGE );
// 	displayBuffer.fillOval( Xx(x) - (BALL_RADIUS - 1),
// 				Yy(y) - (BALL_RADIUS - 1),
// 				2 * BALL_RADIUS,
// 				2 * BALL_RADIUS );
// 	if( halo ) {
// 	    displayBuffer.setPaint( Color.BLACK );
// 	    displayBuffer.drawOval(  Xx(x) - (BALL_RADIUS - 1),
// 				     Yy(y) - (BALL_RADIUS - 1),
// 				     2 * BALL_RADIUS,
// 				     2 * BALL_RADIUS );
// 	}

// 	displayBuffer.setPaint( Color.BLACK );
// 	displayBuffer.drawString( "" + player, Xx(x), Yy(y) );
//     }

//     /**
//      * Draw a force field on the field
//      */
//     public void drawBallForceField( Point2D.Double bff[][] ) {
// 	if( bff == null || bff[0] == null ) {
// 	    return;
// 	}

// 	int X_MAX = bff.length;
// 	int Y_MAX = bff[0].length;
// 	int nx1, ny1, nx2, ny2;

// 	for( int y = 0; y < Y_MAX; y++ ) {
// 	    for( int x = 0; x < X_MAX; x++ ) {
// 		nx1 = (int) (x * FIELD_X / X_MAX);
// 		ny1 = (int) (y * FIELD_Y / Y_MAX);
// 		nx2 = nx1 + (int) (FORCE_TILE_SIZE * bff[x][y].x);
// 		ny2 = ny1 + (int) (FORCE_TILE_SIZE * bff[x][y].y);
// 		//displayBuffer.setPaint( Color.GRAY );
// 		//displayBuffer.drawLine( Xx(nx1), Yy(ny1), Xx(nx2), Yy(ny2) );
// 		//displayBuffer.setPaint( Color.LIGHT_GRAY );
// 		//displayBuffer.drawOval( Xx(nx1), Yy(ny1), 1, 1 );

// 		//float angle = (float) (Math.atan2(bff[x][y].y,bff[x][y].x) / Math.PI);
// 		//debug( x + "," + y + " dx: " + bff[x][y].x + " dy: " + bff[x][y].y + 
// 		//   " - angle: " + angle );
// 		//displayBuffer.setColor( new Color( Math.abs(angle),
// 		//			   1.0f,
// 		//			   1.0f ) );
// 		//displayBuffer.setColor( new Color( (float) (2*Math.sqrt(bff[x][y].x*bff[x][y].x+
// 		//					      bff[x][y].y*bff[x][y].y)),
// 		//		   0.0f,
// 		//		   0.0f ) );
// 		//displayBuffer.setColor( new Color( (float) ((bff[x][y].x+1)/2),
// 		//			   0.0f,
// 		//			   (float) ((bff[x][y].y+1)/2) ));
// 		//displayBuffer.setColor( new Color( (float) Math.abs(bff[x][y].x),
// 		//			   0.0f,
// 		//			   (float) Math.abs(bff[x][y].y) ));
// 		float r, g, b;
// 		if( bff[x][y].x > 0 && bff[x][y].y > 0 ) {
// 		    r = (float) bff[x][y].x;
// 		    g = 0.0f;
// 		    b = (float) bff[x][y].y;
// 		} else if( bff[x][y].x <= 0 && bff[x][y].y > 0 ) {
// 		    r = 0.0f;
// 		    g = (float) -bff[x][y].x;
// 		    b = (float) bff[x][y].y;
// 		} else if( bff[x][y].x <= 0 && bff[x][y].y <= 0 ) {
// 		    r = (float) -bff[x][y].y;
// 		    g = (float) (-bff[x][y].x + -bff[x][y].y);
// 		    if( g > 1.0f ) g = 1.0f;
// 		    b = (float) -bff[x][y].y;
// 		} else {
// 		    r = (float) (bff[x][y].x + -bff[x][y].y);
// 		    if( r > 1.0f ) r = 1.0f;
// 		    g = (float) -bff[x][y].y;
// 		    b = (float) -bff[x][y].y;
// 		}
// 		displayBuffer.setColor( new Color( r, g, b ) );
// 		displayBuffer.fillRect( Xx(nx1), Yy(ny1), 5, 5 );
// 		//displayBuffer.setColor( Color.BLACK );
// 		//displayBuffer.drawLine( Xx(nx1), Yy(ny1), Xx(nx2), Yy(ny2) );

// 	    }
// 	}
//     }

    /**
     * Draw the localization particles for some player
     */
    public void drawParticles( Player player )
    {
	displayBuffer.setPaint( Color.WHITE );
	for( int i = 0; i < 100; i++ ) {
	    displayBuffer.fillOval( Xx( (int) player.getParticleX( i ) ),
				    Yy( (int) player.getParticleY( i ) ),
				    PARTICLE_SIZE,
				    PARTICLE_SIZE );
	}
    }

//     /**
//      * Draw the vision objects for some player
//      */
//     public void drawVisibleBeacons( Vector visibleBeacons, int x, int y ) {
// 	if( visibleBeacons == null) {
// 	    return;
// 	}
// 	displayBuffer.setPaint( Color.BLACK );
// 	for( int i = 0; i < visibleBeacons.size(); i++ ) {
// 	    int beaconNum = ((Integer)visibleBeacons.elementAt(i)).intValue();
// 	    switch(beaconNum) {
// 	    case 0:  // Beacon pink over blue
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx(0), Yy(0));
// 		break;	
// 	    case 1:  // Beacon pink over green
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx(0),
// 				       Yy((int)(FIELD_Y / 2)));
// 		break;	
// 	    case 2:  // Beacon pink over yellow
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx(0),
// 				       Yy((int)FIELD_Y));
// 		break;	
// 	    case 3:  // Yellow goal
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)(FIELD_X / 2)),
// 				       Yy((int)FIELD_Y));
// 		break;	
// 	    case 4:  // Beacon yellow over pink
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)FIELD_X),
// 				       Yy((int)FIELD_Y));
// 		break;	
// 	    case 5:  // Beacon green over pink
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)FIELD_X), 
// 				       Yy((int)(FIELD_Y / 2)));
// 		break;	
// 	    case 6:  // Beacon blue over pink
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)FIELD_X),
// 				       Yy(0));
// 		break;	
// 	    case 7:  // Blue goal
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)(FIELD_X / 2)), 
// 				       Yy(0));
// 		break;
	    
// 	    case 8:
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)((FIELD_X / 2) - (GOAL_WIDTH / 2))), 
// 				       Yy(0));
// 		break;
// 	    case 9:
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)((FIELD_X / 2) + (GOAL_WIDTH / 2))),
// 				       Yy(0));
// 		break;
// 	    case 10:
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)((FIELD_X / 2) - (GOAL_WIDTH / 2))),
// 				       Yy((int)FIELD_Y));
// 		break;
// 	    case 11:
// 		displayBuffer.drawLine(Xx(x), Yy(y),
// 				       Xx((int)((FIELD_X / 2) + (GOAL_WIDTH / 2))),
// 				       Yy((int)FIELD_Y));
// 		break;	

// 	    default:
// 		break;
// 	    }
// 	}
//     }

//     /**
//      * Draw the IR Sensor Data of a player
//      */

//     public void drawIRSensorData( int x, int y, double theta,
// 				  double percent ) {
// 	double IR_START = IR_MIN + HEAD_LENGTH;
// 	double IR_END = IR_MAX + HEAD_LENGTH;
// 	displayBuffer.setPaint( Color.RED );
// 	int ax = x + (int)Math.round(IR_START * Math.cos(theta));
// 	int ay = y + (int)Math.round(IR_START * Math.sin(theta));
// 	int bx = x + (int)Math.round(IR_END * Math.cos(theta));
// 	int by = y + (int)Math.round(IR_END * Math.sin(theta));
	
// 	/*
// 	 * These are kept as doubles to increase accuracy.  By rounding fewer
// 	 * times during the whole calculation, the final rounded values are
// 	 * more correct.
// 	 */

// 	double cx = x + 
// 	    ((IR_RANGE * percent) + IR_START) * Math.cos(theta);
// 	double cy = y + 
// 	    ((IR_RANGE * percent) + IR_START) * Math.sin(theta);
// 	int dx = (int)Math.round(cx + IR_WIDTH * 
// 				 Math.cos(theta + (Math.PI / 2.0)));
// 	int dy = (int)Math.round(cy + IR_WIDTH * 
// 				 Math.sin(theta + (Math.PI / 2.0)));
// 	int ex = (int)Math.round(cx + IR_WIDTH * 
// 				 Math.cos(theta - (Math.PI / 2.0)));
// 	int ey = (int)Math.round(cy + IR_WIDTH * 
// 				 Math.sin(theta - (Math.PI / 2.0)));
// 	displayBuffer.drawLine(Xx(ax), Yy(ay),
// 			       Xx(bx), Yy(by));
// 	displayBuffer.drawLine(Xx(dx), Yy(dy),
// 			       Xx(ex), Yy(ey));
// 	displayBuffer.setPaint( Color.BLACK );
// 	int cxi = (int)Math.round(cx);
// 	int cyi = (int)Math.round(cy);
// 	displayBuffer.fillOval(Xx(cxi), Yy(cyi), 2, 2);
//     }


    public void drawBall( int x, int y )
    {
	int radius = 6;
	displayBuffer.setPaint( Color.ORANGE );
	displayBuffer.fillOval( Xx(x)-radius, Yy(y)-radius, 
				2*radius, 2*radius );	
    }

    public void drawBallParticles( Ball ball )
    {
	displayBuffer.setPaint( Color.RED );
	for( int i = 0; i < 100; i++ ) {
	    displayBuffer.fillOval( Xx( (int) ball.getParticleX( i ) ),
				    Yy( (int) ball.getParticleY( i ) ),
				    PARTICLE_SIZE,
				    PARTICLE_SIZE );
	}	
    }

    /**
     * Draw a player at a certain location
     */
    public void drawPlayer( int x, int y, double theta, double phi, double confidence ) 
    {
	int[] coord = getThreePoints( Xx(x), 
				      Yy(y),
				      Tt(theta),
				      PLAYER_TRIANGLE_TIP,
				      PLAYER_BODY_SIZE );

	int[] xCoord = new int[3];
	int[] yCoord = new int[3];
	
	if ( confidence < 0 )
	    confidence = 1;

	/**
	 * Here we're going to split up the coordinate set into x and
	 * y coord.  That way, we can use handy builtin drawing
	 * stuff.
	 */
	for (int i = 0; i < 6; i +=2) {
	    xCoord[i/2] = coord[i];
	    yCoord[i/2] = coord[i+1];
	}

	displayBuffer.setPaint( new Color( 0.0F, (float) (1-confidence), 1.0F ) );
	displayBuffer.fillPolygon( xCoord, yCoord, 3 );
  
	// Draw the player's head.
	
	coord = getThreePoints( Xx(x),
				Yy(y),
				Tt(theta + phi ),
				PLAYER_TRIANGLE_CENTER,
				PLAYER_HEAD_SIZE );

	/**
	 * Here we're going to split up the coordinate set into x and
	 * y coord.  That way, we can use handy builtin drawing
	 * stuff.
	 */
	for (int i = 0; i < 6; i +=2) {
	    xCoord[i/2] = coord[i];
	    yCoord[i/2] = coord[i+1];
	}

	displayBuffer.setPaint( new Color( 0.0F, (float) (1-confidence), 1.0F ) );
	displayBuffer.fillPolygon( xCoord, yCoord, 3 );
  
	displayBuffer.setPaint( Color.BLACK );
	displayBuffer.fillOval( Xx(x)-3, Yy(y) - 3, 6, 6 );

    }
   

    /**
     * The following function computes the points of an isosceles
     * triangle with given top angle and base length, whose point
     * is at x,y and which is pointing at angle theta.
     */
    private int[] getThreePoints( int x, int y, double theta, 
				  int playerTriangle,
				  double baselength) {

	int[] answers = new int[6];
	if(playerTriangle == PLAYER_TRIANGLE_TIP) {
	    double a = (TOP_ANGLE/2.0);                  // PHI/2
	    double c = (baselength / 3.0);              // L/2
	    double d = c / Math.sin(a);                  // L/2 / sin PHI/2
	    double e = (theta + Math.PI / 2);
	    double f = e + a;
	    double g = e - a;
	    int p2x = (int)Math.round(x + d * Math.cos(f));
	    int p2y = (int)Math.round(y + d * Math.sin(f));
	    int p3x = (int)Math.round(x + d * Math.cos(g));
	    int p3y = (int)Math.round(y + d * Math.sin(g));
	    int p1x = (int)Math.round(x);
	    int p1y = (int)Math.round(y);
	    
	    answers[0] = p1x;
	    answers[1] = p1y;
	    answers[2] = p2x;
	    answers[3] = p2y;
	    answers[4] = p3x;
	    answers[5] = p3y;
	}
	else { // (playerTriangle == PLAYER_TRIANGLE_CENTER)
	    theta -= Math.PI / 2;
	    double a = (TOP_ANGLE/2.0);                  // PHI/2
	    double b = (1.0/Math.sin(a));                // CSC PHI/2
	    double c = (baselength / 3.0);              // L/3
	    double d = c * (1/Math.tan(a));              // L/3 TAN PHI/2
	    double tempX = Math.sqrt(baselength * baselength +
				     d * d)/2.0;
	    double tempPsi = Math.PI - Math.atan(baselength/d);
	    int p2x = (int)Math.round(x + tempX * Math.cos(theta + tempPsi));
	    int p2y = (int)Math.round(y + tempX * Math.sin(theta + tempPsi));
	    int p3x = (int)Math.round(x + tempX * Math.cos(theta - tempPsi));
	    int p3y = (int)Math.round(y + tempX * Math.sin(theta - tempPsi));
	    int p1x = (int)Math.round(x + d * Math.cos(theta));
	    int p1y = (int)Math.round(y + d * Math.sin(theta));
	    
	    answers[0] = p1x;
	    answers[1] = p1y;
	    answers[2] = p2x;
	    answers[3] = p2y;
	    answers[4] = p3x;
	    answers[5] = p3y;
	}
	return answers;
    }
   
// //     /**
// //      * Return true if the mouse is close enough to the specified player
// //      * that we want to turn on a halo around that player
// //      */
// //     private boolean haloed(int player, FieldState f) {
// // 	double a = (TOP_ANGLE/2.0);                  // PHI/2
// // 	double b = (1.0/Math.sin(a));                // CSC PHI/2
// // 	double c = (BASE_LENGTH / 3.0);              // L/3
// // 	double d = c * (1/Math.tan(a));              // L/3 TAN PHI/2
// // 	double tempX = Math.sqrt(BASE_LENGTH * BASE_LENGTH +
// // 				 d * d)/2.0;

// // 	double haloThreshold = tempX * HALO_RATIO;  // How close we
// // 						    // have to be
// // 						    // before we get a
// // 						    // halo.
// // 	double distance = Math.sqrt(Math.pow((Xx(mouseX) - Xx(f.pX[player])),2) +
// // 				    Math.pow((Yy(mouseY) - Yy(f.pY[player])),2));
	
// // 	return (distance < haloThreshold);
// //     }

// //     /**
// //      * Return true if the mouse is close enough to the ball that we
// //      * want it haloed
// //      */
// //     private boolean ballHaloed(FieldState f, int player) {
// // 	/**
// // 	 * How close we have to be before we get a halo.
// // 	 */
// // 	double haloThreshold =  BALL_RADIUS * HALO_RATIO;  
// // 	double distance = Math.sqrt(Math.pow((Xx(mouseX) - Xx(f.ballX[player])),2) +
// // 				    Math.pow((Yy(mouseY) - Yy(f.ballY[player])),2));

// //        	return (distance < haloThreshold);
// //     }



//     //////////////////////////////////////////////////////////


    /**
     * Initialize misc. variables
     */
    public void init() {

	// load the background image
	backgroundImage = Toolkit.getDefaultToolkit().getImage(imageFilename);
	mediaTracker = new MediaTracker(new Frame());
	try {
	    mediaTracker.addImage(backgroundImage, 0);
	    mediaTracker.waitForID(0);
	} catch( InterruptedException e ) {
	    debug( "Error loading " + imageFilename + ": " + e );
	}

	// for the initial display, just show the background image
	displayImage = backgroundImage;
	wholeImage = backgroundImage;
    }


    /**
     * Continue initialization after the frame is ready
     */
    public void init2() {

        Dimension d = getSize();

	// clear the area behind the field
        //displayBuffer.setPaint( Color.BLACK );
	//displayBuffer.fillRect( 0, 0, d.width, d.height );


	wholeImage = frame.createImage( d.width, d.height );
	wholeBuffer = (Graphics2D) wholeImage.getGraphics();
	wholeBuffer.setPaint( Color.BLACK );
	wholeBuffer.fillRect( 0, 0, d.width, d.height );

	// create a displayImage that we'll acutally display
	int width = backgroundImage.getWidth(null);
	int height = backgroundImage.getHeight(null);

	displayImage = frame.createImage( width, height );
	displayBuffer = (Graphics2D) displayImage.getGraphics();

	// draw the background image onto the displayImage
	displayBuffer.drawImage( backgroundImage, 0, 0, this );
    }




    public void update( Graphics g ) {
	paint( g );
    }



    public void paint(Graphics g) {
	if( wholeBuffer != null ) {
	    wholeBuffer.drawImage( displayImage, 0, IMAGE_VERT_OFFSET, this );
	}
	g.drawImage( wholeImage, 0, 0, this );
    }



    /**
     * xX() and yY() convert from canvas positions to real field positions.
     * Xx() and Yy() convert from real field positions to canvas positions.
     */
    public static int xX( int x ) {
	return (int) Math.round( x * FIELD_X / CANVAS_X - FIELD_X / 2 );
    }

    public static int yY( int y ) {
	return (int) Math.round( FIELD_Y / CANVAS_Y * ( CANVAS_Y / 2 - y ) );
    }

    public static int Xx( int X ) {
	return (int) Math.round( CANVAS_X / FIELD_X * ( X + FIELD_X / 2 ) );
    }

    public static int Yy( int Y ) {
	return (int) Math.round( CANVAS_Y / 2 - Y * CANVAS_Y / FIELD_Y );
    }

    public static double Tt( double theta ) {
	theta = -1 * theta;

	while( theta >= 2*Math.PI && theta < 0 ) {
	    if( theta >= 2*Math.PI ) {
		theta -= 2*Math.PI;
	    } else if( theta < 0 ) {
		theta += 2*Math.PI;
	    }
	}

	return( theta );
    }

    /**
     * scaleXx() and scaleYy() simply scale inputs from field
     * positions to canvas positions
     */
    public int scaleXx( int X ) {
	return( (int) (X * CANVAS_X / FIELD_X) );
    }

    public int scaleYy( int Y ) {
	return( (int) (Y * CANVAS_Y / FIELD_Y) );
    }


    ///////////////////////////////////////////////////////////////////////////////////




    /**
     * Send debugging output somewhere, depending on the class variable debug
     */
    public void debug( String s ) {
	if( debug ) {
	    System.out.println( s );
	}
    }













    /*



    //////////////////////////////////////////////////////////
    // variables that came with this class
    final static int maxCharHeight = 15;
    final static int minFontSize = 6;

    final static float dash1[] = {10.0f};
    final static BasicStroke dashed = new BasicStroke(1.0f, 
                                                      BasicStroke.CAP_BUTT, 
                                                      BasicStroke.JOIN_MITER, 
                                                      10.0f, dash1, 0.0f);

    FontMetrics fontMetrics;
    //////////////////////////////////////////////////////////




    FontMetrics pickFont(Graphics2D g2,
                         String longString,
                         int xSpace) {
        boolean fontFits = false;
        Font font = g2.getFont();
        FontMetrics fontMetrics = g2.getFontMetrics();
        int size = font.getSize();
        String name = font.getName();
        int style = font.getStyle();

        while ( !fontFits ) {
            if ( (fontMetrics.getHeight() <= maxCharHeight)
                 && (fontMetrics.stringWidth(longString) <= xSpace) ) {
                fontFits = true;
            }
            else {
                if ( size <= minFontSize ) {
                    fontFits = true;
                }
                else {
                    g2.setFont(font = new Font(name,
                                               style,
                                               --size));
                    fontMetrics = g2.getFontMetrics();
                }
            }
        }

        return fontMetrics;
    }



    public void draw_grid( Tile[][] grid ) {
	Graphics2D g2 = (Graphics2D) this.getGraphics();

        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        Dimension d = getSize();

	// create a temporary buffer to reduce flicker
	Image buffer = createImage( d.width, d.height );
	Graphics2D buffered_g2 = (Graphics2D) buffer.getGraphics();
	
	// clear the area behind the grid
        buffered_g2.setPaint( white );
	buffered_g2.fillRect( 0, 0, d.width, d.height );

        buffered_g2.setPaint( black );
	buffered_g2.drawRect( 1, 1, p.window_size_x-1, p.window_size_y-1 );

	// draw each tile on the grid
	int x, y;

	for( x = 0; x < p.grid_width; x++ ) {
	    for( y = 0; y < p.grid_height; y++ ) {
		
		// if the tile is occupied, draw a cell
		if( grid[x][y].occupied ) {

		    buffered_g2.setPaint( red );
		    buffered_g2.fillOval( (int) (x * p.window_size_x / p.grid_width),
					  (int) (y * p.window_size_y / p.grid_height),
					  4, 4 );
		}

	    }
	}

	// draw the final image to the real graphics buffer
	g2.drawImage( buffer, 0, 0, this );
    }
    */

//     /* Testing purposes */
//     public static void main( String[] args )
//     {
// 	for ( int i = -10; i < 10; i++ )
// 	    System.out.println( i + ", " + Xx(xX(i)) + ", " + xX(Xx(i)) );
// 	for ( int i = -10; i < 10; i++ )
// 	    System.out.println( i + ", " + Yy(yY(i)) + ", " + yY(Yy(i)) );
	
	
//     }

}
