import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.net.*;
import java.io.*;
import javax.swing.BorderFactory; 
import javax.swing.border.Border;
import javax.swing.border.TitledBorder;
import java.util.Vector;
import java.awt.image.*;
import java.awt.geom.Point2D;

public class Monitor implements ActionListener, 
				ItemListener,
				MouseListener,
				MouseMotionListener
{
    public static final String TITLEBAR_STRING = "Aibo Simulator - GUI";

    private Server server;

    private JTextField commandTextField;
    private JLabel statusLabel;
    private JTextArea statusTextArea;
    private JScrollPane statusScrollPane;
    private FieldCanvas fc;
    private JFrame frame;

    //private ReplayVCR rply;
    private int pDrag;
    private int dragY;
    private double dragT;
    private int dragButton;

    public Monitor( Server server )
    {
	this.server = server;
	initializeGui();
	pDrag = -1;
    }

    public void errorMsg( String msg )
    {
	System.err.println( msg );
    }


    private void createFileMenu( JMenuBar menuBar )
    {
	// file menu
	JMenu menu = new JMenu( "File" );
	menu.setMnemonic( KeyEvent.VK_F );
	menu.getAccessibleContext().setAccessibleDescription( "The File Menu" );
	menuBar.add( menu );

	// file->quit
	JMenuItem menuItem = new JMenuItem( "Quit", KeyEvent.VK_Q );
	menuItem.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_Q, ActionEvent.ALT_MASK ) );
	menuItem.getAccessibleContext().setAccessibleDescription( "Quit the program" );
	menuItem.addActionListener( this );
	menu.add( menuItem );
    }

    private void createMenuBar()
    {
        JMenuBar menuBar = new JMenuBar();
        menuBar.setOpaque( true );
        menuBar.setPreferredSize( new Dimension( 600, 20 ) );

	// add the menubar to the frame
        frame.setJMenuBar(menuBar);

	createFileMenu( menuBar );
    }

    public void initializeGui()
    {
	// disable eye candy around the window
        JFrame.setDefaultLookAndFeelDecorated(false);

        // create and set up the window
        frame = new JFrame( TITLEBAR_STRING );
        frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );

	createMenuBar();

	////////////////////////////////////////////////////
	// West side

	JPanel west1Panel = new JPanel();
	west1Panel.setLayout( new BoxLayout(west1Panel,BoxLayout.Y_AXIS) );
        frame.getContentPane().add( west1Panel, BorderLayout.WEST );

	/////////////////////
	// Command Panel

	JPanel commandPanel = new JPanel(new GridLayout(0,1));
	west1Panel.add( commandPanel );

	Border commandBorder = BorderFactory.createEtchedBorder();
	TitledBorder commandTitledBorder = BorderFactory.createTitledBorder( commandBorder, 
									     "Command Panel");
	commandTitledBorder.setTitleJustification(TitledBorder.RIGHT);
	commandPanel.setBorder( commandTitledBorder );

	JPanel commandControlPanel = new JPanel(new BorderLayout());
	commandPanel.add( commandControlPanel );

	// Command entry
	JPanel commandEntryPanel = new JPanel( new BorderLayout() );
	commandControlPanel.add( commandEntryPanel, BorderLayout.NORTH );
	
	JLabel commandEntryLabel = new JLabel( "Command:" );
	commandEntryPanel.add( commandEntryLabel, BorderLayout.WEST );

	commandTextField = new JTextField(20);
	commandTextField.setActionCommand("commandTextField");
	commandTextField.addActionListener(this);
	commandEntryPanel.add( commandTextField, BorderLayout.EAST );

	// Command send button
	JButton commandSendButton = new JButton("Send");
	commandSendButton.addActionListener(this);
	commandControlPanel.add( commandSendButton, BorderLayout.EAST );


	/////////////////////
	// Connection Panel

	GridBagLayout gridbag = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();

	JPanel connectionPanel = new JPanel(gridbag);
	west1Panel.add( connectionPanel );

	Border connectionBorder = BorderFactory.createEtchedBorder();
	TitledBorder connectionTitledBorder = BorderFactory.createTitledBorder( connectionBorder, 
									     "Connection Panel");
	connectionTitledBorder.setTitleJustification(TitledBorder.RIGHT);
	connectionPanel.setBorder( connectionTitledBorder );

	
	////////////////////
	// Status Panel
	JPanel statusPanel = new JPanel();
	statusPanel.setLayout( new BoxLayout(statusPanel,BoxLayout.Y_AXIS) );
	west1Panel.add( statusPanel );

	Border statusBorder = BorderFactory.createEtchedBorder();
	TitledBorder statusTitledBorder = BorderFactory.createTitledBorder( statusBorder, 
									     "Status Panel");
	statusTitledBorder.setTitleJustification(TitledBorder.RIGHT);
	statusPanel.setBorder( statusTitledBorder );

	statusTextArea = new JTextArea();
	statusTextArea.setFont(new Font("Serif", Font.PLAIN, 12));
	statusTextArea.setLineWrap(false);
	statusTextArea.setEditable(false);
	statusTextArea.setWrapStyleWord(true);
	statusScrollPane = new JScrollPane(statusTextArea);
	statusScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	statusScrollPane.setPreferredSize(new Dimension(250, 250));

	statusPanel.add( statusScrollPane );

	JButton statusClearButton = new JButton("Clear");
	statusClearButton.addActionListener(this);
	statusPanel.add( statusClearButton );


	////////////////////////////////////////////////////
	// East side

	JPanel east2Panel = new JPanel();
	east2Panel.setLayout(new BoxLayout(east2Panel, BoxLayout.Y_AXIS));
        frame.getContentPane().add( east2Panel, BorderLayout.EAST );

	fc = new FieldCanvas(frame);
	fc.setPreferredSize( new Dimension( 368, 528 ) );
	fc.addMouseListener( this );
	fc.addMouseMotionListener( this );
	east2Panel.add( fc );

	//rply = new ReplayVCR(this, frame);
	//east2Panel.add(rply);

	JPanel buttonPanel = new JPanel();

	JButton pauseButton = new JButton( "Pause" );
	pauseButton.addActionListener( this );
	buttonPanel.add( pauseButton );

	JButton resumeButton = new JButton( "Resume" );
	resumeButton.addActionListener( this );
	buttonPanel.add( resumeButton );

	JButton blindButton = new JButton( "Blind" );
	blindButton.addActionListener( this );
	buttonPanel.add( blindButton );

	JButton resetButton = new JButton( "Reset" );
	resetButton.addActionListener( this );
	buttonPanel.add( resetButton );

	east2Panel.add( buttonPanel );

        // display the main frame
        frame.pack();
        frame.setVisible(true);

	// continue initialization of the FieldCanvas
	fc.init2();

    }

    public void updateField()
    {
	fc.updateField( server.getPlayers(), server.getBall() );
    }
    

   // Event handling

    /**
     * Handle an action event
     */
    public void actionPerformed(ActionEvent e) 
    {
	// File -> Quit
	if( "Quit".equals(e.getActionCommand()) ) {
	    quit();
	} 

	if( "Pause".equals(e.getActionCommand()) ) {
	    if ( !server.isPaused() )
		server.pause();
	    else
		server.step();
	} 

	if( "Resume".equals(e.getActionCommand()) ) {
	    server.resume();
	} 

	if( "Blind".equals(e.getActionCommand()) ) {
	    server.blind();
	} 

	if( "Reset".equals(e.getActionCommand()) ) {
	    server.reset();
	} 
    }

    /**
     * This function is used when checkboxes are changed
     */
    public void itemStateChanged(ItemEvent e) 
    {
    }


    ////////////////////////////////////////////////////////////
    // Mouse stuff - the Server listens to mouse actions that
    // come from the FieldCanvas

    public void mouseClicked(MouseEvent e) {
	int x = e.getX();
	int y = e.getY();
	//System.out.println( "clicked: (" + x + "," + y + ")" );
    }


    // Invoked when the mouse enters a component.  
    public void mouseEntered(MouseEvent e) {
	//System.out.println( "entered" );
    }

    // Invoked when the mouse exits a component.  
    public void mouseExited(MouseEvent e) {
	//System.out.println( "exited" );

    }


    /**
     * When the mouse is pressed, draw some text containing the 
     * field position at some location
     */
    public void mousePressed(MouseEvent e) {
	int x = e.getX();
	int y = e.getY();
	dragY = y;
	double mx = FieldCanvas.xX( x );
	double my = FieldCanvas.yY( y );

	dragButton = e.getButton();
	    Vector players = server.getPlayers();
	    pDrag = -1;
	    double minDist = 80;
	    for ( int i = 0; i < players.size(); i++ ) {
		Player p = (Player) players.elementAt( i );
		double px = p.getX(); 
		double py = p.getY(); 
		double pt = p.getOrientation(); 
		double d = Math.sqrt( ( px - mx ) * ( px - mx ) +
				      ( py - my ) * ( py - my ) );
		if ( d < minDist ) {
		    minDist = d;
		    pDrag = i;
		    dragT = p.getOrientation();
		}
	    }
	    Ball b = server.getBall();
	    double px = b.getX(); 
	    double py = b.getY(); 
	    double d = Math.sqrt( ( px - mx ) * ( px - mx ) +
				  ( py - my ) * ( py - my ) );
	    if ( d < minDist ) {
		pDrag = 7748;
	    }
	   
	    //}
	    //else if( e.getButton() == MouseEvent.BUTTON3 ) {


	    //System.out.println( "pressed: (" + x + "," + y + ")" );
	    
	    // fc.drawPosition( fc.xX(x), fc.yY(y) );
	    //}

    }

    /**
     * When the mouse is released, repaint the field
     */
    public void mouseReleased(MouseEvent e) {
	    int x = e.getX();
	    int y = e.getY();

	    //System.out.println( "released: (" + x + "," + y + ")" );
	//fc.updateField(fs);
    }


    // Invoked when a mouse button is pressed 
    // on a component and then dragged.  	      
    public void mouseDragged(MouseEvent e) {
	    int x = e.getX();
	    int y = e.getY();

	    if ( pDrag == 7748 ) {
		Ball b = server.getBall();
		b.setPos( FieldCanvas.xX( x ),
			  FieldCanvas.yY( y ) );
		updateField();
	    }
	    else if ( pDrag >= 0 ) {
		Vector players = server.getPlayers();
		Player p = (Player) players.elementAt( pDrag );
		if ( dragButton == MouseEvent.BUTTON1 ) {
		    p.setPos( FieldCanvas.xX( x ),
			      FieldCanvas.yY( y ) );
		}
		else if ( dragButton == MouseEvent.BUTTON3 ) {
		    p.setAng( dragT + ( dragY - y ) * Math.PI / 100 ); 
		}
		updateField();
	    }

	    //System.out.println( "dragged: (" + x + "," + y + ")" );
    }

    // Invoked when the mouse button has been 
    // moved on a component (with no buttons down).
    public void mouseMoved(MouseEvent e) {
	int x = e.getX();
	int y = e.getY();

	//System.out.println( "moved: (" + x + "," + y + ")" );
      	//fc.updateMouse(fc.xX(x),fc.yY(y));
	//fc.updateField(fs);
    }

    private void quit()
    {
	System.exit( 0 );
    }
}

