/**
 * \file Constants.h
 * Entirely too many constants.
 * \todo Fix Desai's evilness in defining constants for really stupid things.
 * \todo Make some constants more precise, using the predefined
 *   constants in various math header files.
 */

#ifndef _CONSTANTS_H
#define _CONSTANTS_H

const int MAX_COMMAND_SIZE = 1024;

const int KICKFILE_CHARS = 64; // num of chars in the kickfile string
			       // of the MoveParamData struct

// decides how twitchy we'll be about setting "ws->frontFeedColliding"
const double FRONT_FEET_COLLIDING_THRESHOLD = 0.5;

// how close the ball can be to the wall before we start freakin out
const double BALL_CLOSE_TO_WALL_DISTANCE_THRESHOLD = 250;
// how misaligned we have to be in a bad wall situation before we start freakin out
const double BALL_CLOSE_TO_WALL_ANGLE_THRESHOLD = 60;

// the number of ball particles
const int NUM_BALL_PARTICLES = 20;

// numbers for other particles
const int NUM_ROBOT_PARTICLES = 100;
const double RESEED_IMPORTANCE = 0.02;

// Constants for random particles
const double RANDOM_PROPORTION = 0.05;

/**
 * These are for the IR data... I did not know where else to put them.
 */
const double PSD_MIN = 100000.0;
const double PSD_MAX = 900000.0;
const double CHEST_CLOSE_THRESHOLD = 140;
//const double CHEST_CLOSE_THRESHOLD = 160;
const int CHEST_CLOSE_COUNT = 5;
const int RING_LENGTH = 180;
const int IR_HISTORY_SIZE = 3;

const int STATE_NAME_MAX_SIZE = 39;
const int ROLE_NAME_MAX_SIZE = 39;

typedef double Distance;
typedef double Slope;
typedef double VAngle;
typedef double NumPixels;
typedef int ObjectID;
typedef double Speed;
typedef double Time;
typedef double AngularSpeed;

const double PI = 3.14159265358979323846264;
const double   G = 9.82e6;
const int BIG_INTEGER = 32767;  // "infinity"  --peggy
const double NO_VISION_DIST = 1e10;

const int LEFT = 0;
const int RIGHT = 1;
const int FORWARDS = 2;
const int BACKWARDS = 3;

//Dan:some constants about the location of a virtual point in the Aibo's neck
const double ShoulderVPointYDist = 95;
const double VPointCameraDist = 80;

//Dan: some more constants about the Aibo's body
const double FrontBackDist = 130;
const double LinkLength = 75;

//Dan: Strategy
const double SupportDist = 800;

//Dan: Keeping track of the ball:
const int BallQueueMax = 12;
const double BallDistMult = 1.1;  

//peggy: more keeping track of the ball
const int BallDistQueueMax = 5;
const int BALL_KNOWN_CONST = 5;
// Mohan::Number of frames the ball has to be seen for before it is
// accepted as being a ball...
const int NUM_FRAMES_SEEN = 2;

// Mohan::Minimum value of ball confidence above which the ball
// estimate is considered to be good enough to act on...
const double BALL_WELL_KNOWN_CONF = 0.60; //0.73;
const double LOW_BALL_CONF_THRESH = 0.60;
const double LOW_BALL_CONF =  0.40;
const double NO_BALL_CONF = 0.10;
const double TEAMMATE_POSE_CONF_THRESH = 0.80;
const double TEAMMATE_POSE_CONF_THRESH_LOW = 0.6;
const int MAX_BALL_DISTANCES = 35;
const double LOC_ESTIMATES_CONF_THRESH = 0.61;

// Some measurements of the field, and some objects on it.
const double FIELD_WIDTH = 3600;
const double FIELD_LENGTH = 5400;
const double GRASS_WIDTH = 4000;  // This is the area in which the aibos
const double GRASS_LENGTH = 6000; // are assumed to be.
const double FIELD_DIAGONAL = 6490;
const Distance MINX = -FIELD_WIDTH  / 2;
const Distance MINY = -FIELD_LENGTH / 2;
const Distance MAXX =  FIELD_WIDTH  / 2;
const Distance MAXY =  FIELD_LENGTH / 2;
const double FIELD_CENTER_X = 0;
const double FIELD_CENTER_Y = 0;
const double GOAL_WIDTH = 800;
const double BORDER_OFFSET = 150;

const double OUT_FIELD_X = 2*MINX;
const double OUT_FIELD_Y = 2*MINY;

// Mohan::Goal left and right edges also used for localization.
const Distance GOAL_LEFT_EDGE_X = -GOAL_WIDTH / 2;
const Distance GOAL_RIGHT_EDGE_X = GOAL_WIDTH / 2;
const Distance GOAL_CENTER_X = 0;

// Mohan's stuff from san fran
const double OPP_GOAL_LEFT_EDGE_X = GOAL_LEFT_EDGE_X;
const double OPP_GOAL_RIGHT_EDGE_X = GOAL_RIGHT_EDGE_X;
const double OPP_GOAL_CENTER_Y = FIELD_LENGTH/2;
const double OWN_GOAL_CENTER_Y = -FIELD_LENGTH/2;

// Mohan::Image and projection related constants.
const Distance FOCAL_LENGTH = 3.27;
const VAngle   ANGLE_OF_VISION_X = 56.9;
const VAngle   ANGLE_OF_VISION_Y = 45.2;
const double   RESOLUTION_X = 416;
const double   RESOLUTION_Y = 320;
const double   PIXELS_PER_MM = 58.6983;
const NumPixels MIN_IMAGE_X = 0;
const NumPixels MAX_IMAGE_X = 207;
const NumPixels MIN_IMAGE_Y = 0;
const NumPixels MAX_IMAGE_Y = 159;
const double   IMAGE_CENTER_X = MAX_IMAGE_X/2;
const double   IMAGE_CENTER_Y = MAX_IMAGE_Y/2;
const double   FOCALPIX  = ( PIXELS_PER_MM * FOCAL_LENGTH );
const double   CAM_ORG_BODY_CEN_Y = 67.5;
const double   CAM_ORG_BODY_CEN_Z = 19.5;
const double   REARLEG_SHOULDER_BODY_CEN_X = 62.5;
const double   REARLEG_SHOULDER_BODY_CEN_Y = 65.0;
const double   REAR_LEG_OFFSET1 = 9.0;
const double   REAR_LEG_OFFSET2 = 4.7;
const double   REAR_LEG_LENGTH1 = 69.5;
const double   REAR_LEG_LENGTH2 = 79.4;

const Distance BEACON_Z = 400.0;
const Distance BEACON_COLOR_Z = 200.0;
const Distance BEACON_RADIUS = 53.0;

const Distance GOAL_Z = 300.0;
const Distance GOAL_INTERIOR_X = 800.0;
const Distance GOAL_EXTERIOR_X = 810.0;
const Distance GOAL_DEPTH = 300.0;
const Distance FRONT_GOAL_DIST = 0.0;
const Distance GOAL_LINE_TO_END = GOAL_DEPTH + FRONT_GOAL_DIST;
const Distance PENALTY_X = 1300.0;
const Distance PENALTY_Y = 650.0;

const Distance CENTER_CIRCLE_RADIUS = 180.0;

const Distance WHITE_LINE_THICKNESS = 25.0;

const Distance ROBOT_BODY_X = 119.0;
const Distance ROBOT_BODY_Y = 80.0;

const Distance ROBOT_CAMERA_Z = 227.0;

const Distance BALL_DIAMETER = 84.0;

const Distance POINT_BEACON_Z = 0.0;

const double   BASE_OF_NECK_TO_BACK_OF_HEAD = 48*PIXELS_PER_MM;
const double   BACK_OF_HEAD_TO_NOSE = 64.42*PIXELS_PER_MM;

// approximate speed of the brain in cycles/sec
const int CYCLES_PER_SECOND = 20;

const Distance VARIANCE_LIMIT = 300.0;
const Distance WELL_LOCALIZED_VARIANCE = 350.0;

const double CAP_MULTIPLIER = 2.0;

// Mohan::Constants for defensive position...
const double DEF_DIST = 1000.0;
const double DEFPOST_X = FIELD_CENTER_X;    //\equiv 0;
const double DEFPOST_Y = -FIELD_LENGTH/4;   // \equiv -1500
const double DEF_NKOFF_X = FIELD_WIDTH/4;  // \equiv 900;
const double DEF_NKOFF_Y = -FIELD_LENGTH/4 - 150; // \equiv -1500;
const double DEF_ORIENT = 45;

// Mohan::Constants for the attacker trying to support the defender...
const double ATT_SUPDEF_X = FIELD_CENTER_X;    //\equiv 0;
const double ATT_SUPDEF_Y = FIELD_LENGTH/8;    // \equiv MAXY/4;


// Mohan::Constants for the keeper position...
const double KEEPER_X = 0;
const double KEEPER_Y = -2600;

// Mohan::Constants for the attacking positions...
const double ATTPOST_KOFF_X = FIELD_CENTER_X;              // \equiv 0;
const double ATTPOST_KOFF_Y = -400;
const double ATTPOST_NKOFF_X = FIELD_CENTER_X;             // \equiv 0;
const double ATTPOST_NKOFF_Y = -FIELD_LENGTH/4 - 150;      // \equiv -1500;

const double ATTSUP_KOFF_X = -FIELD_WIDTH/4;         // \equiv -900;
const double ATTSUP_KOFF_Y = -250;
const double ATTSUP_NKOFF_X = -FIELD_WIDTH/4;        // \equiv -900;
const double ATTSUP_NKOFF_Y = -FIELD_LENGTH/4 - 150; // \equiv -1500;
const double ATTSUP_ORIENT = -45;

const double MESSAGE_TIME_THRESH = 2.5;
const double MESSAGE_CONF_THRESH = 0.75;

// Mohan::Some more distance and variance limits used for field
// markers...
const double MAX_MARKER_DISTANCE = FIELD_DIAGONAL;
const double MAX_MARKER_ANGLE = PI;
const double MAX_MARKER_DIST_VARIANCE = 500.0;
const double MAX_MARKER_ANGLE_VARIANCE = 22 * ( PI / 180 );

// Mohan::Some constants for opponent robots...
const double INTER_OPPONENT_DIST = 600.0;
const double INTER_OPPONENT_X = 500.0;
const double INTER_OPPONENT_Y = 500.0;
const double MIN_OPP_STDEV = 150;
const double MAX_OPP_STDEV = 700;
const double OPP_VAR_DEGRADE_FACTOR = 90000.0;

// Mohan::More general constants...
const double MAX_STDX = 1300;
const double MAX_STDY = 1300;


// Mohan::Some more variances...
const double X_VARIANCE_THRESHOLD = 300.0;
const double Y_VARIANCE_THRESHOLD = 500.0;
const double THETA_VARIANCE_THRESHOLD = (23 * PI / 180);

// Mohan::Some more field constants...
const double GOAL_BOX_WIDTH = 1300.0;
const double GOAL_BOX_DEPTH = 650.0;
const double GOAL_BOX_X_MIN = -GOAL_BOX_WIDTH / 2;
const double GOAL_BOX_X_MAX = GOAL_BOX_WIDTH / 2;
const double OWN_GOAL_BOX_Y_MIN = -FIELD_LENGTH / 2;
const double OWN_GOAL_BOX_Y_MAX = -FIELD_LENGTH / 2 + GOAL_BOX_DEPTH;
const double OPP_GOAL_BOX_Y_MIN = FIELD_LENGTH / 2 - GOAL_BOX_DEPTH;
const double OPP_GOAL_BOX_Y_MAX = FIELD_LENGTH / 2;

const double MAX_PLAYER_X_CHANGE_PER_CYCLE = 60;
const double MAX_PLAYER_Y_CHANGE_PER_CYCLE = 60;
const double MAX_PLAYER_THETA_CHANGE_PER_CYCLE = 0.3;
const double MAX_BALL_X_CHANGE_PER_CYCLE = 100;
const double MAX_BALL_Y_CHANGE_PER_CYCLE = 100;
const double MAX_BALL_THETA_CHANGE_PER_CYCLE = 10;
const double BALL_VAR_DEGRADE_FACTOR = 90000;

// Mohan::Distance at which the ball is assumed to be when the robot
// "has" the ball...
const double HAVE_BALL_DIST = 200;

// Mohan::Distance below which the opponent goal is used by the robot
// for its reactive behavior...
const double CLOSE_TO_OPP_GOAL_DIST = 1300;
const double SEEN_GOAL_HEADING_TH = 10;

// Mohan::Distance below which the goal is considered as a landmark in
// localization...
const double CLOSE_GOAL_DIST = 650;

const int    NO_OBJECTS = 0;
const int    NUM_TEAM_MEMBERS = 4;
const int    NUM_OPPONENTS = 4;

// Mohan::Some constants to use with line pixels...
const int LINE_X = 29;
const int LINE_Y = 44;
const int LINE_ANG = 36;
const int X_RANGE = 100;
const int Y_RANGE = 100;
const int ANG_RANGE = 10;

// Mohan::Some more edge pixs constants...
const int NUM_EDGE_PIX = 100;
const int SCANOFFSET = 5;
const int FIELD_LINE_WIDTH = 25;
// const int FIELD_BORDER_WIDTH = 100;
const int MAX_LINE_DIST_THRESHOLD = 1800;
const int NUM_PIX_OBS = 3;
const int BLUE_GOAL_EDGE_PIX = 0;
const int YELLOW_GOAL_EDGE_PIX = 1;
// const int BORDER_EDGE_PIX = 2;
const int FIELD_EDGE_PIX = 2;



// Mohan::Thresholds and constants to determine line intersections
// from lines in the image plane...
const int MAX_CLUSTERS = 5;
const int LINE_MIN_PIX =  7;
const double DIST_THRESHOLD = 3.5; //2.5
const int MAX_SPREAD = 15;
const int MAX_INTERSECTS = 10;
const int LINE_ANGLE_THRESHOLD = 15;

// Mohan::Constants to use line intersections in localization...
const int MAX_LINE_INT_DIST = 1300;
const double INTERSECT_PROB_SLOPE = ( -0.25 / MAX_LINE_INT_DIST );
const double MAX_LINE_DIST = 1000;
const double LINE_PIX_SLOPE = ( -0.25/ MAX_LINE_DIST );

// const int NUM_BF_INT_POINTS = 6;
// const int NUM_BB_INT_POINTS = 4;
const int NUM_FF_INT_POINTS = 14; // only field intersects exist after
				  // border removal...

const ObjectID INVALID_ID = -1; 

// number of poses to keep in the smooth pose history  -peggy
const int POSE_HISTORY_SIZE = 100;

// Some mathematical constants
const double MAX_VAL = 1.0E10;
const int MAX_OBSERVATIONS = 50;


const double PERCENT = 1E-2;
const double MILLI = 1E-3;
const double MICRO = 1E-6;


// Some movement parameters.
const Speed X_DEGRADE_SPEED = 300;
const Speed Y_DEGRADE_SPEED = 300;
const AngularSpeed THETA_DEGRADE_SPEED = PI/2;

const Speed OPP_X_DEGRADE_SPEED = 2300;
const Speed OPP_Y_DEGRADE_SPEED = 2300;
const Speed OPP_THETA_DEGRADE_SPEED = PI/2;

const Speed BALL_X_SPEED = 1500;
const Speed BALL_Y_SPEED = 1500;
const Speed BALL_THETA_SPEED = 0; // Doesnt matter.


// for faster walks:
const double MAX_XV = 210; // pre-japan: 205; // 236;
const double SIDE_PINCH_XV = 165; // pre-japan: 175; // 134;
// learned large steps
//const double MAX_YV = 300;
// learned small steps, high turnover
const double MAX_YV = 390; // pre-japan: 375; // 387;
const double APPROACH_YV = 289; // This is just a guess
const double DRIBBLE_YV = 250;  // a guess for now  -peggy
const double PINCH_YV = 335;  // wow, actually not a guess!  -peggy
const double MAX_THETAV = 153; // pre-japan: 146; // 150;
const double BACK_LEG_TURN_THETAV = 30;
// Previously 40 - grossly overestimated...
const double SLOW_CHIN_PINCH_THETAV = 18; // lab: 38; //lisb: 36; // old field: 48.5;
// Previously 92 - slightly overestimated...
const double CHIN_PINCH_THETAV = 93; // pre-japan: 90; //lisb: 95; // old field: 106;
const double MIN_YV = -295; // pre-japan: -280; // Backwards is asymmetrical

/* // for faster walks: */
/* const double MAX_XV = 225; */
/* // learned large steps */
/* //const double MAX_YV = 300; */
/* // learned small steps, high turnover */
/* const double MAX_YV = 354; */
/* const double APPROACH_YV = 305; */
/* const double MAX_THETAV = 153; */
/* const double BACK_LEG_TURN_THETAV = 30; */
/* const double SLOW_CHIN_PINCH_THETAV = 43; */
/* const double CHIN_PINCH_THETAV = 95.5; //112.5; */
/* const double MIN_YV = -302; // Backwards is asymmetrical */

// for slower walks
//const double MAX_XV = 135;
//const double MAX_YV = 250;
//const double MAX_THETAV = 90;
//const double CHIN_PINCH_THETAV = 74;  // 83;
//const double BACK_LEG_TURN_THETAV = 17;
//const double MIN_YV = -170; // Backwards is asymmetrical

// Maintain the equation of the edges and lines in the ax+by+c = 0 format.
const double edgeA[] = {INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID, 1, 0, 1, 0};

const double edgeB[] = {INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID, 0, 1, 0, 1,};

const double edgeC[] = {INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			INVALID_ID,INVALID_ID,INVALID_ID, -MINX, -MAXY, -MAXX, -MINY,};

const double edgeGamma[] = {INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			    INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			    INVALID_ID,INVALID_ID,INVALID_ID,INVALID_ID,
			    INVALID_ID,INVALID_ID,INVALID_ID, 3*PI/2, PI, PI/2, 2*PI,};

// Some exceptions

const int NO_ROBOT_ID_EXCEPTION = 0;

// Some options

#define AIBO 0
#define LINUX 1

// Some global map constants.

const int MAX_OBJECTS = 20;

// Some movement parameters

const double X_QUANT = 500;
const double Y_QUANT = 500;
const double X_ERROR = 150;
const double Y_ERROR = 150;
const double THETA_ERROR = 10;
const double MIN_DX = 0;
const double MIN_DY = 0;
const double MIN_DTHETA = 0;

// Some configuration files

const char LAN_CONF_FILE[] = "/MS/OPEN-R/SYSTEM/CONF/WLANCONF.TXT";

// Some other system params.

const int IP_INCREMENT = 10;

// Joint and Sensor constants

enum JOINT_t {
  NECK_TILT,
  HEAD_PAN,
  HEAD_TILT,
  RFLEG_J1,
  RFLEG_J2,
  RFLEG_J3,
  LFLEG_J1,
  LFLEG_J2,
  LFLEG_J3,
  RRLEG_J1,
  RRLEG_J2,
  RRLEG_J3,
  LRLEG_J1,
  LRLEG_J2,
  LRLEG_J3,
  TAIL_PAN,
  TAIL_TILT,
  MOUTH,
  NUM_JOINTS
}; 

enum SENSOR_t {
  ACCELEROMETER_Y,
  ACCELEROMETER_X ,
  ACCELEROMETER_Z,
  HEAD_SENSOR,
  CHIN_SENSOR,
  HEAD_DISTANCE_SENSOR_NEAR, // 50 to 500 mm
  HEAD_DISTANCE_SENSOR_FAR,  // 200 to 1500 mm
  CHEST_DISTANCE_SENSOR,     // 100 to 900 mm
  RFLEG_SW,
  LFLEG_SW,
  RRLEG_SW,
  LRLEG_SW,
  BACK_SENSOR_FRONT,
  BACK_SENSOR_MID,
  BACK_SENSOR_REAR,
  WIRELESS_LAN,
  MICROPHONE,
  SPEAKER,
  NUM_SENSORS
};


// Mohan::Constants added for color learning on the robot...
const int LEARN_NONE = 0;
const int LEARN_GW = 1;
const int LEARN_YELLOWG = 2;
const int LEARN_YELLOWB = 3;
const int LEARN_PINK_Y = 4;
const int LEARN_PINK_B = 5;
const int LEARN_BLUEG = 6;
const int LEARN_BLUEB = 7;
const int LEARN_BALL = 8;
const int LEARN_OPP = 9;
const int LEARN_SAVE = 10;
const int LEARN_CLEAR = 11;
const int LEARN_CLEAR_RELOAD = 12;
const int LEARN_REVERSE = 13;
const int LEARN_RESET_Y = 14;
const int LEARN_RESET_L = 15;

// Mohan::Maximum values for the LAB color space channels - used in
// color learning tasks...
const int MAXVAL_L = 442;
const int MAXVAL_A = 90;
const int MAXVAL_B = 90;    


// Constant representing the various colors the robot is trained to
// recognize...
const int NUM_COLORS = 10;
const int PINK = 0;
const int YELLOW = 1;
const int BLUE = 2;
const int ORANGE = 3;
const int GREEN = 4;
const int RED = 5;
const int DBLUE = 6;
const int WHITE = 7;
const int FGREEN = 8;
const int BLACK = (NUM_COLORS - 1);
const int CLEARALL = 4;


#endif // _CONSTANTS_H
