#ifndef _PFLOCALIZATION_H
#define _PFLOCALIZATION_H

#include "../LocalizationInterface.h"
#include "Particles.h"


class PFLocalization: public LocalizationInterface 
{
 private:

  // At most 2% of particles are replaced by reseed values
  // Feel free to change this if you think that more or less would be better
  static const int MAX_RESEED = (int) (RESEED_IMPORTANCE * NUM_ROBOT_PARTICLES);

  // Observation history constants
  // Probably shouldn't be changed
  static const int HIST_CAPACITY = 20;
  static const Time MAX_TIME = 3.0;
  static const double MAX_DISTANCE = 150.0;
  static const AngDeg MAX_ROTATION = 10.0;
  static const double MIN_CERTAINTY = 0.55;

  // Should I just use a single estimate instead of using particles?
  // "-s" switch
  bool useSingleEstimate;
  
  // Should I turn off the use of reseeding estimates
  // "-n" switch
  bool noReseed;

  // Array of particles
  Particles *particles;

  // History for each landmark
  ObservationHistory **history;

  // Private methods
  void UpdateWorldStateSingleEstimate( WorldState *WS,
				       Vector2D vel, AngDeg angVel,
				       Time updateTime,
				       Observation histLandmarks[], 
				       int numHist );
  void UpdateWorldStateWithParticles( WorldState *WS,
				      Vector2D vel, AngDeg angVel,
				      Time updateTime,
				      Observation seenLandmarks[], 
				      int numSeen,
				      Observation histLandmarks[], 
				      int numHist );
  int getReseedEstimates( Observation *obs, int numObs, 
			  Particle *estimates );

  /* Define your methods here */

 public:
  PFLocalization( WorldState *WS, bool single, bool nr );
  ~PFLocalization();

  void UpdateWorldState( WorldState *WS );
  void getInfoMessage( char *buffer );
} ;

#endif //_PFLOCALIZATION_H
