#ifndef _PARTICLES_H
#define _PARTICLES_H

#include "Observations.h"
#include "Utils.h"

class Particle 
{
 protected:
  double M_prob;
  Point2D M_pos;
  AngDeg M_ang;

 public:
  Particle();

  Particle( double prob,
	    Point2D pos,
	    AngDeg ang,
	    Rectangle boundary );
  
  double getProbability();
  Point2D getPosition();
  AngDeg getAngle();

  void setProbability( double prob );
  void adjustProbability( double prob, double maxIncrease,
			  double maxDecrease );
  void degradeProbability( double factor );
  bool setPosition( Point2D pos, Rectangle boundary );
  void setAngle( AngDeg ang );
  void setParticle( double prob, Point2D pos, AngDeg ang,
		    Rectangle boundary );

   void moveRelative( Vector2D relativeDisplacement,
		     AngDeg rotation, Rectangle boundary );
  void moveGlobal( Vector2D globalDisplacement,
		   AngDeg rotation, Rectangle boundary );

  void randomWalk( double maxDistance, AngDeg maxRotation,
		   Rectangle boundary );

  void placeRandomly( Rectangle boundary );

  double getDistanceToPoint( Point2D p );
  AngDeg getAngleToPoint( Point2D p );

  void getInfoString( char *buffer, char ID,
		      Point2D pos, AngDeg ang );
} ;


typedef double Time;

class Particles
{
 protected:
  Particle *M_particles;
  Particle *M_particlesSwap;
  int M_numParticles;
  Rectangle M_boundary;
  Time M_lastMotionUpdate;

  double M_maxProbIncrease;
  double M_maxProbDecrease;
  double M_deltaDist;
  double M_deltaAng;
  double M_degradeFactor;

  double getProbabilitySum( Particle *p, int numParticles );

  void updateProbabilities( Observation* observation, int numObservations );
			    

  int sampleParticles( Particle *src, int srcSize, int numSlots, 
		       double &sum, Particle* dst, double weight );

 public:
  Particles();
  Particles( int numParticles,
	     double maxProbIncrease, double maxProbDecrease,
	     double deltaDist, double deltaAng,
	     double degradeFactor, Rectangle boundary );

  ~Particles();

  void updateObservation( Observation* observation, int numObservations,
			  Particle *reseedValues, int numReseedValues );

  void updateMotion( Vector2D velocity, AngDeg rotation, 
		     int numObservations, Time updateTime );
  void reset();

  Particle getPoseEstimate();
  int getNumParticles();

  void getInfoMessage( char *buffer );
} ;

#endif //_PARTICLES_H
