/**
 * \file Types.h 
 *
 * Definition of basic types
 */

//#include "Constants.h"
#include "Geometric.h"

#ifndef Types_H
#define Types_H

/**
 * Different search types for SearchForBallTask
 */
enum search_type { SEARCH_TURN_ONLY, SEARCH_BACK_UP, SEARCH_DISTANCE };

/**
 * Different type of formations
 */
enum formation_t {
  DEFENSIVE_FORMATION,
  OFFENSIVE_FORMATION,
  FORMATION_ILLEGAL
};

/**
 * Uniquely identifies each player
 */
enum PLAYER_ID_t {
  PLAYER_A,
  PLAYER_B,
  PLAYER_C,
  PLAYER_D,
  NUM_PLAYERS
};


/** 
 * Defines a number of different roles
 */
enum role_t {
    ATTACKER_ROLE,
    DEFENDER_ROLE,
    SUPPORTER_ROLE,
    KEEPER_ROLE,
    ROLE_ILLEGAL
};


/**
 * Enumerated type for team color
 */
enum side_t {
  SIDE_BLUE,
  SIDE_RED,
  SIDE_ILLEGAL
};


/**
 * Whether we have the kickoff or not.
 */
enum kickoff_t {
  KICKOFF_OWN,
  KICKOFF_OPPONENT,
  KICKOFF_NONE
};


/**
 * Enumerated type for initial behavior
 */
enum set_play_t {
  PLAY_KICK,
  PLAY_DASH,
  PLAY_NORMAL,
  PLAY_ILLEGAL
};


/**
 * State of the game
 */
enum game_state_t {
  GAME_STATE_READY,
  GAME_STATE_SET,
  GAME_STATE_PLAYING,
  GAME_STATE_PENALIZED,
  GAME_STATE_FINISHED,
  GAME_STATE_INITIAL
};



/**
 * A datatype representing the state of the AIBO with respect
 * to whether or not it has fallen.  If the AIBO has not
 * fallen, then use NOT_FALLEN.  Otherwise, use the value
 * associated with the direction that the AIBO has fallen.
 */
enum FALLEN_STATE {
  NOT_FALLEN,
  FELL_LEFT,
  FELL_RIGHT,
  FELL_FRONT,
  FELL_REAR
};

/**
 * A message to hold simple ping/pong information.
 */
struct PongMessage {
  PLAYER_ID_t youAre;
};


/**
 * A message to contain latest info on who is alive.
 */
struct AliveMessage {
  bool alive[NUM_PLAYERS];
};

/**
 * a message from a teammate about the ball
 */
struct BallMessage {
  // the teammate's ID
  //  int ID;
  // true if the teammate sees the ball right now
  bool seesBall;
  // Ball relative distance...
  double BallDist;
  // best estimate of ball location
  double ballX;
  double ballY;
  // estimate of how likely the ball is to be there
  double ballCertainty;
  int consecutiveNotBall;
  bool stopLookBall;
  bool goToBall;
  bool cantGetToBall;
};


struct PositionMessage {
  //Teammate's ID
  // int ID;
  role_t role;
  double xpos, ypos, theta, confidence;
  bool haveBall;
  game_state_t GameSt;
  MHGauss Opps[NUM_OPPONENTS];
};


/**
 * A message describing a command to a player
 */
struct CommandMessage {
  role_t role;
  bool goToBall;
  bool goBack;
  bool becomeCommander;
};
#endif
