/**
 * \file WorldState.cc
 *
 */
#include <stdio.h>

#include "WorldState.h"

//#include "Brain.h"
/*
#include <OPENR/OObject.h>
#include <OPENR/OSubject.h>
#include <OPENR/OObserver.h>
*/

/**
 * Thresholds for the sensors, they won't be activated for a custom sense
 * until the sensor goes above this level
 */
const double WorldState::senseThresholds[] = {
  -1, -1, -1, 10, 1, -1, -1, -1, 1, 1, 1, 1, 10, 10, 10, -1, -1, -1
};


//WorldState::WorldState(){throw NO_ROBOT_ID_EXCEPTION;}

WorldState::WorldState( PLAYER_ID_t MyID )//, CommunicationInterface* com )
  :bodyPos(0,0),  
   bodyVel(0,0),
   bodyAngVel(0),
   bodyTiltMove(0),
   frontHeight(0), 
   Cambody(4,4), Camneck(4,4), Body(4,4), Neckbody(4,4) {

  //chinpinchopenloop = false;
  //seeyellowtaskside = false;
  //seeyellowvisionside = false;
  aboutToMove = false;

  //Ben::
  isPasser = false;
  tmid = 0;
	
  weScored = theyScored = false;
  flashAllLeds = flashNoLeds = false;

  fall_direction = NOT_FALLEN;

  remainingCapacity = (unsigned short)0;

  currentRole = ATTACKER_ROLE;

  /**
   * This should be re-written to use something more intelligent.
   */
  if( MyID == PLAYER_D ) {
    kickoffTaker = true;
  } 
  else {
    kickoffTaker = false;
  }
  strategyGoToBall = true;

  // The boolean is set to true iff the robot (in behavior) stops and
  // stares at the ball because the robot and the ball are in the goal
  // box zone...
  stopLookBall = false;
  robotGoBack = true;

  // The boolean set to true iff the robot for some reason (out of
  // zone, obstructed etc) cannot get to the ball though it does see
  // it...
  cantGetToBall = false;
  KeeperAmI = false;
  
  //communication = com;

  lastBallInfoTime = 0;
  noBallInfoTime = 0;

  time = 0;
  previousTime = 0;

  headPan = 0;
  headTilt = 0;
  headTilt2 = 0;

  ballAngle = 0;
  ballDistance = FIELD_DIAGONAL;
  for (int i=0; i < BallDistQueueMax; i++) {
    lastBallDistance[i] = FIELD_DIAGONAL;
  }

  bodyAng = 0;
  Body_Tilt = 0.0;
  Body_Roll = 0.0;

  bodyCertainty = 0;

  ballLookingDirX = 0;
  ballLookingDirY = 1;
  ballLookingDirZ = 0;

  ballVel = Point2D(0,0);
  timeleft = 1000;
  xcontact = 0;

  // This says that we haven't set our ID using the network
  M_ID_set = false;
  M_ID = MyID;
  
  for(int i=0; i < NUM_PLAYERS; i++) {
    lastHeard[i] = 0;
    for(int j=0; j < NUM_PLAYERS; j++) {
      alive[i][j] = false;
    }
  }

  pinkBlueInVision = false;
  pinkYellowInVision = false;
  yellowPinkInVision = false;
  bluePinkInVision = false;
 
  resetObservations();

  firstPlay = PLAY_ILLEGAL;
  side = SIDE_BLUE;
  gameState = GAME_STATE_INITIAL;
  prevGameState = GAME_STATE_INITIAL;
  kickoff = KICKOFF_NONE;
//    for(int i = 0; i < NUM_PENALTY; i++) {
//      penalties[i] = 0;
//    }
  ownScore = 0;
  opponentScore = 0;

  // Mohan::Initialize drop-in variables...
  DropInTime = 10;  // Time since last drop-in set high...
  DropInTeam = -1;  // Team that caused drop-in set to be -ve...

  //sharedMMStatusInfo = NULL;

  ballBox = NULL;
  ballInVision = false;
  SureBall = false;

  ConsecutiveNotBall = 0;
  ConsecutiveBall = 0;
  BallKnown = false;
  haveBall = false;
  newCNB = 0;

  penaltyShot = false;

  globalBallEstimate = Point2D( 0, 0 );

  numBoundingBoxes = 0;
  blueGoalInVision = false;
  yellowGoalInVision = false;
  opponentsInVision = false;
  teammatesInVision = false;
  ownGoalInVision = false;
  opponentGoalInVision = false;
  
  for( int i = 0; i < NUM_JOINTS; i++ ) {
    joints[i] = 0.0;
  }
  for( int i = 0; i < NUM_SENSORS; i++ ) {
    sensors[i] = 0.0;
  }
  previousNeckTilt = previousHeadPan = previousHeadTilt = 0.0;
  headPanLambda = 0.29;
  //headPanLambda = 1.0;
  ffLambda = 0.15;
  frontFeetCount = 0.0;
  frontFeetColliding = false;

  closeChestIRCount = 0;
  notCloseChestIRCount = BIG_INTEGER;

  for( int i = 0; i < RING_LENGTH; i++ ) {
    IRRing[i] = (char) 0;
  }
  nearIRHistoryIdx = 0;
  farIRHistoryIdx = 0;
  nearIRHighCount = 0;
  farIRHighCount = 0;

  Neckbody.M[0][0] = 1; Neckbody.M[0][1] = 0; Neckbody.M[0][2] = 0; Neckbody.M[0][3] = 0;
  Neckbody.M[1][0] = 0; Neckbody.M[1][1] = 1; Neckbody.M[1][2] = 0; Neckbody.M[1][3] = CAM_ORG_BODY_CEN_Y;
  Neckbody.M[2][0] = 0; Neckbody.M[2][1] = 0; Neckbody.M[2][2] = 1; Neckbody.M[2][3] = CAM_ORG_BODY_CEN_Z;
  Neckbody.M[3][0] = 0; Neckbody.M[3][1] = 0; Neckbody.M[3][2] = 0; Neckbody.M[3][3] = 1;

  Body2Neck = Neckbody.Inverse(); 

  for (int i = 0; i < NUM_SENSORS; i++) {
    lapseCycles[i] = 0;
    lastDuration[i] = 0;
  }

  AccXHistId = 0;
  AccYHistId = 0;
  AccZHistId = 0;
  for( int i = 0; i < ACC_HIST_MAX; i ++ ) {
    AccXHist[i] = 0;
    AccYHist[i] = 0;
    AccZHist[i] = G;
  }


  particles = new Particles( 100,//NUM_ROBOT_PARTICLES,
			     0.1,//MAX_PROB_INCREASE,
			     0.06,//MAX_PROB_DECREASE,
			     150,//DELTA_DIST,
			     30,//DELTA_ANG,
			     0.9998,//DEGRADE_FACTOR,
			     GRASS_BOUNDARY );//FIELD_BOUNDARY );

  // Mohan::Set a point to represent a point outside the field
  // (team-relative global coordinates)...
  OutFieldPoint = Point2D( OUT_FIELD_X, OUT_FIELD_Y );

  // Mohan:: A dummy pose to set and return when the actual values do
  // not exist...
  DummyPose.Position = OutFieldPoint;
  DummyPose.Orientation = 0.0;
  DummyPose.Confidence = 0.0;

  // A dummy Opponent gaussian well outside the field...
  DummyOpp.Gms.mean = OutFieldPoint;
  DummyOpp.Gms.stdev = Point2D( MAXX, MAXY );
  DummyOpp.updated = false;
  DummyOpp.seen = false;

  // Mohan::Initialize opponents...
  for( int i = 0; i < NUM_OPPONENTS; i++ ) {
    Opponents[i] = DummyOpp;
  }

  // Mohan::Initialize opponents...
  for( int i = 0; i < NUM_TEAM_MEMBERS; i++ ) {
    Teammates[i] = DummyOpp;
  }

  // Mohan:;Initialize variable that stores number of opponents the
  // robot has (currently) retained estimates for...
  PrevNumOpps = 0;
  PrevNumTms = 0;

  // initialize the data concerning our teammates
  for( int i = 0; i < NUM_PLAYERS; i++ ) {
    teammateState[i].ID = (PLAYER_ID_t)i;
    teammateState[i].seesBall = false;
    teammateState[i].BallDist = NO_VISION_DIST;
    teammateState[i].ballPosition.setPoint(0,0);
    for( int j = 0; j < NUM_OPPONENTS; j++ ) {
      teammateState[i].Opps[j] = DummyOpp;
    }
  }

  // Mohan:: A pose stored to be used in behavior when the robot does
  // not get visual information and moves/does not move...
  uncappedSmoothPose.Position.setPoint(0.0,0.0);
  uncappedSmoothPose.Orientation = 0.0;
  uncappedSmoothPose.Confidence = 0.0;
  SmoothPose = uncappedSmoothPose;

  // create four points that represent the four corners of the field
  LL.setPoint( MINX, MINY );
  UL.setPoint( MINX, MAXY );
  LR.setPoint( MAXX, MINY );
  UR.setPoint( MAXX, MAXY );

  // create four lines that represent the edges of the field
  lbot = Line2D::makeLineFromTwoPoints( LL, LR );
  ltop = Line2D::makeLineFromTwoPoints( UL, UR );
  llft = Line2D::makeLineFromTwoPoints( UL, LL );
  lrgt = Line2D::makeLineFromTwoPoints( UR, LR );

  LineNum = -1;

  // A cheap hack to estimate if the robot cannot get to ball because
  // it is being blocked by other robots, walls etc...
  BallDistID = 0;
  for( int i = 0; i < MAX_BALL_DISTANCES; i ++ ) {
    BallDistsArr[i] = NO_VISION_DIST;
  }

  // Mohan::Variables initialized for the task of keeping the robots
  // (other than the keeper) from entering their own goal box zones...

  // Initialize the end points of the lines...
  LineAP1 = Point2D( GOAL_BOX_X_MIN, OWN_GOAL_BOX_Y_MIN ); // ( 950, 100 )
  LineAP2 = Point2D( GOAL_BOX_X_MIN, OWN_GOAL_BOX_Y_MAX ); // ( 950, 620 )

  LineBP1 = Point2D( GOAL_BOX_X_MIN, OWN_GOAL_BOX_Y_MAX ); // ( 950, 620 )
  LineBP2 = Point2D( GOAL_BOX_X_MAX, OWN_GOAL_BOX_Y_MAX ); // ( 1950, 620 )

  LineCP1 = Point2D( GOAL_BOX_X_MAX, OWN_GOAL_BOX_Y_MAX ); // ( 1950, 620 )
  LineCP2 = Point2D( GOAL_BOX_X_MAX, OWN_GOAL_BOX_Y_MIN ); // ( 1950, 100 )

  LineDP1 = Point2D( GOAL_BOX_X_MIN, OPP_GOAL_BOX_Y_MAX ); // ( 950, 4300 )
  LineDP2 = Point2D( GOAL_BOX_X_MIN, OPP_GOAL_BOX_Y_MIN ); // ( 950, 3780 )

  LineEP1 = Point2D( GOAL_BOX_X_MIN, OPP_GOAL_BOX_Y_MIN ); // ( 950, 3780 )
  LineEP2 = Point2D( GOAL_BOX_X_MAX, OPP_GOAL_BOX_Y_MIN ); // ( 1950, 3780 )

  LineFP1 = Point2D( GOAL_BOX_X_MAX, OPP_GOAL_BOX_Y_MIN ); // ( 1950, 3780 )
  LineFP2 = Point2D( GOAL_BOX_X_MAX, OPP_GOAL_BOX_Y_MAX ); // ( 1950, 4300 )

  // Initialize the lines that form the goal boxes...
  LineA = Line2D::makeLineFromTwoPoints( LineAP1, LineAP2 );
  LineB = Line2D::makeLineFromTwoPoints( LineBP1, LineBP2 );
  LineC = Line2D::makeLineFromTwoPoints( LineCP1, LineCP2 );
  LineD = Line2D::makeLineFromTwoPoints( LineDP1, LineDP2 );
  LineE = Line2D::makeLineFromTwoPoints( LineEP1, LineEP2 );
  LineF = Line2D::makeLineFromTwoPoints( LineFP1, LineFP2 );


  // Data collection to estimate the function between actual and
  // measured landmark distances...
  CollectData = false;

  for( int i = 0; i < 4500; i ++ ) {
    MeasuredDistHist[i] = 0.0;
  }
  Measdistcount = 0;

  // Values for stability test calculations...
  StartStabilityTest = false;
  NumArrVals = 0;

  NumTrials = 0;
  HistCleaned = false;
  PrevTime = 0;

  // Values for the accelerometer - collision experiments...
  NumAccVals = 0;
  NumCalVals = 0;
  accsumx = 0;  accsumy = 0;  accsumz = 0;
  tslavs = 0;

  CollClass = 0;

  V_x = 0; V_y = 0; V_z = 0;
  R_xy = 0; R_yz = 0; R_zx = 0;
  NewCollVals = false;


  // A variable for the movie generation (from the robot's point fo
  // view)...
  TimeToStop = false;

  // Mohan::Variables for Mazda's surveillance task...
  markerinframe = false;
  min_marker_id = OBJ_PINK_BLUE;
  min_marker_offset = 90;
  min_marker_angle = 0;
  angleft = 90;
  angright = 90;


  // Mohan::Parameters for the color learn task...
  RunColorLearnTask = false;
  ColorLearnOption = LEARN_NONE;
  ColorClearOption = 0;
  LearnGoalColorFromBeacon = false;

  SpecialCMapUpdate = false;
  LookForGoalBlob = false;

  GoalInView = false;
  GoalIVCount = 0;

  // Default value for the color space variable...
  ColorSpace = 0;

  // Variables that control the change in color channel variation
  // limits...
  ChangedLimits = false;
  ColorChanged = 0;
  YLimit = 39;
  CbLimit = 31;
  CrLimit = 31;

  ObjectPanAngle = 0;
  ObjectTiltAngle = 0;

  // Mohan::Parameters that allow the robot to learn the color better
  // by localizing to special positions on the field...
  ReinforceColor = false;
  ReinforcePos = -1;
  ReinforceCMapUpdate = false;

//    // Nate: initialize 2005 GameController data
//    gc_player_idx = MyID;
//    gc_team_idx = TEAM_BLUE;
//    gc_opp_team_idx = TEAM_RED;

  // New ball representation information initialized...
  RelBall = OutFieldPoint;
  GlobBall = OutFieldPoint;
  RelBallHeading = 0.0;
  RelBallDistance = 0.0;

  RelBallGauss.mean = OutFieldPoint;
  RelBallGauss.stdev = Point2D( 1e3, 1e3 );
  BallConfidence = 0.0;

  RecentlyStartedDegrade = false;

  BallVelComputed = false;
  NumBallVals = 0;
  BallValsIdx = 0;

  BallRelVel = Point2D( 0, 0 );
  BallRelVelDir = 0;

  SpecialVelUpdate = false;

  //GCtime = 10.0;
  LastSignalTime = 0.0;

  UsingMergedEstimate = false;

  pointStartTime = 0;
  supportPoint = Point2D( GOAL_BOX_X_MAX, OPP_GOAL_BOX_Y_MAX );
  resetSupportPoint = true;

  Body_Height = 0.0;

  // Mohan::Flag for learning colors a bit - Dr. Lavender's demo...
  specialLearn = false;
  AskForBall = false;
  specialDone = false;
  SP_GWLearnt = false;
  SP_BallColLearnt = false;

  // TEMPO FLAG FOR DEBUGGING...
  ChangedTargetPoint = false;

  prevpictime = -1;
  manpiccap = false;

  prevBMesgTime = -1;
  prevPMesgTime = -1;
}




/**
 * Ask if we think a certain robot is alive.
 */
bool WorldState::isAlive(PLAYER_ID_t ID) {
  if(ID == M_ID) {
    return true;
  }
  return alive[M_ID][ID];
}

bool WorldState::isAlive(int ID) {
  return isAlive((PLAYER_ID_t)ID);
}



/**
 * This is for general-purpose data initialization activities.
 */
void WorldState::GetData() {
//    if( communication != NULL ) {
//      UpdateNetworkStatus();
//      // Mohan::receive stuff from teammates...
//      UpdateFromTeammates();
//    }
  // At this point, the joints and sensor arrays have been populated -
  // good enough to compute the collision-detection values...
  if( COLLISION_DETECT ) {
    GenerateTestSample();
    if( NewCollVals ) {
      InCollision = CollidingAmI();
    }
  }
}



/**
 * This is for general-purpose data-processing activities.
 */
void WorldState::Update() {
  UpdateBallInfo();
  Vector2D Ball( getBallDistance(), getBallAngle(), POLAR );
  //UpdateRelBalls( Ball );
  //ComputeBallVel();

  // Mohan:: April 15, 2005-An attempt to maintain an unimodal
  // distribution for the ball position in the robot's coordinate
  // frame - currently we shall try a Gaussian as a representation of
  // the ball position relative to the robot...
  UpdateBallGaussian();

  // check for updates for all of the custom senses
  updateSenses();

  // only let the smooth pose change by so much each cycle
  updateSmoothPoseWithCap();

  // use ball particles to update some higher-level information about
  // where we think the ball is
  //UpdateBallEstimate();

  // Mohan::New means of updating the ball position as a Gaussian...
  UpdateGlobalBall();

  // Mohan::time to set the "cantGetToBall" flag. For now it happens
  // iff the robot is a defender and the ball is in the top third of
  // the field and the robot is sufficiently far away from the ball as
  // then the defender cannot go to the ball even if it is the
  // closest...
  if( currentRole == DEFENDER_ROLE && BallInTopThird() 
      && BallYDisplacement() >= 1500 ) {
    cantGetToBall = true;
  }
  else {
    cantGetToBall = false;
  }

  // Now that the robot has "sensed" and "merged" everything, send
  // updated info to teammates...
  //UpdateToTeammates();

  // Mohan::Reset the previoustime variable to reflect the current
  // time value...
  previousTime = time;
}


/**
 * Mohan::Function returns true iff the ball is moving away from the
 * robot - relative motion:-)...
 */
bool WorldState::BallMovingAway() {
  if( ( BallRelVelDir > 30 && BallRelVelDir < 150 && 
	RelBallGauss.mean.getY() > 200 ) ||
      ( BallRelVelDir > -30 && BallRelVelDir < -150 &&
	RelBallGauss.mean.getY() < - 200 ) ) {
    cout << "\n Ball moving away... \n";
    return true;
  }
  return false;
}



/**
 * Mohan::Function returns true if the ball is seen to be
 * approaching the robot - currently this implies that the ball
 * velocity is above a threshold and the direction of approach is
 * towards the robot...
 */
bool WorldState::BallApproaching() {
  int vel_thresh = 900, high_vel_thresh = 1800;
  int low_ang = -130;
  int high_ang = -50;
  int buffer = 20;

  // If the ball velocity is suitably high and the direction of
  // approach is within certain negative thresholds, the ball is
  // approaching...
  if( BallRelVel.getMagnitude() > vel_thresh &&
      BallRelVelDir > low_ang && BallRelVelDir < high_ang ) {
    return true;
  }
  // Else if the ball velocity is really high and the angle is
  // sufficiently negative, consider it still as aproaching the
  // robot...
  else if( BallRelVel.getMagnitude() > high_vel_thresh &&
	   BallRelVelDir > low_ang-buffer && 
	   BallRelVelDir < high_ang+buffer ) {  
    return true;
  }
  else {
    return false;
  }
}




/**
 * Mohan::Function that assigns a standard deviation to the relative
 * ball position as a function of the ball distance...
 */
Point2D WorldState::GetBallStdevs( double distance ) {
  // A cubic for generating the standard deviation as a function of
  // the ball distance...
  double a0 = -20.205902;
  double a1 = 0.816216;
  double a2 = -0.000417;
  double a3 = 0.0000001068;

  double stdev = max( ( a0 + a1 * distance + a2 * pow( distance, 2 ) +
			a3 * pow( distance, 3 ) ), 100 );

  return Point2D( stdev, stdev );
}



/**
 * Mohan::Function that attempts to calculate the ball velocity and
 * direction and if successful returns true - computation happens if
 * enough data vals stored in array...
 */
bool WorldState::ComputeRelBallVel() {
  static double Reduction_factor = 0.968936;
  int MinNumVals = 8;
  if( NumBallVals < MinNumVals ) {
    return false;
  }

  Point2D p1, p2;
  Point2D sum1( 0, 0 );
  Point2D sum2( 0, 0 );
  // Same old strategy of dividing up the queue into two halves and
  // computing the average in each to obtain the overall velocity and
  // direction...
  int loop1 = NumBallVals/2;
  int loop2 = NumBallVals - loop1;
  int i = BallValsIdx;

  double start_time = BallTimeVals[i]; 

  // "Older" half position average...
  for( int j = 0; j < loop1; j++ ) {
    sum1 += BallValsQueue[ i ];
    i = (i+1) % NumBallVals;
  }
  p1 = sum1/loop1;

  double stop_time = BallTimeVals[i];

  // "Newer" half position average...
  for( int j = 0; j < loop2; j ++ ){
    sum2 += BallValsQueue[i];
    stop_time = BallTimeVals[i-1];
    i = (i+1) % NumBallVals;
  }
  p2 = sum2/loop2;

  // The velocity is the displacement divided by time - we do time/2
  // because we are considering the average times in each half to
  // compute velocity...

  // Note: if the velocity update has to occur while the robot does
  // not see the ball, the direction needs to change each time but the
  // velocity magnitude is just a factor of the previous one...
  if( !SpecialVelUpdate ) {  
    BallRelVel = ( p2 - p1 ) / ( 0.5 * fabs( stop_time - start_time ) );
  }
  else {
    BallRelVel *= Reduction_factor;
  }

  BallRelVelDir = ( p2 - p1 ).getDirection();
  return true;
}



/**
 * Mohan::Function performs a motion update of the ball in relative
 * frame of the robot based on the movement commands that the robot
 * executes - also update the ball position history to be used for
 * velocity computation...
 */
void WorldState::Motion_Update_Gaussian() {
  double timediff = getTime() - previousTime;
  Vector2D vel = getVelocity();
  AngDeg angVel = getAngularVelocity();

  // Based on robot's movement, obtain the change in pose in robot's
  // coordinate frame...
  Vector2D deltaPos = vel * timediff;
  AngDeg deltaAng = angVel * timediff;

  // Mohan::There are two options to consider here: When the robot
  // "has" the ball but cannot see it, such as when it is doing a
  // chinpinchturn or when it is dribbling, the ball position shall be
  // set to be a constant distance ahead of the robot (in robot
  // relative frame). In all other cases, it shall eb updated to
  // compensate for the robot's motion...
  if( haveBall || ControlDribble ) {
    // Set ball position to a fixed point in front of the robot in
    // robot-relative frame of reference...
    RelBallGauss.mean = Point2D( 0, HAVE_BALL_DIST );
    // Suitably change the values in the ballposition queue too so
    // that it does not have stale values when the "haveBall" flag is
    // reset...
    for( int i = 0; i < NumBallVals; i ++ ) {
      BallValsQueue[i] = RelBallGauss.mean;
    }
  }
  else {
    Point2D CurrBallMean = RelBallGauss.mean;
    RelBallGauss.mean = CurrBallMean.globalToRelative( deltaPos, deltaAng );

    // Update the ball positions stored, if any, to be used to compute
    // velocity...
    for( int i = 0; i < NumBallVals; i ++ ) {
      BallValsQueue[i].globalToRelative( deltaPos, deltaAng );
    }

    // Next, we need to compute the ball velocity and direction - the
    // case wherein we need to compute velocity even when the robot does
    // not see the ball is taken care of in the "ComputeRelBallVel()"
    // function...
    if( ComputeRelBallVel() ) {
      BallVelComputed = true;
      // If velocity computation is successful, update the current known
      // ball position suitably based on time...
      RelBallGauss.mean += Point2D( BallRelVel.getMagnitude(),
				    BallRelVelDir,
				    POLAR ) * timediff;

//       cout << "\n Ball (Vel,dir): (" << BallRelVel.getMagnitude()
// 	   << "," << BallRelVelDir << ")\n";
      
      // Some computation, as before, on when the ball is going to
      // reach the robot...
      double ball_xvel = BallRelVel.getX();
      double ball_yvel = BallRelVel.getY();

      double ball_x = RelBallGauss.mean.getX();
      double ball_y = RelBallGauss.mean.getY();
      
      timeleft = 1000;
      xcontact = 0;
      if( BallRelVelDir < 0 && ball_y > 0 ) {
	timeleft = fabs( ball_y/ball_yvel ); // timeleft is positive.
	xcontact = ball_x + ball_xvel*timeleft;
      }

    }
  }

  if( BALLDEBUG ) {
    cout << "\n After motion update: (" << RelBallGauss.mean.getX()
	 << "," <<  RelBallGauss.mean.getY() << ")\n";
  }
}



/**
 * Mohan::Function returns true if the input Gaussian represents a
 * valid point on the field - for now only a stdev check...
 */
bool WorldState::ValidGauss( Gauss g ) {
  return ( g.stdev.getX() < 1e4 &&
	   g.stdev.getY() < 1e4 );
}



/**
 * Mohan::Function that causes the robot to update the relative ball
 * position (in the robot's coordinate frame) using an unimodal
 * Gaussian...
 */
void WorldState::UpdateBallGaussian() {
  BallVelComputed = false;

  Gauss CurrBallGauss;
  // First, update the current ball estimate (Gaussian) corresponding
  // to the motion of the robot...
  Motion_Update_Gaussian();

  // Generate the Gaussian corresponding to the current ball using the
  // current frame's ball estimate, if any...
  if( ballInVision ) {
    double balldist = getBallDistance();
    double ballangle = getBallAngle(); 
    CurrBallGauss.mean.setPoint( balldist, ballangle, POLAR);
    CurrBallGauss.stdev = GetBallStdevs( balldist );

    if( isnan( CurrBallGauss.mean.getX() ) ||
	isnan( CurrBallGauss.mean.getY() ) ) {
      cout << "\n ZERO BY ZERO ERROR 1 " << CurrBallGauss.mean << endl;
    }

    // Reset flag - flag when true implies that we had just started
    // degrading the Gaussian...
    RecentlyStartedDegrade = false;

    // If the robot was previously engaged in updating position based
    // on velocity even when the ball was not seen, it should clear
    // the entire array when a new ball is seen again (and reset the
    // flag of course:-)...
    if( SpecialVelUpdate ) {
      for( int i = 0; i < BallQueueMax; i ++ ) {
	BallValsQueue[ i ] = Point2D( 0, 0 );
	BallTimeVals[ i ] = 0;
      }
      NumBallVals = 0;
      BallValsIdx = 0;
      SpecialVelUpdate = false;
    }

    // Store current ball position in the queue to compute velocity...
    BallTimeVals[ BallValsIdx ] = time;
    BallValsQueue[ BallValsIdx++ ] = CurrBallGauss.mean;
    NumBallVals = ( ( NumBallVals >= BallQueueMax ) ? BallQueueMax : 
		    NumBallVals+1 );

    // Reset index suitably...
    if( BallValsIdx >= BallQueueMax ) {
      BallValsIdx = 0;
    }

    if( BALLDEBUG ) {
      cout << "\n Old (" << RelBallGauss.mean.getX() << ","
	   << RelBallGauss.mean.getY() << "," << RelBallGauss.stdev.getX()
	   << "," << RelBallGauss.stdev.getY() << ") and New ("
	   << CurrBallGauss.mean.getX() << "," << CurrBallGauss.mean.getY()
	   << "," << CurrBallGauss.stdev.getX() << "," << CurrBallGauss.stdev.getY()
	   << ") \n";
    }

    RelBallGauss = GaussianMerge( CurrBallGauss, RelBallGauss );

    RelBallGauss.stdev.setX( max( RelBallGauss.stdev.getX(), 100 ) );
    RelBallGauss.stdev.setY( max( RelBallGauss.stdev.getY(), 100 ) );

    if( isnan( RelBallGauss.mean.getX() ) ||
	isnan( RelBallGauss.mean.getY() ) ) {
      cout << "\n ZERO BY ZERO ERROR 2 " << RelBallGauss.mean << endl;
      RelBallGauss.mean = OutFieldPoint;
      RelBallGauss.stdev = Point2D( 1e5, 1e5 );
    }

    if( BALLDEBUG ) {
      cout << "\n Rel Ball Merged (" << RelBallGauss.mean.getX() << "," 
	   << RelBallGauss.mean.getY() << ", and, " << RelBallGauss.stdev.getX()
	   << "," << RelBallGauss.stdev.getY() << ") \n"; 
    }
  }
  // If not, degrade the known position of the ball, if it is known
  // that is;-)...
  else if( ValidGauss( RelBallGauss ) && !haveBall ) {
    RelBallGauss = GaussianDegrade( RelBallGauss, previousTime, time, 
				    MAX_BALL_X_CHANGE_PER_CYCLE, 
				    MAX_BALL_Y_CHANGE_PER_CYCLE, 
				    MAX_BALL_THETA_CHANGE_PER_CYCLE, 
				    BALL_VAR_DEGRADE_FACTOR );
    if( BALLDEBUG ) {
      cout << "\n Rel Ball Degraded (" << RelBallGauss.mean.getX() << "," 
	   << RelBallGauss.mean.getY() << ", and, " << RelBallGauss.stdev.getX()
	   << "," << RelBallGauss.stdev.getY() << ") \n"; 
    }

    if( !RecentlyStartedDegrade ) {
      RecentlyStartedDegrade = true;
    }

    // If the ball has been recently lost from view, we can still
    // perform the velocity updates as long as the probability is
    // reasonably high...
    if( RecentlyStartedDegrade && 
	BallConfidence >= LOW_BALL_CONF_THRESH ) {
      SpecialVelUpdate = true;
    }
    else {
      SpecialVelUpdate = false;
    }

    if( !SpecialVelUpdate ) {
      // If ball is not seen and velocity can no longer be used
      // safely, reset the ball velocity array...
      for( int i = 0; i < BallQueueMax; i ++ ) {
	BallValsQueue[ i ] = Point2D( 0, 0 );
	BallTimeVals[ i ] = 0;
      }
      NumBallVals = 0;
      BallValsIdx = 0;
    }
  }

  RelBall = RelBallGauss.mean;
  // also compute a heading for that relball
  RelBallHeading = normalizeAngle( RelBall.getDirection() - 90 );
  RelBallDistance = RelBall.getMagnitude();
}



/**
 * Mohan::Funtion calculates the global ball position based on the
 * Gaussian estimation technique...
 */
void WorldState::UpdateGlobalBall() {
  UsingMergedEstimate = false;

  Point2D relBallPos = RelBallGauss.mean;
  if( relBallPos == OutFieldPoint ) {
    globalBallEstimate = OutFieldPoint;
    BallConfidence = 0.0;
  }

  else {
    globalBallEstimate = relBallPos.relativeToGlobal( SmoothPose.Position, SmoothPose.Orientation );
    // compute the angle to the ball in field coordinates
    globalBallAngle = normalizeAngle((globalBallEstimate - SmoothPose.Position).getDirection() - 90);
    
    double a0 = 0.9916256;     // 0.9916256
    double a1 = 0.000080114;   // 8.0114 e-5
    double a2 = -0.00000215426;// -2.1543 e-6
    double a3 = 0.00000000124; // 1.24 e-9
    double stdevx =  RelBallGauss.stdev.getX();

    if( stdevx >= MAX_STDX ) {
      BallConfidence = NO_BALL_CONF;
    }
    else {
      BallConfidence = minmax( 0.0, ( a0 + a1 * stdevx + a2 * pow( stdevx, 2 ) +
				      a3 * pow( stdevx, 3 ) ), 1.0 );
      
      if( isnan( globalBallEstimate.getX() ) ||
	  isnan( globalBallEstimate.getY() ) ) {
	cout << "\n ZERO BY ZERO ERROR 2 " << globalBallEstimate << endl;
	globalBallEstimate = OutFieldPoint;
      }

      if( BALLDEBUG ) {
	cout << "\n using own ball estimate:("
	     << globalBallEstimate.getX() << ","
	     << globalBallEstimate.getY() << ","
	     << BallConfidence << ")\n";
      }
    }
  }

  if( BALLDEBUG ) {
    cout << "\n Global Ball: (" << globalBallEstimate.getX() << ","
	 << globalBallEstimate.getY() << " and " << BallConfidence << ") \n";
  }

  // Now if the robot's confidence of the ball is really low, it takes
  // the teammates estimates into account to arrive at an estimate...
  if( BallConfidence <= LOW_BALL_CONF ) {
    double BallWeights[NUM_TEAM_MEMBERS];
    Point2D TeamBallEsts[NUM_TEAM_MEMBERS];
    
    bool BallInfoReceived = false;
    int NumMessages = 0;

    double sumdist = 0.0;
    for( int i = 0; i < NUM_TEAM_MEMBERS; i ++ ) {
      // Mohan::If this teammate isn't the robot under consideration
      // and its ball message isn't too old, use it to obtain the
      // merged ball estimate - first sum the ball distances so that
      // individual weights can be calculated...
      if( teammateState[i].ID != M_ID &&
	  ( time - teammateState[i].lastPositionMessageTime <= MESSAGE_TIME_THRESH ) &&
	  ( time - teammateState[i].lastBallMessageTime <= MESSAGE_TIME_THRESH ) ) {
	sumdist += teammateState[i].BallDist;

	TeamBallEsts[i] = Point2D( teammateState[i].ballPosition.getX(), 
				   teammateState[i].ballPosition.getY() );
	BallInfoReceived = true;
	++ NumMessages;
      }
    }
    
    // Now to merge the teammates' estimates...
    Point2D MergedBallEst(0,0);
    double sumweights = 0.0;
    double maxprob = 0;
    // If an usable teammate ball estimate exists...
    if( BallInfoReceived ) {
      for( int i = 0; i < NUM_TEAM_MEMBERS; i ++ ) {
	if( teammateState[i].ID != M_ID &&
	    ( fabs(time - teammateState[i].lastPositionMessageTime) <= MESSAGE_TIME_THRESH ) &&
	    ( fabs(time - teammateState[i].lastBallMessageTime) <= MESSAGE_TIME_THRESH ) ) {
	  // Consider three things as weights:ball confidence, position
	  // confidence and ball distance...
	  double distweight = ( ( NumMessages > 1 ) ? 
				( 1 - teammateState[i].BallDist/sumdist ) :
				1.0 );
	  // cout << "\n Distweight: " << distweight << endl;
	  distweight = minmax(0.0, distweight, 1.0);
	  
	  BallWeights[i] = ( ( teammateState[i].ballCertainty + 
			       teammateState[i].pose.Confidence + 
			       distweight ) / 3.0 );
	  // cout << "\n BallWeights: " << BallWeights[i] << endl;
	  maxprob = max( maxprob, BallWeights[i] );
	  
	  sumweights += BallWeights[i];
	  // Calculate weighted average of the ball positions sent in by
	  // the teammates...
// 	  cout << "\n Input ball: " << teammateState[i].ballPosition 
// 	       << "<" << TeamBallEsts[i] << endl;
	  MergedBallEst += ( teammateState[i].ballPosition * BallWeights[i] );
	  // cout << "\n Merged to get " << MergedBallEst << endl;
	}
      }
      
      if( sumweights <= 0.0 ||
	  MergedBallEst == Point2D(0, 0 ) ) {
	if( BALLDEBUG ) {
	  cout << "\n DIVIDE BY ZERO - DUMBO MOHAN "
	       << sumweights << MergedBallEst << "\n";
	}
      }
      else {
	// Perform overall ball position computation, by dividing by
	// overall weight...
	globalBallEstimate = MergedBallEst/sumweights;
	// For now overall ball confidence is the same as the best
	// overall weight of one of the teammates...
	BallConfidence = minmax(0, maxprob, 1);
	UsingMergedEstimate = true;
	if( BALLDEBUG ) {
	  cout << "\n using merged teammate ball estimate:(" 
	       << globalBallEstimate.getX() << ","
	       << globalBallEstimate.getY() << ","
	       << BallConfidence << ")\n";
	}
      }
    }
  }
  
  GlobBall = globalBallEstimate;
  
//   GenericDir tempdir;
//   // Debug check...
//   if( BallCloseToFieldBorder( tempdir ) ) {
//     cout << "\n Well - what do I do now... \n";
//   }
}



/**
 * Update smoothPose, not adjusting any component more than
 * some maximum amount
 */
void WorldState::updateSmoothPoseWithCap() {

  // update the history
  for( int i = POSE_HISTORY_SIZE - 2; i >= 0; i-- ) {
    SmoothPoseLocationHistory[i+1] = SmoothPoseLocationHistory[i];
  }
  SmoothPoseLocationHistory[0] = SmoothPose.Position;

  double diffX = uncappedSmoothPose.Position.getX()- SmoothPose.Position.getX();
  double diffY = uncappedSmoothPose.Position.getY() - SmoothPose.Position.getY();
  AngDeg diffT = normalizeAngle( uncappedSmoothPose.Orientation - SmoothPose.Orientation );

//   double maxX = MAX_XV * timeInFrame * CAP_MULTIPLIER;
//   double maxY = MAX_YV * timeInFrame * CAP_MULTIPLIER;
//   double maxT = MAX_THETAV * timeInFrame * CAP_MULTIPLIER;

//   if( fabs(diffX) > maxX ) {
//     if( diffX > 0 ) diffX = maxX;
//     else diffX = -maxX;
//   }

//   if( fabs(diffY) > maxY ) {
//     if( diffY > 0 ) diffY = maxY;
//     else diffY = -maxY;
//   }

//   if( fabs(diffT) > maxT ) {
//     if( diffT > 0 ) diffT = maxT;
//     else diffT = -maxT;
//   }

  SmoothPose.Position.setPoint( SmoothPose.Position.getX() + diffX,
				SmoothPose.Position.getY() + diffY );

  // make sure SmoothPose doesn't go off the field
  // if( SmoothPose.Position.getX() < 0 ) SmoothPose.Position.setX( 0 );
//   else if( SmoothPose.Position.getX() > MAXX ) SmoothPose.Position.setX( MAXX );
//   if( SmoothPose.Position.getY() < 0 ) SmoothPose.Position.setY( 0 );
//   else if( SmoothPose.Position.getY() > MAXY ) SmoothPose.Position.setY( MAXY );

  SmoothPose.Orientation = normalizeAngle( SmoothPose.Orientation + diffT );
}



Time WorldState::getTime()
{
  return time;
}

void WorldState::setPose( Point2D pos, AngDeg ang, double certainty )
{
  // Use team-centric coordinates
  if ( side == SIDE_RED ) {
    bodyPos = pos * -1;
    bodyAng = normalizeAngle( ang + 180 );
  }
  else { // side == SIDE_BLUE
    bodyPos = pos;
    bodyAng = ang;
  }
  bodyCertainty = certainty;
}

void WorldState::setVelocities( Vector2D vel, AngDeg angVel )
{
  bodyVel = vel;
  bodyAngVel = angVel;
}

void WorldState::resetObservations()
{
  numObservations = 0;
  NumOpponents = 0;
  NumTeammates = 0;
}

void WorldState::addObservation( ObjectT obj, double dist, 
				 AngDeg ang, double certainty ) {
  if( numObservations >= MAX_OBSERVATIONS ) {
    cout << "Error: cannot add another observation" << endl;
    return;
  }

  Observation *obs = &observations[ numObservations ];
  obs->setObject( obj );
  obs->setDistance( dist );
  obs->setHeading( ang );
  obs->setCertainty( certainty );
  numObservations++;
}

/**
 * Mohan::A separate function to add opponent observation - similar to
 * the one above for adding other observations of landmarks etc. This
 * makes processing of opponent observations easier...
 */
void WorldState::addOppObservation( ObjectT obj, double dist,
				    AngDeg ang, double certainty ) {
  if( NumOpponents >= NUM_OPPONENTS ) {
    cout << "Error: cannot add another opponent observation" << endl;
    return;
  }
  Observation *OppObs = &OppObservations[ NumOpponents ];
  OppObs->setObject( obj );
  OppObs->setDistance( dist );
  OppObs->setHeading( ang );
  OppObs->setCertainty( certainty );
  NumOpponents++;
}


/**
 * Mohan::A separate function to add teammate observation - similar to
 * the one above for adding other observations of landmarks etc. This
 * makes processing of teammate observations easier...
 */
void WorldState::addTmsObservation( ObjectT obj, double dist,
				    AngDeg ang, double certainty ) {
  if( NumTeammates >= NUM_TEAM_MEMBERS ) {
    cout << "Error: cannot add another teammate observation" << endl;
    return;
  }
  Observation *TmsObs = &TmsObservations[ NumTeammates ];
  TmsObs->setObject( obj );
  TmsObs->setDistance( dist );
  TmsObs->setHeading( ang );
  TmsObs->setCertainty( certainty );
  NumTeammates++;
}



AngDeg WorldState::getHeadPan() 
{
  return headPan;
}

AngDeg WorldState::getHeadTilt() 
{
  return headTilt;
}

Point2D WorldState::getLocation()
{
  return bodyPos;
}

AngDeg WorldState::getOrientation()
{
  return bodyAng;
}

double WorldState::getPoseCertainty() {
  return bodyCertainty;
}

/**
 * Mohan::Function that returns the entire pose of the robot -
 * position + orientation + certainty...
 */
Pose WorldState::getMyPose() {
  Pose pose;
  pose.Position = bodyPos;
  pose.Orientation = bodyAng;
  pose.Confidence = bodyCertainty;
  return pose;
}


int WorldState::getObservations( Observation *obs )
{
  for ( int i = 0; i < numObservations; i++ )
    obs[ i ] = observations[ i ];
  return numObservations;
}

Vector2D WorldState::getVelocity() {
  return bodyVel;
}

AngDeg WorldState::getAngularVelocity() {
  return bodyAngVel;
}

double WorldState::distanceToPoint( Point2D p )  {
  // Mohan::Maybe we want to use the Smoothpose - as below??
  //return SmoothPose.Position.getDistanceTo( p );
  return bodyPos.getDistanceTo( p );
}


/**
 * Takes a point, and returns an angle from our x-axis to point
 */
AngDeg WorldState::angleToPoint( Point2D p ) {
  // Mohan::Maybe we want to use the Smoothpose - as below??
  Vector2D relativeDisplacement = p.globalToRelative( bodyPos,
						      bodyAng );
//    Vector2D relativeDisplacement = p.globalToRelative( SmoothPose.Position,
//  						      SmoothPose.Orientation );
//   Vector2D relativeDisplacement =
//     p.globalToRelative( bodyPos, bodyAng );
 //  cerr << bodyPos << ", " << bodyAng << " is " 
//        << relativeDisplacement.getDirection()
//        << " from " << p << endl;
  return relativeDisplacement.getDirection();
}

void WorldState::getInfoMessage( char *buffer ) {
  char estBuffer[ 1000 ];
  Point2D pos = bodyPos;
  AngDeg ang = bodyAng;
  if ( side == SIDE_RED ) {
    pos *= -1;
    ang = normalizeAngle( ang + 180 );
  }
  sprintf( estBuffer, "(e %d %d %d %d %d %d)", (int) pos.getX(), 
	   (int) pos.getY(), (int) ang, (int) bodyCertainty, 
	   (int) headPan, (int) headTilt );
  
  strcpy( buffer, "(info " );
  strcat( buffer, estBuffer );
}


/**
 * Return true if something has set off the close-chest sensor for
 * more than a certain number of cycles
 */
bool WorldState::objectCloseToChest() {
  return( closeChestIRCount > CHEST_CLOSE_COUNT );
}

/**
 * If this returns true, it indicates that the ball disappeared from
 * vision at roughly (roughness controlled by gapSize) the time we
 * started to sense an object close to the chest IR sensor.
 * This is a good indicator of whether we have NEWLY gotten very close
 * to the ball.  To check if we continue to be near the ball, use
 * BallStillFelt().  -peggy
 */
bool WorldState::BallFirstFelt() {
  const int gapSize = 3;
  return (!BallKnown && (closeChestIRCount + gapSize > ConsecutiveNotBall) );
}

/**
 * If this returns true, it indicates that we are continuing to get
 * intermittent confirmation from the chest IR sensor that the ball
 * is very close.
 * This is a good indicator of whether we STILL are very close to the
 * ball, given that we once acquired it.  To check if we have just 
 * gotten very close to the ball for the first time, use BallFirstFelt().
 * -peggy
 */
bool WorldState::BallStillFelt() {
  const int framesUntilLost = 13;
  return (notCloseChestIRCount < framesUntilLost);
}



/**
 * Add a value to the array of saved near IR values
 */
void WorldState::addNearIRValue( double d ) {
  // keep track of the number of max vals
  if( d >= 500 ) {
    nearIRHighCount++;
  } 

  // keep a running average of the non-max vals
  else {
    nearIRHighCount = 0;
    nearIRHistory[ nearIRHistoryIdx ] = d;
    nearIRHistoryIdx++;
    if( nearIRHistoryIdx >= IR_HISTORY_SIZE ) {
      nearIRHistoryIdx = 0;
    }
  }
}

/**
 * Add a value to the array of saved far IR values
 */
void WorldState::addFarIRValue( double d ) {
  if( d >= 1500 ) {
    farIRHighCount++;
  } else {
    farIRHighCount = 0;
    farIRHistory[ farIRHistoryIdx ] = d;
    farIRHistoryIdx++;
    if( farIRHistoryIdx >= IR_HISTORY_SIZE ) {
      farIRHistoryIdx = 0;
    }
  }
}

/**
 * Return an average of the values in the near IR history array
 */
double WorldState::nearIRAverage() {
  double sum = 0.0;
  for( int i = 0; i < IR_HISTORY_SIZE; i++ ) {
    sum += nearIRHistory[i];
  }
  return( sum / IR_HISTORY_SIZE );
}

/**
 * Return an average of the values in the far IR history array
 */
double WorldState::farIRAverage() {
  double sum = 0.0;
  for( int i = 0; i < IR_HISTORY_SIZE; i++ ) {
    sum += farIRHistory[i];
  }
  return( sum / IR_HISTORY_SIZE );
}


/**
 * Mohan::Add a value to the array of saved X-axis Accelerometer
 * values...
 */
void WorldState::NewAccXVal( double x ) {
  AccXHist[ AccXHistId ] = x;
  AccXHistId ++;
  if( AccXHistId >= ACC_HIST_MAX ) {
    AccXHistId = 0;
  }
}


/**
 * Mohan::Add a value to the array of saved Y-axis Accelerometer
 * values...
 */
void WorldState::NewAccYVal( double y ) {
  AccYHist[ AccYHistId ] = y;
  AccYHistId ++;
  if( AccYHistId >= ACC_HIST_MAX ) {
    AccYHistId = 0;
  }
}


/**
 * Mohan::Add a value to the array of saved X-axis Accelerometer
 * values...
 */
void WorldState::NewAccZVal( double z ) {
  AccZHist[ AccZHistId ] = z;
  AccZHistId ++;
  if( AccZHistId >= ACC_HIST_MAX ) {
    AccZHistId = 0;
  }
}


/**
 * Mohan::Return an average of the values in X-axis Accelerometer
 * History array...
 */
double WorldState::AccXAvg() {
  double xsum = 0;
  
  for( int i = 0; i < ACC_HIST_MAX; i ++ ) {
    xsum += AccXHist[i];
  }
  return( xsum / ACC_HIST_MAX ); 
}


/**
 * Mohan::Return an average of the values in Y-axis Accelerometer
 * History array...
 */
double WorldState::AccYAvg() {
  double ysum = 0;
  
  for( int i = 0; i < ACC_HIST_MAX; i ++ ) {
    ysum += AccYHist[i];
  }
  return( ysum / ACC_HIST_MAX ); 
}

/**
 * Mohan::Return an average of the values in X-axis Accelerometer
 * History array...
 */
double WorldState::AccZAvg() {
  double zsum = 0;
  
  for( int i = 0; i < ACC_HIST_MAX; i ++ ) {
    zsum += AccZHist[i];
  }
  return( zsum / ACC_HIST_MAX ); 
}


/**
 * Mohan::Returns ID of the robot...
 */
PLAYER_ID_t WorldState::getMyID() {
  return M_ID;
}


/**
 * Mohan::Returns the ID of the orange ball...
 */
int WorldState::getBallID() {
  return OBJ_BALL;
}


int WorldState::getTeamID() {
  if( side == SIDE_BLUE ){
    return OBJ_BLUE_ROBOT;
  }
  else {
    return OBJ_RED_ROBOT;
  }
}


int WorldState::getOpponentID() {
  if( side == SIDE_BLUE ){
    return OBJ_RED_ROBOT;
  }
  else {
    return OBJ_BLUE_ROBOT;
  }
}


int WorldState::getOwnGoalID() {
  if( side == SIDE_BLUE ){
    return OBJ_BLUE_GOAL;
  }
  else {
    return OBJ_YELLOW_GOAL;
  }
}


int WorldState::getOpponentGoalID() {
  if( side == SIDE_BLUE ){
    return OBJ_YELLOW_GOAL;
  }
  else {
    return OBJ_BLUE_GOAL;
  }
}


/**
 * Mohan::Just returns the ball bounding box for you to play
 * with :-)...
 */
BoundingBox* WorldState::getBallBox() {
  if( !ballInVision ) {
    return NULL;
  }
  return ballBox;
}

void WorldState::UpdateNetworkStatus() {
//    // Respond to any ping messages
//    while( communication->NumberofDataObjects(PING_MESSAGE) > 0 ) {
//      recvData* data = communication->ReceiveData(PING_MESSAGE);
//      if( data == NULL ) {
//        cout << "Error: received null data for ping message" << endl;
//      } else {
//        PLAYER_ID_t fromID = (PLAYER_ID_t)data->DogFrom;
//        if(DEBUG) {
//  	cout << "Just received a ping from " << fromID << "." << endl;
//        }
//        lastHeard[fromID] = 0;
//        PongMessage msg;
//        msg.youAre = fromID;
//        communication->SendData( PONG_MESSAGE,
//  			       reinterpret_cast<byte *>(&msg),
//  			       sizeof(msg),
//  			       (sndDOGNUMBER)fromID);
//        delete data;
//      }
//    }

//    // If we don't have an ID, try to get one.
//    if( !M_ID_set ) {
//      if( communication->NumberofDataObjects(PONG_MESSAGE) > 0 ) {
//        // Use a pong message to determine our ID
//        recvData* data = communication->ReceiveData(PONG_MESSAGE);
//        if( data == NULL ) {
//  	cout << "Error: received null data for pong message" << endl;
//        } else { 
//  	PLAYER_ID_t fromID = (PLAYER_ID_t)data->DogFrom;
//  	lastHeard[fromID] = 0;
//  	PongMessage *msg = reinterpret_cast<PongMessage *>(data->data);
//  	M_ID = msg->youAre;
//  	M_ID_set = true;
//  	if(DEBUG) {
//  	  cout << "Setting my ID to " << M_ID << endl;
//  	}
//  	delete data;
//        }
//        // Now clear out the rest of the stack.  In the future, this
//        // could be used to determine who heard us.
//        while( communication->NumberofDataObjects(PONG_MESSAGE) > 0 ) {
//  	recvData* data = communication->ReceiveData(PONG_MESSAGE);
//  	if( data == NULL ) {
//  	  cout << "Error: received null data for pong message" << endl;
//  	} else { 
//  	  delete data;
//  	}
//        }
//      } else { // Request some pong messages by sending some pings
//        static int pingTimer = 0;
//        if (pingTimer == CYCLES_BETWEEN_PINGS) {
//  	if(DEBUG) {
//  	  cout << "Don't know who I am, pinging." << endl;
//  	}
//  	int msg = 0;
//  	communication->SendData( PING_MESSAGE,
//  				 reinterpret_cast<byte *>(&msg),
//  				 sizeof(msg),
//  				 ALL);
//  	pingTimer = 0;
//        }
//        pingTimer++;
//      }
//    } 
//    else {  
//      // We have an ID so we can do normal network things.
//      // Now we tally up what we've heard to determine who is up and who
//      // is not.  If someone has not been heard from in a while, we send
//      // them a ping. If it's been even longer, we declare them not up
//      // anymore (and continue to ping them).
//      // cout << "\n Robot had ID \n";
//      alive[M_ID][M_ID] = true;
//      for(int i = 0; i < NUM_PLAYERS; i++) {
//        if(i != M_ID) {
//  	// Process any pong messages from these players
//  	while( communication->NumberofDataObjects(PONG_MESSAGE) > 0 ) {
//  	  recvData* data = communication->ReceiveData(PONG_MESSAGE);
//  	  if( data == NULL ) {
//  	    cout << "Error: received null data for pong message" << endl;
//  	  } else { 
//  	    PLAYER_ID_t fromID = (PLAYER_ID_t)data->DogFrom;
//  	    lastHeard[fromID] = 0;
//  	    if(DEBUG) {
//  	      cout << "Receiving pong from " << fromID << endl;
//  	    }
//  	  }
//  	  delete data;
//  	}
//  	if(!isAlive(i) && lastHeard[i] < CONSIDER_DEAD_THRESHOLD) {  
//  	  // Back up?
//  	  alive[M_ID][i] = true;
//  	  if(DEBUG) {
//  	    cout << i << " appears to be back from the dead." << endl;
//  	  }
//  	}
//  	if(lastHeard[i] >= START_PINGING_THRESHOLD) {
//  	  // Start pinging
//  	  if((lastHeard[i] - START_PINGING_THRESHOLD) % CYCLES_BETWEEN_PINGS == 0) {
//  	    if(DEBUG) {
//  	      cout << "Pinging " << i << endl;
//  	    }
//  	    // Send a ping
//  	    int msg = 0;
//  	    communication->SendData( PING_MESSAGE,
//  				     reinterpret_cast<byte *>(&msg),
//  				     sizeof(msg),
//  				     (sndDOGNUMBER)i);
//  	  }
//  	}
//  	if(lastHeard[i] >= CONSIDER_DEAD_THRESHOLD && isAlive(i)) {
//  	  alive[M_ID][i] = false;
//  	  if(DEBUG) {
//  	    cout << "Haven't heard from " << i << " in a long time.  Considering it dead." << endl;
//  	  }
//  	}
//  	// Now we age everyone's (except ours) "last heard" count
//  	lastHeard[i]++;
//        }
//      }
//    }
}



/**
 * Mohan::Function that updates the ball information in the
 * worldstate. This is called in Update() above...
 */
void WorldState::UpdateBallInfo() {
  if( ballInVision ) {
    BallKnown = true;
    ConsecutiveNotBall = 0;
    ++ConsecutiveBall;
    SureBall = ( ConsecutiveBall >= NUM_FRAMES_SEEN ) ? true : false;
  }
  else {
    ConsecutiveBall = 0;
    SureBall = false;
    ConsecutiveNotBall++;
    if( ConsecutiveNotBall > BALL_KNOWN_CONST ) {
      BallKnown = false;
    }
  }
  if (ballInVision || haveBall) {
    newCNB = 0;
  } else {
    newCNB++;
  }
}


/**
 * check all of the senses to see if they are being activated
 * moved from JointsAndSensors 16feb04 by peggy.
 */
void WorldState::updateSenses() {
  int i;

  // update lapseCycles and lastDuration for each input
  // (only use the electrostatic sensors and the foot sensors)
  for( i = HEAD_SENSOR; i < WIRELESS_LAN; i++ ) {
    // if the sensor is active, set lapseCycles to 0 and increase lastDuration
    if( sensors[i] > senseThresholds[i] ) {
      // if lapseCycles isn't zero, then we've been inactive for some
      // time: reset lastDuration in this case.  otherwise, increase lastDuration
      if( lapseCycles[i] > 0 ) {
	prevDuration[i] = lastDuration[i];
	prevLapseCycles[i] = lapseCycles[i]; // Update the previous ones
	lastDuration[i] = 1;
      } else {
	lastDuration[i]++;
      }
      lapseCycles[i] = 0;
    }
    // if the sensor isn't active, increase lapseCycles
    else {
      lapseCycles[i]++;
    }
  }
}


/**
 * Return true if the sense specified by which_sense has been
 * detected.
 * moved from JointsAndSensors 16feb04 by peggy.
 */
//  bool WorldState::getCustomSense( CUSTOM_SENSE_t which_sense ) {
//    if( which_sense >= 0 && which_sense < NUM_CUSTOM_SENSES ) {
//      switch( which_sense ) {
//        /////////////////////////////////////////////////////////////
//      case CS_BF_TAP:
//        // Make sure this isn't a double tap
//        return( lapseCycles[BACK_SENSOR_FRONT] == DOUBLE_TAP_PAUSE_LIMIT &&
//  	      lastDuration[BACK_SENSOR_FRONT] >= 1 &&
//  	      lastDuration[BACK_SENSOR_FRONT] <= CYCLES_PER_SECOND &&
//  	      prevLapseCycles[BACK_SENSOR_FRONT] >= DOUBLE_TAP_PAUSE_LIMIT );
//        /////////////////////////////////////////////////////////////
//      case CS_BM_TAP:
//          // Make sure this isn't a double tap
//        return( lapseCycles[BACK_SENSOR_MID] == DOUBLE_TAP_PAUSE_LIMIT &&
//  	      lastDuration[BACK_SENSOR_MID] >= 1 &&
//  	      lastDuration[BACK_SENSOR_MID] <= CYCLES_PER_SECOND &&
//  	      prevLapseCycles[BACK_SENSOR_MID] >= DOUBLE_TAP_PAUSE_LIMIT );
//        /////////////////////////////////////////////////////////////
//      case CS_BR_TAP:
//        // Make sure this isn't a double tap
//        return( lapseCycles[BACK_SENSOR_REAR] == DOUBLE_TAP_PAUSE_LIMIT &&
//  	      lastDuration[BACK_SENSOR_REAR] >= 1 &&
//  	      lastDuration[BACK_SENSOR_REAR] <= CYCLES_PER_SECOND &&
//  	      prevLapseCycles[BACK_SENSOR_REAR] >= DOUBLE_TAP_PAUSE_LIMIT );
//        /////////////////////////////////////////////////////////////
//      case CS_BF_PRESS:
//        return( lapseCycles[BACK_SENSOR_FRONT] == 0 &&
//  	      lastDuration[BACK_SENSOR_FRONT] == CYCLES_PER_SECOND );
//        /////////////////////////////////////////////////////////////
//      case CS_BM_PRESS:
//        return( lapseCycles[BACK_SENSOR_MID] == 0 &&
//  	      lastDuration[BACK_SENSOR_MID] == CYCLES_PER_SECOND );
//        /////////////////////////////////////////////////////////////
//      case CS_BR_PRESS:
//        return( lapseCycles[BACK_SENSOR_REAR] == 0 &&
//  	      lastDuration[BACK_SENSOR_REAR] == CYCLES_PER_SECOND );
//        /////////////////////////////////////////////////////////////
//      case CS_HEAD_TAP:
//        // Make sure this isn't a double tap
//        return( lapseCycles[HEAD_SENSOR] == DOUBLE_TAP_PAUSE_LIMIT &&
//  	      lastDuration[HEAD_SENSOR] >= 1 &&
//  	      lastDuration[HEAD_SENSOR] <= 3*CYCLES_PER_SECOND &&
//  	      prevLapseCycles[HEAD_SENSOR] >= DOUBLE_TAP_PAUSE_LIMIT );
//        /////////////////////////////////////////////////////////////
//      case CS_HEAD_PRESS_LONG:
//        return( lapseCycles[HEAD_SENSOR] == 0 &&
//  	      lastDuration[HEAD_SENSOR] == 3*CYCLES_PER_SECOND );
//        /////////////////////////////////////////////////////////////
//      case CS_CHIN_TAP:
//        return( lapseCycles[CHIN_SENSOR] == 1 &&
//  	      lastDuration[CHIN_SENSOR] < CYCLES_PER_SECOND );
//        /////////////////////////////////////////////////////////////
//      case CS_HEAD_DOUBLE_TAP:
//        return( lapseCycles[HEAD_SENSOR] == 1 &&
//  	      lastDuration[HEAD_SENSOR] >= 1 &&
//  	      lastDuration[HEAD_SENSOR] <= 3*CYCLES_PER_SECOND &&
//  	      prevLapseCycles[HEAD_SENSOR] < DOUBLE_TAP_PAUSE_LIMIT &&
//  	      prevDuration[HEAD_SENSOR] >= 1 &&
//  	      prevDuration[HEAD_SENSOR] <= 3*CYCLES_PER_SECOND );
//        /////////////////////////////////////////////////////////////
//      default:
//        fprintf(stderr, "JointsAndSensors::getCustomSense(): undefined custom sense %d\n", which_sense);
//        return( false );
//      }
    
//    } else {
//      fprintf(stderr, "JointsAndSensors::getCustomSense(): invalid sense request %d\n", which_sense);
//      return( false );
//    }
//  }



 


/**
 * Nate: returns true if we are at a point p on the field within
 * some threshold
 */
bool WorldState::IsAtPoint( Point2D p, double distanceThreshold ) {
//   Point2D mypos = SmoothPose.Position;
//   AngDeg mytheta = SmoothPose.Orientation;
//   Point2D temp = p;
//   Point2D posdisp = temp.globalToRelative( mypos, mytheta );
//   double dist = posdisp.getMagnitude();
  
  // Mohan:;Is it not easier to just do a single function call as
  // below, since both points are in global frame of reference
  // anyway...
  double dist = SmoothPose.Position.getDistanceTo( p );
  if( dist < distanceThreshold ) {
    return true;
  } 
  else {
    return false;
  }
}


/**
 * Function returns true iff the given pose is on the field...
 */
bool WorldState::IsOnField( Pose pose ) {
  return FIELD_BOUNDARY.isInside( pose.Position );
}


/**
 * Mohan::Function that returns true iff the given Point2D is on
 * field...
 */
bool WorldState::IsOnField( Point2D p ) {
  return FIELD_BOUNDARY.isInside( p );
}


/**
 * Mohan:Function that returns true iff the given Gauss structure
 * represents a point that is on field...
 */
bool WorldState::IsOnField( Gauss g ) {
  return ( FIELD_BOUNDARY.isInside( g.mean ) &&
	   g.stdev.getX() < MAX_STDX && 
	   g.stdev.getY() < MAX_STDY );
}


/**
 * Copied (and modified to be heading (y-axis) instead
 * of angle (x-axis)) from Particles.cc.  -peggy 24feb04
 */
// gjk - should we be using smoothpose here?
AngDeg WorldState::getHeadingToPoint( Point2D p )
{
  Vector2D relativeDisplacement = 
    p.globalToRelative( SmoothPose.Position, SmoothPose.Orientation );
  return normalizeAngle( relativeDisplacement.getDirection() - 90 );
}



/**
 * Just what it says....   -peggy 24feb04
 */
AngDeg WorldState::HeadingToFieldCenter() 
{
  return getHeadingToPoint( FIELD_BOUNDARY.getPosCenter() );
}



/**
 * Mohan::Function to return ball observation distance (ONLY WORKS IF
 * THE BALL IS IN THE VISUAL FRAME - NOT TO BE USED FOR THE BALL
 * ESTIMATE CASE)...
 */
double WorldState::getBallDistance() {
  if( !ballInVision ) {
    return NO_VISION_DIST;
  }

  for( int i = 0; i < numObservations; i ++ ) {
    if( observations[i].getObject() == OBJ_BALL ) {
      return observations[i].getDistance();
    }
  }
  return NO_VISION_DIST;
}



/**
 * Mohan::Function to return ball observation angle (in degrees wrt
 * x-axis)...
 */
AngDeg WorldState::getBallAngle() {
  if( !ballInVision ) {
    return -1000.0;
  }
  
  for( int i = 0; i < numObservations; i ++ ) {
    if( observations[i].getObject() == OBJ_BALL ) {
      return observations[i].getHeading();
    }
  }
  return -1000.0;
}



/**
 * Mohan::Function returns the probability of seeing ball in the
 * current frame of analysis - returns 0.0 if ball not seen...
 */
double WorldState::getBallConfidence() {
  if( !ballInVision ) {
    return 0.0;
  }
  for( int i = 0; i < numObservations; i++ ) {
    if( observations[i].getObject() == OBJ_BALL ) {
      return observations[i].getCertainty();
    }
  }
  return 0.0;
}


/**
 * Nate: the center of the opponent goal
 */
Point2D WorldState::getOppGoalPoint() {
  double x = ( MAXX + MINX ) / 2;
  double y = MAXY;
  return Point2D(x,y);
}


/**
 * Mohan::Function returns the color of the opponent's goal...
 */
int WorldState::getOppGoalColor() {
  return( ( side == SIDE_BLUE ) ? YELLOW : BLUE );
}


/**
 * Nate: depending on where we are, return some point inside the
 * opponent goal.  When we're within the goalbox min/max
 * x-coordinates, linearly vary with respect to the current position.
 * When we're outside, return the opposite corner.
 */
Point2D WorldState::getBestOppGoalPoint() {

  ///////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////
  // HISTORICAL MARKER
  //
  // At this point, on the ninth day in the fifth month of the year
  // 2005, did Peter create this block of code.  May it forever remain
  // unsullied by the hands of grad students.

  // Sorry, but Nate and Peter messed up the math here and I am now
  // (June 3, 2005) changing the original piece of code;-) -- Mohan...

  double myX = SmoothPose.Position.getX();
  double x;
  // scale factor is 1, we dribble across the goal from the right edge
  // to the left edge.  as scale factor increases, we dribble more and
  // more towards the center of the goal.
  double scaleFactor = 1.75;
  if( myX < GOAL_BOX_X_MIN*scaleFactor ) {
    x = GOAL_RIGHT_EDGE_X;
  } 
  else if( myX > GOAL_BOX_X_MAX*scaleFactor ) {
    x = GOAL_LEFT_EDGE_X;
  } 
  else {
    // Mohan::OK, at this point, in all the rush at Atlanta, Nate and
    // Peter forgot to add in a proper slope - I am forced to correct
    // it in the interests of the team:-)

    // Previously: x = -myX/scaleFactor; Anything > scaleFactor*400,
    // (GOAL_WIDTH/2=400), causes the robot to overshoot in the other
    // direction...
    double slope_x = ( GOAL_WIDTH/2 ) / ( scaleFactor*GOAL_BOX_WIDTH/2 );
    x = -myX * slope_x;
  }

  //
  // 
  ///////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////
  double y = MAXY;
  Point2D Target( x, y );

  // Flag to be removed later once the purpose is served...
  ChangedTargetPoint = false;

  // Mohan::Now, an attempt at trying to make the robot avoid
  // opponents, as it dribbles the ball...
  
  // Of course, if the robot is not sure of its pose, it should not
  // try anything crazy...
  if( SmoothPose.Confidence <= TEAMMATE_POSE_CONF_THRESH ) {
    return Target;
  }
  
  // First, to figure out if the current chosen target point is in a
  // dangerous direction...
  Point2D RelTarget = Target;
  RelTarget.globalToRelative( SmoothPose.Position, SmoothPose.Orientation );

  // If the relative direction of the normal target position is
  // "safe", just use that point as the target point - If not, choose
  // a new target point, making sure that the goal is not overshot
  // though...
//    if( !SafeDirection( RelTarget.getDirection() ) ) {
//      //     cout << "\n Opponent caused a change in target from: (" 
//      // 	 << Target.getX() << "," << Target.getY() << ") to: (";
    
//      // For now, a very simple strategy - find out the goal edge that
//      // the current target point is closest to...
//      Point2D left( GOAL_LEFT_EDGE_X, MAXY ), right( GOAL_RIGHT_EDGE_X, MAXY );
//      double dist1 = Target.getDistanceTo( left );
//      double dist2 = Target.getDistanceTo( right );
//      double offset = 100;

//      // Then, move the target point a bit in the opposite direction...
//      double new_x = ( ( dist1 > dist2 ) ? Target.getX()-offset : 
//  		     Target.getX()+offset );
    
//      // Bound it to be within the goal dimensions though...
//      new_x = minmax( GOAL_LEFT_EDGE_X, new_x, GOAL_RIGHT_EDGE_X );
//      Target.setX( new_x );

//      // cout << Target.getX() << "," << Target.getY() << ")" << endl;

//      // Flag to be removed later once the purpose is served...
//      ChangedTargetPoint = true;
//      // cout << "\n Opponent based target change...\n";
//    }

  // Before this target point is returned, one more crazy check - to
  // be performed iff the robot is close to the opponent goal...
  if( DistanceToOppGoal() <= CLOSE_TO_OPP_GOAL_DIST && 
      ( opponentGoalInVision || OpponentGoalBlobInVision() ) ) {
    bool TargChanged = false;
    Point2D NewTarget = Target;
    AngDeg GlobTargHead;

    Line2D OppGoalLine, TargLine;

    // Determine the line representing the opponent's goal line...
    OppGoalLine = Line2D::makeLineFromTwoPoints( Point2D(MINX, MAXY), 
						 Point2D(MAXX, MAXY) );
    if( opponentGoalInVision ) {
      // Determine angle to seen opp goal in global coordinates - we
      // use +90 because the "heading" function returns angles wrt the
      // robot's y-axis...
      GlobTargHead = normalizeAngle( HeadingToSeenOppGoal() + 90 
				     + SmoothPose.Orientation );
    }
    else { // Need not check explicitly - if(OpponentGoalBlobInVision())
      // Determine angle to seen (possible) opp goal blob in global
      // coordinates - we use +90 because the "heading" function
      // returns angles wrt the robot's y-axis...
      GlobTargHead = normalizeAngle( HeadingToSeenOppGoalBlob() + 90 
				     + SmoothPose.Orientation );
    }
    // Determine the line through the robot's current global position
    // at the angle of the seen/probable opp goal blob...
    TargLine = Line2D::makeLineFromPositionAndAngle( SmoothPose.Position,
						     GlobTargHead );
    
    // The intersection of the desired target line with the
    // opponent's goal line provides the desired target point...
    NewTarget = TargLine.getMyIntersection( OppGoalLine );
    // If this new target point is valid, use it...
    if( NewTarget.getY() > 0 ) {
      TargChanged = true;
      ChangedTargetPoint = true;
      // cout << "\n Vision feedback based target change...\n";
    }
    Target = ( TargChanged ? NewTarget : Target );
  }

  return Target;
}



/**
 * Nate: The angle in field coordinates to the opponents goal from a
 * destination point given as the argument...
 */ 
AngDeg WorldState::GlobalAngleToOppGoal( Point2D dest ) 
{
  double x = ( MAXX + MINX ) / 2;
  double y = MAXY;
  Point2D oppGoal( x, y );
  
  double angle = normalizeAngle( atan2Deg( (oppGoal-dest).getY(),
					   (oppGoal-dest).getX() ) - 90 );
  return angle;
}



/**
 * Mohan:: I have just brought in this function from GlobalMap.cc and
 * changed it to the current format in WorldState...
 */
AngDeg WorldState::HeadingToOppGoal() {
  double xBase = ( MAXX + MINX ) / 2;
  double xDelta = GOAL_INTERIOR_X / 2;
  double y = MAXY;
  
  //theta1 and theta2 are by default in degrees...
  //Dan: angleToPoint returns angle from x-axis; HeadingtoOppGoal from y.
  Point2D p1(xBase + xDelta,y);
  double theta1 = normalizeAngle(angleToPoint( p1 ) - 90);

  Point2D p2(xBase - xDelta,y);
  double theta2 = normalizeAngle(angleToPoint( p2 ) - 90);

  double target = ( Vector2D( 1.0, theta1, POLAR ) + 
		    Vector2D( 1.0, theta2, POLAR ) ).getDirection();
  return normalizeAngle( target );
}



/**
 * Mohan::Returns true if a candidate opponent goal blob has been seen
 * in vision, in the current frame...
 */
bool WorldState::OpponentGoalBlobInVision() {
  if( side == SIDE_BLUE ) {
    return( YGBlobObsr.getObject() == OBJ_YELLOW_BLOB );
  }
  else {
    return( BGBlobObsr.getObject() == OBJ_BLUE_BLOB );    
  }
}



/**
 * Mohan::Returns the observation corresponding to the opponent goal
 * colored blob that could possibly be part of the opponent goal - USE
 * WITH CARE - ASSUMES THAT A GOAL COLORED BLOB HAS BEEN SEEN IN
 * CURRENT FRAME. ALSO, THIS BLOB COULD BE NOISE...
 */
Observation WorldState::OppGoalBlobObsr() {
  if( !OpponentGoalBlobInVision() ) {
    cout << "\n DUMBO - crashing - check for opp goal blob first!!\n";
  }
  return( ( side == SIDE_BLUE ) ? YGBlobObsr : BGBlobObsr );
}



/**
 * Mohan::Provides a heading (wrt robot y-axis) to the candidate blob
 * that could possibly be the opponent goal - USE WITH CARE - USE IFF
 * ROBOT IS VERY CLOSE TO THE GOAL, ASSUMES THAT CANDIDATE BLOB IS
 * SEEN...
 */
AngDeg WorldState::HeadingToSeenOppGoalBlob() {
  if( OpponentGoalBlobInVision() ) {
    return( normalizeAngle( OppGoalBlobObsr().getHeading() - 90 ) );
  }
  cout << "\n DUMBO - referencing NULL \n";
  return 0.0;
}



/**
 * uses vision first, then localization if the goal isn't currently
 * seen.  --peggy 8mar04
 */
AngDeg WorldState::HeadingToSeenOppGoal() {
  ObjectT soughtObject;
  
  if( SIDE_BLUE == side ) { // opp goal is yellow
    soughtObject = OBJ_YELLOW_GOAL;
  }
  else { // opp goal is blue
    soughtObject = OBJ_BLUE_GOAL;
  }
  for( int i=0; i < numObservations; i++ ) {
    if( observations[i].getObject() == soughtObject ) {
      //fprintf(stderr, "OPP GOAL SEEN: heading %f\n", observations[i].getHeading() - 90);
      return( normalizeAngle( observations[i].getHeading() - 90 ) );
    }
  } 
  // if we get here without returning, then we can't see the goal
  //fprintf(stderr, "OPP GOAL ---NOT--- SEEN: heading %f\n", HeadingToOppGoal());
  return HeadingToOppGoal();
}



/**
 * Mohan::Gives the heading (wrt robot's y-axis) to the point to shoot
 * to when the robot tries to shoot to target but cannot find the
 * target...
 */
AngDeg WorldState::HeadingWhenTargetUnseen() {
  Point2D target;
  target = ( OnRightSideOfField() ?
	     Point2D( MINX, MAXY/4 ) :
	     Point2D( MAXX, MAXY/4 ) );

  return( normalizeAngle( angleToPoint( target ) - 90 ) );
}



/**
 * Mohan:: returns the heading (wrt y-axis) to own goal...
 *
 * Nate (via Dan): new implementation, will behave strangely if called
 * from inside the goal box
 */
AngDeg WorldState::HeadingToOwnGoal() {
  double xBase = ( MAXX + MINX ) / 2;
  double xDelta = GOAL_INTERIOR_X / 2;
  double y = MINY;
  
  // theta1 and theta2 are by default in degrees. The -90 is to get
  // the angles wrt the y-axis of the robot (pointing straight
  // ahead)...
  Point2D p1(xBase + xDelta,y);
  double theta1 = normalizeAngle(angleToPoint( p1 ) - 90);

  Point2D p2(xBase - xDelta,y);
  double theta2 = normalizeAngle(angleToPoint( p2 ) - 90);
	
  double target = ( Vector2D( 1.0, theta1, POLAR ) + 
		    Vector2D( 1.0, theta2, POLAR ) ).getDirection();
  return normalizeAngle( target );

  /*
  if (side == SIDE_RED && theta1 > theta2)
    theta2 += 360;
  else if (side == SIDE_BLUE && theta2 > theta1)
    theta1 += 360;
  return normalizeAngle( (theta1 + theta2) / 2);
  return normalizeAngle (theta1 + normalizeAngle(theta2-theta1)/2);					 
  */
}



/**
 * Dan:: returns the distance to own goal...
 */
double WorldState::DistanceToOwnGoal() {
  double x = GOAL_CENTER_X;
  double y = MINY;
  
  Point2D p(x,y);

  return distanceToPoint( p );
}

/**
 * Dan:: returns the distance to opp goal...
 */
double WorldState::DistanceToOppGoal() {
  double x = GOAL_CENTER_X;
  double y = MAXY;
  
  Point2D p(x,y);

  return distanceToPoint( p );
}



/**
 * Mohan::Similar to the function for heading, the robot first tries
 * to obtain the distance to the Opponent's goal based on vision
 * objects and if that is not successful, it uses the standard
 * approach of taking the known global coordinates of the goal - USE
 * WITH CAUTION AND EVEN THEN IFF THE ROBOT IS CLOSE TO THE OPPONENT'S
 * GOAL...
 */
double WorldState::DistanceToSeenOppGoal() {
  ObjectT ReqObject;
  
  if( side == SIDE_BLUE ) {
    ReqObject = OBJ_YELLOW_GOAL;
  }
  else {
    ReqObject = OBJ_BLUE_GOAL;
  }
  for( int i=0; i < numObservations; i++ ) {
    if( observations[i].getObject() == ReqObject ) {
      return( observations[i].getDistance() );
    }
  } 

  return DistanceToOppGoal();
}



/**
 * Mohan::returns the heading (wrt y-axis) to the left corner of the
 * opponent's half...
 */
AngDeg WorldState::HeadingToOppLeftCorner() {
  double x = MINX;
  double y = MAXY;
  
  Point2D p(x, y);
  //The -90 is to get the angle wrt the y-axis of the robot (pointing
  //straight ahead)...
  return normalizeAngle( angleToPoint( p ) - 90 );
}


/**
 * Mohan::returns the heading (wrt y-axis) to the right corner of the
 * opponent's half...
 */
AngDeg WorldState::HeadingToOppRightCorner() {
  double x = MAXX;
  double y = MAXY;

  Point2D p(x, y);
  //The -90 is to get the angle wrt the y-axis of the robot (pointing
  //straight ahead)...
  return normalizeAngle( angleToPoint( p ) - 90 );
}

/**
 * peggy: returns the heading to the midpoint of the right half
 * of the offensive end of the field (just right of the offense
 * right goal post).
 */
AngDeg WorldState::HeadingToOppMidRightSide() {
  double x = (MAXX - FIELD_CENTER_X)/2.0 + FIELD_CENTER_X;
  // double x = FIELD_CENTER_X + GOAL_RIGHT_EDGE_X;
  double y = MAXY;

  Point2D p(x, y);
  //The -90 is to get the angle wrt the y-axis of the robot (pointing
  //straight ahead)...
  return normalizeAngle(angleToPoint(p) - 90);
}

/**
 * peggy: returns the heading to the midpoint of the left half
 * of the offensive end of the field (just left of the offense
 * left goal post).
 */
AngDeg WorldState::HeadingToOppMidLeftSide() {
  double x = (FIELD_CENTER_X - MINX)/2.0 + MINX;
  // double x = FIELD_CENTER_X + GOAL_LEFT_EDGE_X;
  double y = MAXY;

  Point2D p(x,y);
  //The -90 is to get the angle wrt the y-axis of the robot (pointing
  //straight ahead)...
  return normalizeAngle(angleToPoint(p) - 90); 
}


/**
 * Mohan::returns the heading (wrt y-axis) to the left corner of the
 * opponent's goal...
 */ 
AngDeg WorldState::HeadingToOppGoalLeftCorner() {
  double x = GOAL_LEFT_EDGE_X;
  double y = MAXY;
  
  Point2D p(x, y);
  // The -90 is to get the angle wrt the y-axis of the robot (pointing
  // straight ahead)...
  return normalizeAngle( angleToPoint( p ) - 90 );
}


/**
 * Mohan::returns the heading (wrt y-axis) to the right corner of the
 * opponent's goal...
 */
AngDeg WorldState::HeadingToOppGoalRightCorner() {
  double x = GOAL_RIGHT_EDGE_X;
  double y = MAXY;
  
  Point2D p(x, y);
  // The -90 is to get the angle wrt the y-axis of the robot (pointing
  // straight ahead)...
  return normalizeAngle( angleToPoint( p ) - 90 );
}


/*
 * Dan: returns if point is in own goal box.
 */
bool WorldState::InOwnGoalBox(Point2D p) {
  bool WithinX = (p.getX() <= GOAL_BOX_X_MAX) && (p.getX() > GOAL_BOX_X_MIN);
  bool WithinY;
  WithinY = (p.getY() <= OWN_GOAL_BOX_Y_MAX && p.getY() > OWN_GOAL_BOX_Y_MIN );

  return (WithinX && WithinY);
}



/**
 * Nate: are we in a place on the field where the angle to the goal
 * posts is really oblique?
 * peggy 7mar05: updated for new field
 */
bool WorldState::InOffensiveWasteland() {
  double m, b;
  double y;
  // return true if we're above either of two lines describing the
  // offensive wasteland.

  // the 3/4 fraction comes from the fact that we want each line to
  // hit the side of the field 1/2 of the way from the goal to the
  // beacon.  -peggy 7mar05
  // looks like mohan changed this to 1/2, to make the lines hit the
  // side of the field *at* the beacon now.  -peggy 21apr05

  // m = (MAXY*3.0/4.0 - MAXY) / (MAXX - GOAL_RIGHT_EDGE_X);
  // m = (MAXY*1.0/2.0 - MAXY) / (MAXX - GOAL_RIGHT_EDGE_X);
  // b = MAXY - m * GOAL_RIGHT_EDGE_X;

  // Mohan::Jun28,05 - Changing the region to extend from the beacon
  // to the corresponding penalty area corner...
  
  // First, on the right side of the field, the slope is given by...
  m = ( ( MAXY*1.0/2.0 - ( MAXY - GOAL_BOX_DEPTH ) ) / 
	( MAXX - GOAL_BOX_X_MAX ) );
  // The intercept then becomes...
  b = ( MAXY/2.0 ) - ( m * MAXX );
  
  // Sample point test case...
  y = m * SmoothPose.Position.getX() + b;

  // Return true iff the robot is above the line and to the right of
  // the goal...
  if( SmoothPose.Position.getY() >= y &&
      SmoothPose.Position.getX() >= GOAL_RIGHT_EDGE_X ) {
    //cout << "In RIGHT offensive wasteland" << endl;
    return true;
  }
  
  // Then, on the left side of the field, the slope is given by...
  // m = (MAXY - MAXY*3.0/4.0) / (GOAL_LEFT_EDGE_X - MINX);
  // m = (MAXY - MAXY*1.0/2.0) / (GOAL_LEFT_EDGE_X - MINX);
  // b = MAXY - m * GOAL_LEFT_EDGE_X;
  m = ( ( MAXY*1.0/2.0 - ( MAXY - GOAL_BOX_DEPTH ) ) / 
	( MINX - GOAL_BOX_X_MIN ) );
  // The intercept then becomes...
  b = ( MAXY/2.0 ) - ( m * MINX );
  
  // Sample point test case...
  y = m * SmoothPose.Position.getX() + b;

  // Return true iff the robot is above the line and to the left of
  // the goal...
  if( SmoothPose.Position.getY() >= y && 
      SmoothPose.Position.getX() <= GOAL_LEFT_EDGE_X ) {
    //cout << "In LEFT offensive wasteland" << endl;
    return true;
  }
  
  // if we're not above either line, return false
  //cout << "Not" << endl;
  return false;
}



/**
 * Nate: heading to the point right in front of the opponent's goal
 */
AngDeg WorldState::HeadingToOffensiveKey() {
  if( side == SIDE_BLUE ) {
    return getHeadingToPoint( Point2D( 0,  1940 ) );
  } else {
    return getHeadingToPoint( Point2D( 0, -1940 ) );
  }
}


/**
 * Mohan::returns the pose of the teammate whose ID is specified as
 * input...
 */
Pose WorldState::GetTeammatePose( int tmid ) {
  for( int i = 0; i < (NUM_TEAM_MEMBERS-1); i++ ) {
    if( teammateState[i].ID == tmid ) {
      return teammateState[i].pose;
    }
  }
  cout << "\n No such teammate exists...\n";
  return DummyPose;
}


/**
 * Mohan::returns the pose of the teammate who currently is the
 * attacker...
 */
Pose WorldState::GetPoseOfAttacker() {
  for( int i = 0; i < (NUM_TEAM_MEMBERS-1); i++ ) {
    if( teammateState[i].Role == ATTACKER_ROLE ) {
      return teammateState[i].pose;
    }
  }
  cout << "\n No such teammate exists...\n";
  return DummyPose;
}

/**
 * Kurt: gets the home position
 */

Pose WorldState::GetStartingPose() {
  return startingPose;
}


/**
 * Mohan::Function returns the position that the robot has to go to if
 * it decides to spread out to find the ball...
 */
Pose WorldState::GetSpreadPose() {
  return spreadPose;
}


/**
 * Mohan::returns the pose of the teammate who currently is the
 * supporter...
 */
Pose WorldState::GetPoseOfSupporter() {
  for( int i = 0; i < (NUM_TEAM_MEMBERS-1); i++ ) {
    if( teammateState[i].Role == SUPPORTER_ROLE ) {
      return teammateState[i].pose;
    }
  }
  cout << "\n No such teammate exists...\n";
  return DummyPose;
}


/**
 * Mohan::returns the pose of the teammate who currently is the
 * defender...
 */
Pose WorldState::GetPoseOfDefender() {
  for( int i = 0; i < (NUM_TEAM_MEMBERS-1); i++ ) {
    if( teammateState[i].Role == DEFENDER_ROLE ) {
      return teammateState[i].pose;
    }
  }
  cout << "\n No such teammate exists...\n";
  return DummyPose;
}


/**
 * Mohan::returns the pose of the teammate who currently is the
 * keeper...
 */
Pose WorldState::GetPoseOfKeeper() {
  for( int i = 0; i < (NUM_TEAM_MEMBERS-1); i++ ) {
    if( teammateState[i].Role == KEEPER_ROLE ) {
      return teammateState[i].pose;
    }
  }
  cout << "\n No such teammate exists...\n";
  return DummyPose;
}


/**
 * Mohan::returns the pose of the teammate with a specific role...
 */
Pose WorldState::GetPoseOfRobotWithRole( role_t role ) {
  for( int i = 0; i < (NUM_TEAM_MEMBERS-1); i++ ) {
    if( teammateState[i].Role == role ) {
      return teammateState[i].pose;
    }
  }
  cout << "\n No such teammate exists...\n";
  return DummyPose;
}



/**
 * Mohan::eturns the smoothly varying pose...
 */
Pose WorldState::GetSmoothPose() {
  return SmoothPose;
}



/**
 * Returns the distance moved since the specified number of frames ago,
 * according to SmoothPoseLocationHistory.  If framesAgo is larger than 
 * POSE_HISTORY_SIZE, it will return the distance moved since 
 * POSE_HISTORY_SIZE frames ago.  The distance moved is averaged over the
 * specified number of frames ("window") after framesAgo.  -peggy
 */ 
Distance WorldState::averageDistanceMovedSince( int framesAgo, int window ) {
  framesAgo = min(framesAgo, POSE_HISTORY_SIZE);
  if( framesAgo - window < 0 ) {  // keep the for loop below from accessing negative indices
    window = framesAgo;
  }
  Point2D avgPosn;
  avgPosn.setX(0);
  avgPosn.setY(0);

  for( int i=0; i < window; i++ ) {
    avgPosn = avgPosn + SmoothPoseLocationHistory[framesAgo - 1 - i];
  }
  avgPosn = avgPosn / (double)window;

  return( SmoothPose.Position - avgPosn ).getMagnitude();
}



/**
 * Mohan::Fucntion that returns true if the robot is within a certain
 * buffer zone around the goal box...
 */
bool WorldState::NearOwnGoalBox() {
  if( uncappedSmoothPose.Confidence <= 0.65 ) {
    return false;
  }

  double x = uncappedSmoothPose.Position.getX();
  double y = uncappedSmoothPose.Position.getY();
  //cout << "\n Robotx, Roboty (" << x << "," << y << ")" << endl;

  double buffer = 200;
  double buffer2 = 150;

  return( x >= ( GOAL_BOX_X_MIN - buffer2 ) &&
	  x <= ( GOAL_BOX_X_MAX + buffer2 ) &&
	  y <= ( OWN_GOAL_BOX_Y_MAX + buffer ) &&
	  y >= ( OWN_GOAL_BOX_Y_MIN )
	  );
}



/**
 * Selim: Function that returns true if the ball is right along the
 * opponent goal line
 */
bool WorldState::BallAtOpponentGoalLine() {
  if( uncappedSmoothPose.Confidence <= 0.65 || 
      !ballInVision ) {
    return false;
  }

  double x = globalBallEstimate.getX();
  double y = globalBallEstimate.getY();

  // The buffer determines how far from the line we allow the ball to
  // be while still considering it "along the line"...
  double buffer = 250;

  return( x >= ( GOAL_LEFT_EDGE_X - buffer/2 ) &&
	  x <= ( GOAL_RIGHT_EDGE_X + buffer/2 ) &&
	  y >= ( OPP_GOAL_BOX_Y_MAX - buffer )
	  );
}


/**
 * Mohan::Fucntion that returns true if the ball is within a certain
 * buffer zone around the goal box - a modification of the function
 * below, for the keeper...
 */
bool WorldState::BallInKeeperZone() {
  double x = globalBallEstimate.getX();
  double y = globalBallEstimate.getY();
  double buffer = 200;

  return( x >= ( GOAL_BOX_X_MIN - buffer ) &&
	  x <= ( GOAL_BOX_X_MAX + buffer ) &&
	  y <= ( OWN_GOAL_BOX_Y_MAX + buffer )
	  );
}



/**
 * Mohan::Fucntion that returns true if the ball is within a certain
 * buffer zone around the goal box...
 */
bool WorldState::BallNearOwnGoalBox() {
  // Ideally we should not check if the localization estimate is bad
  // but for now...
  if( uncappedSmoothPose.Confidence <= 0.65 || 
      BallConfidence < BALL_WELL_KNOWN_CONF ) {
    return false;
  }
  
  double x = globalBallEstimate.getX();
  double y = globalBallEstimate.getY();
  double buffer = 100;

  return( x >= ( GOAL_BOX_X_MIN - buffer ) &&
	  x <= ( GOAL_BOX_X_MAX + buffer ) &&
	  y <= ( OWN_GOAL_BOX_Y_MAX + buffer )
	  );
}


/**
 * Nate: like above, but for the opponent's goal box
 */
bool WorldState::BallNearOppGoalBox( double buffer ) {
  // Ideally we should not check if the localization estimate is bad
  // but for now - NOT ANYMORE - MOHAN...
//   if( uncappedSmoothPose.Confidence <= 0.65 || 
//       BallConfidence < BALL_WELL_KNOWN_CONF ) {
//     return false;
//   }
  
  double x = globalBallEstimate.getX();
  double y = globalBallEstimate.getY();

  return( x >= ( GOAL_BOX_X_MIN - buffer ) &&
	  x <= ( GOAL_BOX_X_MAX + buffer ) &&
	  y >= ( OPP_GOAL_BOX_Y_MIN - buffer )
	  );
}



/**
 * Mohan::Function returns true if the ball is in the robot's
 * half...
 */
bool WorldState::BallInOwnHalf() {
  if( SmoothPose.Confidence <= 0.65 ||
      BallConfidence < BALL_WELL_KNOWN_CONF ) {
    return false;
  }

  double buffer = 100;
  return( globalBallEstimate.getY() <= buffer );
}




/**
 * Mohan::Function returns true if the ball is in the opponent's
 * half...
 */
bool WorldState::BallInOpponentHalf() {
  if( SmoothPose.Confidence <= TEAMMATE_POSE_CONF_THRESH || 
      BallConfidence < BALL_WELL_KNOWN_CONF ) {
     return false;
  }
  
  double buffer = 100;
  return( globalBallEstimate.getY() >= buffer );
}




/**
 * Given what team we're on, return true if we're on the right side of
 * the field...
 */
bool WorldState::OnRightSideOfField() {
  double x = SmoothPose.Position.getX();
  return( x > 0 );
}



/**
 * Which side of the field we're on, in field coordinates: where
 * right is x > 1450
 */
bool WorldState::OnAbsoluteRightSideOfField() {
  if ( side == SIDE_BLUE )
    return OnRightSideOfField();
  if ( side == SIDE_RED )
    return !OnRightSideOfField();
  cerr << "Something wrong with side: " << side << endl;
  return false;
}

/**
 * Facing top of field in team coordinates?
 */
bool WorldState::FacingTopOfField() {
  AngDeg t = SmoothPose.Orientation;
  return t > -90 && t <= 90;
}

/**
 * True if we're facing the top of the field, in field coordinates:
 * top of field => our theta in [-90,90]
 */
bool WorldState::FacingAbsoluteTopOfField() {

  if ( side == SIDE_BLUE )
    return FacingTopOfField();
  if ( side == SIDE_RED )
    return !FacingTopOfField();

  cerr << "Something wrong with side: " << side << endl;
  return false;
}


/**
 * Functions to test confidence in our position, ball position, and
 * both.
 */
bool WorldState::BallConfident() {
  return BallConfidence >= BALL_WELL_KNOWN_CONF;
}
bool WorldState::BodyConfident() {
  return SmoothPose.Confidence >= TEAMMATE_POSE_CONF_THRESH;
}
bool WorldState::BallAndBodyConfident() {
  return BallConfident() && BodyConfident();
}


/**
 * Mohan::Function returns true if the ball is in the top third of
 * the field - depends on the side the robot is defending...
 */
bool WorldState::BallInTopThird() {
//   if( BallConfidence < BALL_WELL_KNOWN_CONF || 
//       SmoothPose.Confidence < TEAMMATE_POSE_CONF_THRESH ) {
//     return false;
//   }

  double y = globalBallEstimate.getY();
  return( y >=  ( MINY + ( 2.0/3.0 * FIELD_LENGTH ) ) );
}


/**
 * Nate: like ballintopthird, but bottom instead!
 */
bool WorldState::BallInBottomThird() {
  if( BallConfidence < BALL_WELL_KNOWN_CONF || 
      SmoothPose.Confidence < TEAMMATE_POSE_CONF_THRESH ) {
    return false;
  }

  double y = globalBallEstimate.getY();
  return( y <= ( MINY + ( 1.0/3.0 * FIELD_LENGTH ) ) );
}


  /**
   * Is the ball in the middle third of the field?
   */
bool WorldState::BallInMiddleThird() {
  if( BallConfidence < BALL_WELL_KNOWN_CONF || 
      SmoothPose.Confidence < TEAMMATE_POSE_CONF_THRESH ) {
    return false;
  }
  
  double y = globalBallEstimate.getY();
  return( y >= ( MINY + ( 1.0/3.0 * FIELD_LENGTH ) ) &&
	  y <= ( MINY + ( 2.0/3.0 * FIELD_LENGTH ) ) );
}


/**
 * Mohan::Function returns true if the ball is in the correct zone for
 * a switch to happen between the attacker and defender roles...
 */
bool WorldState::BallInSwitchZone() {
  return ( globalBallEstimate.getY() >= MINY/2 &&
	   globalBallEstimate.getY() <= MAXY/4 );
}


/**
 * Mohan::Function returns the x-axis displacement between the robot
 * and the ball in global coordinates...
 */
double WorldState::BallXDisplacement() {
  return( globalBallEstimate.getX() - SmoothPose.Position.getX() );
}


/**
 * Mohan::Function returns the y-axis displacement between the robot
 * and the ball in global coordinates...
 */
double WorldState::BallYDisplacement() {
  return( globalBallEstimate.getY() - SmoothPose.Position.getY() );
}


/**
 * Mohan::Function is a "weak" clone of the function above that has a
 * low confidence limit for the defender to move up...
 */
bool WorldState::BallWellInOppZone() {
//   if( SmoothPose.Confidence < 0.65 ) {
//     return false;
//   }
  double y = globalBallEstimate.getY();
  return( y >= (2.0/3.0 * MAXY) );
}


/**
 * Not Mohan::returns true is ball is on green carpet...
 */
bool WorldState::BallOnGrass() {
  return GRASS_BOUNDARY.isInside( globalBallEstimate );
}



/**
 * Mohan::A version of the 'ball close to borders' function that just
 * returns true iff the ball is close to a field edge...
 */
bool WorldState::BallCloseToBorders() {
  Point2D edge = ClosestEdgePointToBall();
  Point2D ball = globalBallEstimate;

  // Sanity checks to ensure ball stays on the field...
  if( ball.getX() >= MAXX ) ball.setX(MAXX-1);
  if( ball.getX() <= MINX ) ball.setX(MINX+1);
  if( ball.getY() >= MAXY ) ball.setY(MAXY-1);
  if( ball.getY() <= MINY ) ball.setY(MINY+1);

  // Near the opponent's goal box the robot does not pause its
  // attack...
  if( BallCloseToOppGoalBox() ) {
    return false;
  }

  // If the ball is not close to the edge, do not worry about this
  // case...
  if( ball.getDistanceTo(edge) <= 
      BALL_CLOSE_TO_WALL_DISTANCE_THRESHOLD ) {
    return true;
  }
  
  return false;
}



/**
 * Nate: returns a global Point2D of the closest point on the edge
 * of the field to the ball
 */
Point2D WorldState::ClosestEdgePointToBall() {
  // create a point that represents the position of the ball on the field
  Point2D ball = globalBallEstimate;

  // determine the closest point on each edge line to the ball point
  Point2D pbot = lbot.getPointOnLineClosestTo( ball );
  Point2D ptop = ltop.getPointOnLineClosestTo( ball );
  Point2D plft = llft.getPointOnLineClosestTo( ball );
  Point2D prgt = lrgt.getPointOnLineClosestTo( ball );

  // determine the smallest of the above 4 points
  double smallestDistance = 9999999.0;
  Point2D closestPoint;
  double d;
  d = pbot.getDistanceTo(ball);
  if( d < smallestDistance ) {
    smallestDistance = d;
    closestPoint = pbot;
    LineNum = 0;
  }
  d = ptop.getDistanceTo(ball);
  if( d < smallestDistance ) {
    smallestDistance = d;
    closestPoint = ptop;
    LineNum = 1;
  }
  d = plft.getDistanceTo(ball);
  if( d < smallestDistance ) {
    smallestDistance = d;
    closestPoint = plft;
    LineNum = 2;
  }
  d = prgt.getDistanceTo(ball);
  if( d < smallestDistance ) {
    smallestDistance = d;
    closestPoint = prgt;
    LineNum = 3;
  }
  
  return closestPoint;
}



/**
 * Nate: Returns true if the robot is in a bad situation in which to
 * approach the ball: namely, the ball is close to a wall and the
 * robot is also close to that wall.  If we return true, we'll also
 * set turnDirection to be the lateral direction in which we should
 * move to better approach the ball.
 */ 
//  bool WorldState::BadBallOnWallSituation( GenericDir& turnDirection ) {
//    // Mohan::Was using this function temporarily - June6, 2005 - but
//    // now it has been replaced by modified version right below...
//    return false;

//    // get the closest edge point to the ball, the ball, and our position
//    Point2D edge = ClosestEdgePointToBall();
//    Point2D ball = globalBallEstimate;
//    Point2D player = SmoothPose.Position;

//    // since the ball is relative, it can go outside the field which
//    // messes with this calculation.  So we'll force it inside the field
//    if( ball.getX() >= MAXX ) ball.setX(MAXX-1);
//    if( ball.getX() <= MINX ) ball.setX(MINX+1);
//    if( ball.getY() >= MAXY ) ball.setY(MAXY-1);
//    if( ball.getY() <= MINY ) ball.setY(MINY+1);

//    // if the ball is along the defensive back wall of the field, or in
//    // either of the offensive corners, don't worry about it and return
//    // false
//    if( ball.getY() < 2.0*MINY/3 || // def zone
//        ball.getY() > 2.0*MAXY/3 ) { // off zone
//      return false;
//    }

//    // if the ball isn't close to the wall, return false and don't worry
//    // about setting turnDirection
//    if( ball.getDistanceTo(edge) >= BALL_CLOSE_TO_WALL_DISTANCE_THRESHOLD ) {
//      cout << "ball distance to edge is small, returning false: " << ball.getDistanceTo(edge) << endl;
//      return false;
//    }

//    // compute a vector from the player to the ball and from the ball to the edge
//    Vector2D playerBall = ball - player;
//    Vector2D ballEdge = edge - ball;

//    // compute the difference in angles between those two vectors: if
//    // the difference is small enough, we're already lined up well with
//    // the ball and we're not in a bad location
//    AngDeg alpha = normalizeAngle( playerBall.getDirection() - ballEdge.getDirection() );
//    if( fabs(alpha) < BALL_CLOSE_TO_WALL_ANGLE_THRESHOLD ) {
//      cout << "angle is small enough, returning false: " << alpha << endl;
//      return false;
//    }

//    // depending on the angle, set turnDirection
//    if( alpha > 0 ) {
//      turnDirection = DIR_LEFT;
//    } 
//    else {
//      turnDirection = DIR_RIGHT;
//    }
//  //   cout << "*** ball in bad situation, alpha: " << alpha
//  //        << ", turnDirection: " << turnDirection << endl;
//    return true;
//  }



/**
 * Mohan: A variation of the function above - returns true if the
 * robot is in a bad situation in which to approach the ball: namely,
 * the ball is close to the field border - the robot should not
 * approach it head on now...
 */ 
//  bool WorldState::BallCloseToFieldBorder( GenericDir& turnDirection ) {
//    Point2D edge = ClosestEdgePointToBall();
//    Point2D ball = globalBallEstimate;
//    Point2D robot = SmoothPose.Position;

//    // Sanity checks to ensure ball stays on the field...
//    if( ball.getX() >= MAXX ) ball.setX(MAXX-1);
//    if( ball.getX() <= MINX ) ball.setX(MINX+1);
//    if( ball.getY() >= MAXY ) ball.setY(MAXY-1);
//    if( ball.getY() <= MINY ) ball.setY(MINY+1);

//    // If the ball is not close to the edge, do not worry about this
//    // case...
//    if( ball.getDistanceTo(edge) >= 
//        BALL_CLOSE_TO_WALL_DISTANCE_THRESHOLD ) {
//  //     cout << "Ball far from borders: " << ball.getDistanceTo(edge) 
//  // 	 << " LiA\n";
//      return false;
//    }

//    // If the ball is well into the defensive or offensive zone, and in
//    // front of the goals, do not make any changes in path...
//    if( BallCloseToOppGoalBox() ) { 
//      // cout << "\n Ball well into offensive zone - LiA \n";
//      return false;
//    }

//    // Determine a vector from the robot to the ball...
//    Vector2D RobotBall = ball - robot;
//    double CurrApprAng = RobotBall.getDirection();
//    double DesApprAng = DesiredBallApproachAng();
  
//    if( DesApprAng == -1 ) {
//      cout << "\n Calculation error in closest line \n";
//      return false;
//    }

//    cout << "\n Ball close to border lines, angle: " 
//         << CurrApprAng << ", DesAng: " << DesApprAng << "\n";
  
//    // Some more calculations to determine direction of approach...
//    CompensAng = normalizeAngle( CurrApprAng - DesApprAng );

//    if( fabs( CompensAng ) <= 20 ) {
//      cout << "\n Angdiff too small: " << CompensAng << " LiA\n";
//      return false;
//    }
//    else {
//      if( CompensAng < 0 ) {
//        turnDirection = DIR_RIGHT;
//        cout << "\n Angdiff -ve : " << CompensAng << " - ACW\n";
//      }
//      else {
//        turnDirection = DIR_LEFT;
//        cout << "\n Angdiff +ve : " << CompensAng << " - CW\n";
//      }
//    }

//    return true;
//  }



/**
 * Mohan::Function returns true if the ball is well into the offensive
 * half and its x-ccordinates are withina buffer region of the goal
 * box...
 */
bool WorldState::BallCloseToOppGoalBox() {
  Point2D ball = globalBallEstimate;
  int buffer = 50;

  if( ball.getY() > 2.0*MAXY/3 &&
      fabs( ball.getX() ) <= GOAL_BOX_WIDTH/2 + buffer ) {
    return true;
  }
  return false;
}



/**
 * Mohan::Function returns the desired ball approach angle, given that
 * it is close to a field border - THIS IS THE ANGLE MADE BY THE
 * ROBOT'S Y-AXIS WRT THE GLOBAL X-AXIS...
 */
AngDeg WorldState::DesiredBallApproachAng() {
//   // If it is the left or right edge of the field...
//   if( LineNum >= 2 ) {
//     return( BallInLeftHalf() ? 60 : 120 );
//   }
//   // Bottom edge of field - team relative...
//   else if( LineNum == 0 ) {
//     return 90;
//   }
//   // Top edge of field - team relative...
//   else if( LineNum == 1 ) {
//     return -90;
//   }

  // If it is the left or right edge of the field...
  if( LineNum >= 2 ) {
    // As a conservative approach, robot attempts to get to the ball
    // along the line of the field edge - angle depends on the current
    // robot orientation...
    if( SmoothPose.Orientation > -90 &&
	SmoothPose.Orientation < 90 ) {
      return 90;
    }
    else {
      return -90;
    }
  }
  // Bottom edge of field - team relative...
  else if( LineNum == 0 ) {
    // Angle of approach depends on the ball position...
    return( BallInLeftHalf() ? 180 : 0 );
  }
  // Top edge of field - team relative...
  else if( LineNum == 1 ) {
    // Angle of approach depends on the ball position...
    return( BallInLeftHalf() ? 0 : 180 );
  }
  else {
    cout << "\n Cannot trust stupid humans to do any math right, can we? \n";
    return -1;
  }
}



/**
 * Mohan::Function returns true if the robot is in the opponent's half
 * of the field...
 */
bool WorldState::InOpponentHalf() {
  double x = SmoothPose.Position.getX();
  double y = SmoothPose.Position.getY();
  
  double buffer = 250;
  return( x >= MINX &&
	  x <= MAXX &&
	  y >= buffer
	  );
}



/**
 * Mohan::Function returns true iff the robot is in the top third of
 * the field (team-relative coordinates)...
 */
bool WorldState::RobotInTopThird() {
  double x = SmoothPose.Position.getX();
  double y = SmoothPose.Position.getY();
  
  double buffer = 100;
  return( x >= MINX && x <= MAXX &&
	  y >= ( MINY + ( 2.0/3.0 * FIELD_LENGTH ) + buffer ) );
}



/**
 * Mohan::Fucntion returns true if the robot is the closest to the
 * ball...
 */
bool WorldState::ClosestToBall() {
  if( BallConfidence < BALL_WELL_KNOWN_CONF ) {
    return false;
  }
  double myballdist = particleBallEstimate.getMagnitude();

  for( int i = 0; i < NUM_TEAM_MEMBERS; i ++ ) {
    if( i == M_ID || teammateState[i].Role == KEEPER_ROLE ) {
      continue;
    }
    if( ( getTime()-teammateState[i].lastBallMessageTime ) < 1.0  && 
	teammateState[i].BallDist < myballdist ) {
      return false;
    }
  }
  return true;
}



/**
 * Mohan::Function returns the supporting position that the robot
 * needs to go to...
 */
Point2D WorldState::GetSupportPoint() {
  double supportX = 0, supportY = 0;
  // For a defender the support point is the same as the defensive
  // post...
  if( currentRole == DEFENDER_ROLE ) {
    // Set the x,y-coordinates based on the ball location on the
    // field...
    return( GetDefensivePost() );
  }

  // Nate: the point returned by this function is fluctuating wildly
  // over small periods of time.  This is an attempt to simplify it to
  // reduce variance.
  else if( true ) {

    // if we've just transitioned into the supporter behavior, we'll
    // reset the support point
    if( resetSupportPoint ) {
      resetSupportPoint = false;

      if( BallConfident() ) {

	if( globalBallEstimate.getX() > 0 ) {
	  supportPoint.setX( GOAL_BOX_X_MAX );
	} else {
	  supportPoint.setX( GOAL_BOX_X_MIN );
	}

	// if we're not supposed to go back, stay in the top half of the
	// field
	if( !robotGoBack ) {
	  if( globalBallEstimate.getY() < 0 ) {
	    supportPoint.setY( 0 );
	  } else {
	    supportPoint.setY( globalBallEstimate.getY() - 300 );
	  }
	}

	// if we're supposed to go back, allow the robot to choose
	// y-positions all over the field
	else if( robotGoBack ) {
	  if( globalBallEstimate.getY() < 0 ) {
	    supportPoint.setY( globalBallEstimate.getY() + 300 );
	  } else {
	    supportPoint.setY( globalBallEstimate.getY() - 300 );
	  }
	}
	
      } 
      // not confident about the ball
      else {
	// if we're not supposed to go back, stay in the top half of
	// the field...
	if( !robotGoBack ) {
	  supportPoint.setX( GOAL_BOX_X_MAX );
	  supportPoint.setY( MAXY / 2.0 );
	} else if( robotGoBack ) {
	  supportPoint.setX( GOAL_BOX_X_MIN );
	  supportPoint.setY( 0 );
	}	
      }

      if( supportPoint.getY() > MAXY-800 ) supportPoint.setY( MAXY-800 );
      else if( supportPoint.getY() < MINY+800 ) supportPoint.setY( MINY+800 );
      
      double tempx = supportPoint.getX();
      double tempy = supportPoint.getY();

      supportPoint.setX( minmax( MINX, tempx, MAXX ) );
      supportPoint.setY( minmax( MINY, tempy, MAXY ) );
     
      return supportPoint;
    }
    
    // if we're not confident in the ball position, don't change the
    // support point
    if( !BallConfident() ) {
      return supportPoint;
    }

    // if we need to change from the left side of the field to the right
    if( supportPoint.getX() < 0 &&
	globalBallEstimate.getX() < -200 ) {
      supportPoint.setX( GOAL_BOX_X_MAX );
    }
    // if we need to change from the right side of the field to the left
    else if( supportPoint.getX() > 0 &&
	     globalBallEstimate.getX() > 200 ) {
      supportPoint.setX( GOAL_BOX_X_MIN );
    }
    
    // if we're not supposed to go back, stay in the top half of the
    // field
    if( !robotGoBack ) {
      // if the difference in Y is big enough, reset the y-coordinate
      // of the support point
      if( fabs(supportPoint.getY()-globalBallEstimate.getY()) > 800 ) {
	if( globalBallEstimate.getY() < 0 ) {
	  supportPoint.setY( 0 );
	} else {
	  supportPoint.setY( globalBallEstimate.getY() - 300 );
	}
      }

    }

    // if we're supposed to go back, allow the robot to choose
    // y-positions all over the field
    else if( robotGoBack ) {
      // if the difference in Y is big enough, reset the y-coordinate
      // of the support point
      if( fabs(supportPoint.getY()-globalBallEstimate.getY()) > 800 ) {
	if( globalBallEstimate.getY() < 0 ) {
	  supportPoint.setY( globalBallEstimate.getY() + 300 );
	} else {
	  supportPoint.setY( globalBallEstimate.getY() - 300 );
	}
      }      
    }

    if( supportPoint.getY() > MAXY-800 ) supportPoint.setY( MAXY-800 );
    else if( supportPoint.getY() < MINY+800 ) supportPoint.setY( MINY+800 );

    double tempx = supportPoint.getX();
    double tempy = supportPoint.getY();
    
    supportPoint.setX( minmax( MINX, tempx, MAXX ) );
    supportPoint.setY( minmax( MINY, tempy, MAXY ) );
    
    return supportPoint;
  }


  ////////////////////////////////////////////////////////////////////////
  // Nate: THE BLOCK BELOW DOESN'T EVER GET EXECUTED!!!!!!!!!!!!!!!!!!!!
  ////////////////////////////////////////////////////////////////////////

  // For the attackers, if the ball is in the defensive half, support
  // point should be appropriately in its own half: a point depending
  // on the position of the ball...
  else if( !BallInOpponentHalf() && robotGoBack ) {
    // Position as a function of the ball position - fix up the
    // default coordinates first...
    // cout << "\n Here... \n";
    if( BallInLeftHalf() ) {
      supportX = GOAL_BOX_X_MAX;
    }
    else if( BallInRightHalf() ) {
      supportX = GOAL_BOX_X_MIN;
    }
    else {
      supportX = FIELD_CENTER_X;
    }

    Point2D post = Point2D( supportX, ATT_SUPDEF_Y );
    
    // Next, set the x-ccordinate based on the ball x-slope...
    // It comes down to be 1/2...
    double slope_x = ( FIELD_WIDTH/ 4 ) / ( FIELD_WIDTH/2 ); 
    
    // To be used later...
    // double slope_x = ( GOAL_BOX_WIDTH/ 2 ) / FIELD_WIDTH; 
    double x_diff = minmax( MINX, globalBallEstimate.getX(), MAXX ) * slope_x;
    post.setX( post.getX() + x_diff );

    // Set the y-coordinates based on the ball y-slope...
    double slope_y = ( ( MAXY/4 - OWN_GOAL_BOX_Y_MAX ) /
		       ( FIELD_CENTER_Y - MINY ) );
    double y_diff = minmax( MINY, globalBallEstimate.getY(), FIELD_CENTER_Y ) * slope_y;
    post.setY( post.getY() + y_diff );

    return post;
  }

  // For the attackers, if the command is not given to go back, the
  // support position should be based on the ball's position... 
  else {
    // If ball is in the defensive half, the robot should move back
    // closer to the field center...
    if( !BallInOpponentHalf() ) {
      // cout << "\n There - 1 ... \n";
      if( BallInLeftHalf() ) {
	supportX = GOAL_BOX_X_MIN;
      }
//       else if( BallInRightHalf() ) {
// 	supportX = GOAL_BOX_X_MAX;
//       }
      else {
	supportX = GOAL_BOX_X_MAX;
      }
      Point2D post = Point2D( supportX, MAXY/2 );
      
      // Set the y-coordinates based on the ball y-slope...
      double slope_y = ( ( post.getY() - MINY/3 ) /
			 ( FIELD_CENTER_Y - MINY ) );
      double y_diff = minmax( MINY, globalBallEstimate.getY(), FIELD_CENTER_Y ) * slope_y;
      post.setY( post.getY() + y_diff );
    
      return post;
    }
    // If the ball is in the offensive half, the support point should
    // be slightly ahead and to one side of the ball...
    else {
      // cout << "\n There - 2 ... \n";
      supportY = MAXY/2; // OPP_GOAL_BOX_Y_MIN - 300;
      if( BallInLeftHalf() ) {
	supportX = GOAL_BOX_X_MAX;
      }
//       else if( BallInRightHalf() ) {
// 	supportX = GOAL_BOX_X_MIN;
//       }
      else {
	supportX = GOAL_BOX_X_MIN;
	// supportX = FIELD_CENTER_X;
      }

      Point2D post = Point2D( supportX, supportY );
      double buffer = 150;

      // Set the y-coordinates based on the ball y-slope...
      double slope_y = ( ( OPP_GOAL_BOX_Y_MIN - buffer - supportY ) /
			 ( MAXY - FIELD_CENTER_Y ) );
      double y_diff = minmax( FIELD_CENTER_Y, globalBallEstimate.getY(), MAXY ) * slope_y;
      post.setY( post.getY() + y_diff );
      
      return post;
    }

    // The robot, while walking from its current position to the
    // support position, should not pass through the current known
    // ball position - this causes weird looking role switches - TO
    // DO...
  }
}




/**
 * Mohan::Function returns true iff the current (well known) position
 * of the ball is in the left half of the field...
 */
bool WorldState::BallInLeftHalf() {
  if( BallConfidence < BALL_WELL_KNOWN_CONF ) {
    return false;
  }

  double buffer = 150;
  return globalBallEstimate.getX() <= buffer;
}




/**
 * Mohan::Function returns true iff the current (well known) position
 * of the ball is in the right half of the field...
 */
bool WorldState::BallInRightHalf() {
  if( BallConfidence < BALL_WELL_KNOWN_CONF ) {
    return false;
  }

  double buffer = 150;
  return globalBallEstimate.getX() >= -buffer;
}




/**
 * Mohan::Function returns true iff the robot working as the defender
 * role is in the half with the beacons that have pink on top...
 */
bool WorldState::DefenderInLeftHalf() {
  double buffer = 150;
  for( int i = 0; i < NUM_TEAM_MEMBERS; i ++ ) {
    if( i == M_ID || teammateState[i].Role != DEFENDER_ROLE ) {
      continue;
    }
    
    // If the robot playing the defender is on the left and within a
    // buffer region of the field center x-coordinate, return true...
    if ( time - teammateState[i].lastPositionMessageTime < 1 ) {
      return teammateState[i].pose.Position.getX() <= buffer;
    }
  }
  return false;
}



/**
 * Mohan::Function returns true iff the robot working as the defender
 * role is in the half with the beacons that have pink at the
 * bottom...
 */
bool WorldState::DefenderInRightHalf() {
  double buffer = 150;
  for( int i = 0; i < NUM_TEAM_MEMBERS; i ++ ) {
    if( i == M_ID || teammateState[i].Role != DEFENDER_ROLE ) {
      continue;
    }

    // If the robot playing the defender is on the right and within a
    // buffer region of the field center x-coordinate, return true...
    if ( time - teammateState[i].lastPositionMessageTime < 1 ) {
      return teammateState[i].pose.Position.getX() >= -buffer;
    }
  }
  return false;
}



/**
 * Mohan::Function returns true if the robot playing as the attacker
 * and actually going to ball is on the left side (pink on top in
 * beacons) of the field...
 */
bool WorldState::AttackerInLeftHalf() {
  double buffer = 150;
  for( int i = 0; i < NUM_TEAM_MEMBERS; i ++ ) {
    if( i == M_ID || teammateState[i].Role != ATTACKER_ROLE ) {
      continue;
    }
    // If the robot playing as the attacker and going to ball
    // (currently) is on the left and within a buffer region of the
    // field center x-coordinate, return true...
    if( ( time - teammateState[i].lastPositionMessageTime < 1 ) && 
	teammateState[i].goToBall ) {
      return teammateState[i].pose.Position.getX() <= buffer;
    }
  }
  return false;
}



/**
 * Mohan::Function returns true if the robot playing as the attacker
 * and actually going to ball is on the right side (pink at bottom in
 * beacons) of the field...
 */
bool WorldState::AttackerInRightHalf() {
  double buffer = 150;
  for( int i = 0; i < NUM_TEAM_MEMBERS; i ++ ) {
    if( i == M_ID || teammateState[i].Role != ATTACKER_ROLE ) {
      continue;
    }
    // If the robot playing as the attacker and going to ball
    // (currently) is on the right and within a buffer region of the
    // field center x-coordinate, return true...
    if( ( time - teammateState[i].lastPositionMessageTime < 1 ) && 
	teammateState[i].goToBall ) {
      return teammateState[i].pose.Position.getX() >= -buffer;
    }
  }
  return false;
}



/**
 * Mohan::Function returns true if the robot is alone - not in
 * communication with any other robot...
 */
bool WorldState::RobotAlone() {
  for( int i = 0; i < NUM_TEAM_MEMBERS; i ++ ) {
    if( i == M_ID ) {
      continue;
    }
    if( alive[M_ID][i] ) {
      return false;
    }
  }
  return true;
}


/**
 * Mohan::Function returns true if the robot is reasonable well
 * localized - certainty above a constant...
 */
bool WorldState::WellLocalized() {
  return( SmoothPose.Confidence >= TEAMMATE_POSE_CONF_THRESH_LOW );
}



/**
 * Mohan::Function returns true is the keeper is at its post...
 */
bool WorldState::KeeperInPlace() {
  static double BUFF_DIST = 150;
  static double BUFF_ANG = 15;

  Point2D currpost = SmoothPose.Position;
  AngDeg currang = SmoothPose.Orientation;

  Pose Kp = GetKeeperDefaultPose();

  return( currpost.getDistanceTo( Kp.Position ) <= BUFF_DIST &&
	  fabs( currang - Kp.Orientation ) <= BUFF_ANG );
}



/**
 * Mohan:;Function returns the base position that the keeper has to
 * attain under default conditions...
 */
Pose WorldState::GetKeeperDefaultPose() {
  Pose Keeper;
  Keeper.Position.setPoint( KEEPER_X, KEEPER_Y );
  Keeper.Orientation = 0;
  Keeper.Confidence = 1;
  return Keeper;
}


/**
 * Mohan::Function returns the keeper position when the ball
 * position is known - a function of the ball position...
 */
Pose WorldState::GetKeeperPose() {
  double y_buffer = 150;
  double x_buffer = 100;

  // Position as a function of the ball position...  Fix up the
  // default coordinates first...
  Pose BasePose;
  BasePose.Position = Point2D( KEEPER_X, KEEPER_Y+y_buffer );
  BasePose.Orientation = 0.0;

  // Set the y-coordinates based on the ball y-slope - the robot stays
  // ahead of default y-ccord by the buffer distance and then walks
  // back as the ball moves from the field center to miny...
  double slope_y = ( y_buffer / ( FIELD_CENTER_Y - MINY ) );
  
  double y_diff = minmax( MINY, globalBallEstimate.getY(), FIELD_CENTER_Y ) * slope_y;
  BasePose.Position.setY( BasePose.Position.getY() + y_diff );
  
  // Next, set the x-ccordinate based on the ball x-slope - the robot
  // should move from the center to the edge of the goal as the ball
  // moves from the center to the edge of the field...
  double slope_x = ( (GOAL_INTERIOR_X - x_buffer)/2.0 ) / ( FIELD_WIDTH/2.0 ); 

  double x_diff = minmax( MINX, globalBallEstimate.getX(), MAXX ) * slope_x;
  BasePose.Position.setX( BasePose.Position.getX() + x_diff );
  
  // Robot desired orientation based on the angle from the x-axis
  // made by the line pointing from the robot's current position
  // to the global ball location - also converted wrt the y-axis
  // and bounded within limits...
  double Robot2BallAngle = ( globalBallEstimate - 
			     SmoothPose.Position ).getDirection();
  Robot2BallAngle = minmax( -45, normalizeAngle( Robot2BallAngle - 90 ), 45 );
  BasePose.Orientation = normalizeAngle( Robot2BallAngle );

  return BasePose;
}


/**
 * Returns true if an opponent is to the left and close enough to
 * interfere with ball acquisition.  -peggy
 */
bool WorldState::OpponentImmediatelyLeft() {
  static double OppDistThresh = 400;
  Point2D opp, relopp;
  double dist, ang;

  if (PrevNumOpps <= 0) {
    return false;
  }
  
  for (int i = 0; i < PrevNumOpps; i++) {
    opp = Opponents[i].Gms.mean;
    relopp = opp.globalToRelative( SmoothPose.Position, SmoothPose.Orientation );
    dist = sqrt( relopp.getX() * relopp.getX() + relopp.getY() * relopp.getY() );
    ang = atan2Deg( relopp.getY(), relopp.getX() ) - 90;
    if( dist <= OppDistThresh && ang > 0  ) {
      return true;
    }
  }
  return false;
}

/**
 * Returns true if an opponent is to the right and close enough to
 * interfere with ball acquisition.  -peggy
 */
bool WorldState::OpponentImmediatelyRight() {
  static double OppDistThresh = 400;
  Point2D opp, relopp;
  double dist, ang;

  if (PrevNumOpps <= 0) {
    return false;
  }
  
  for (int i = 0; i < PrevNumOpps; i++) {
    opp = Opponents[i].Gms.mean;
    relopp = opp.globalToRelative( SmoothPose.Position, SmoothPose.Orientation );
    dist = sqrt( relopp.getX() * relopp.getX() + relopp.getY() * relopp.getY() );
    ang = atan2Deg( relopp.getY(), relopp.getX() ) - 90;
    if( dist <= OppDistThresh && ang < 0  ) {
      return true;
    }
  }
  return false;
}

/**
 * Returns true if an opponent is both closer than maxdist and in the specified heading
 * range.  -peggy  (warning...this won't work right if you expect the heading range to
 * include the discontinuity at 180/-180)
 */
bool WorldState::OpponentCloseInHeadingRange(double maxdist, 
					     double minheading, 
					     double maxheading) {
  Point2D opp, relopp;
  double dist, heading;

  if ((PrevNumOpps <= 0) || (minheading > maxheading)) {
    return false;
  }

  for (int i=0; i < PrevNumOpps; i++) {
    opp = Opponents[i].Gms.mean;
    relopp = opp.globalToRelative( SmoothPose.Position, SmoothPose.Orientation );
    dist = sqrt(relopp.getX() * relopp.getX() + relopp.getY() * relopp.getY());
    heading = normalizeAngle(atan2Deg(relopp.getY(), relopp.getX()) - 90);
    if (dist < maxdist && heading > minheading && heading < maxheading) {
      return true;
    }
  }
  return false;
}

/**
 * Returns true if any opponent is within threshold of this robot.  -peggy
 */
bool WorldState::OpponentWithinDist(double threshold) {
  Point2D opp, relopp;
  double dist;

  if (PrevNumOpps <= 0) {
    return false;
  }
  
  for (int i = 0; i < PrevNumOpps; i++) {
    opp = Opponents[i].Gms.mean;
    relopp = opp.globalToRelative( SmoothPose.Position, SmoothPose.Orientation );
    dist = sqrt( relopp.getX() * relopp.getX() + relopp.getY() * relopp.getY() );
    if( dist <= threshold ) {
      return true;
    }
  }
  return false;
}


/**
 * Mohan::Function that returns true if there exists an opponent that
 * is within buffer region of 30 degrees and a distance of 500mm from
 * the robot...  (i.e. if the opponent is in front and close enough to
 * interfere with ball acqusition. -p)
 */
bool WorldState::OpponentInWay() {
  static double OppDistTH = 500;
  static double OppAngTH = 30;

  if( PrevNumOpps <= 0 ) {
    return false;
  }
  Point2D opp, relopp;
  double dist, ang;

  for( int i = 0; i < PrevNumOpps; i++ ) {
    opp = Opponents[i].Gms.mean;
    relopp = opp.globalToRelative( SmoothPose.Position, SmoothPose.Orientation );
    dist = sqrt( relopp.getX() * relopp.getX() + relopp.getY() * relopp.getY() );
    ang = atan2Deg( relopp.getY(), relopp.getX() ) - 90;
    if( dist <= OppDistTH && fabs(ang) <= OppAngTH ) {
      return true;
    }
  }
  return false;
}


/**
 * Mohan::Function returns true if there is atleast one opponent near
 * (within a threshold buffer region around) the goal box - used for
 * keeper attack:-)...
 */
bool WorldState::OpponentNearOwnGoalBox() {
  static double buffer = 100;

  if( PrevNumOpps <= 0) {
    return false;
  }

  double posx, posy;
  for( int i = 0; i < PrevNumOpps; i ++ ) {
    posx = Opponents[i].Gms.mean.getX();
    posy = Opponents[i].Gms.mean.getY();
    
    if( posx >= ( GOAL_BOX_X_MIN - buffer ) &&
	posx <= ( GOAL_BOX_X_MAX + buffer ) &&
	posy <= ( OWN_GOAL_BOX_Y_MAX + buffer )
	) {
      return true;
    }
  }
  
  return false;
}



/**
 * Mohan::Function that returns the position (Point2D in global
 * coordinates) of the opponent that is the closest to the robot's own
 * goal box...
 */
Point2D WorldState::GetOppClosestToOwnGoalBox() {
  static double buffer = 100;

  if( !OpponentNearOwnGoalBox() ) {
    return Point2D(0,0);
  }
  
  double posx, posy;
  double tmpoppdist, oppdist = 10000;
  Point2D Opppos, RetOpp;
  RetOpp.setPoint( 0, 0 );

  for( int i = 0; i < PrevNumOpps; i ++ ) {
    Opppos = Opponents[i].Gms.mean;
    posx = Opppos.getX();
    posy = Opppos.getY();
    
    if( posx >= ( GOAL_BOX_X_MIN - buffer ) &&
	posx <= ( GOAL_BOX_X_MAX + buffer ) &&
	posy <= ( OWN_GOAL_BOX_Y_MAX + buffer )
	) {
      tmpoppdist = SmoothPose.Position.getDistanceTo( Opppos );
      if( tmpoppdist <= oppdist ) {
	RetOpp = Opppos;
	oppdist = tmpoppdist;
      }
    }
  }
  return RetOpp;
}



/**
 * Mohan::Function that provides the defensive support point that a
 * robot (usually the defender) needs to go to...
 */
Point2D WorldState::GetDefensivePost() {
   // Position as a function of the ball position...  Fix up the
   // default coordinates first...
   Point2D post = Point2D( DEFPOST_X, DEFPOST_Y );
   
   // Set the y-coordinates based on the ball y-slope...
//    double slope_y = ( ( FIELD_CENTER_Y - DEFPOST_Y ) /
// 		      ( MAXY - FIELD_CENTER_Y ) );
   double slope_y = 0.0;
   
   double y_diff = minmax( FIELD_CENTER_Y, 
			   globalBallEstimate.getY(), MAXY ) * slope_y;
   post.setY( post.getY() + y_diff );
 
   // Next, set the x-ccordinate based on the ball x-slope...
   // It comes down to be 1/2...
   // double slope_x = ( FIELD_WIDTH/ 4 ) / ( FIELD_WIDTH/2 ); 
   double slope_x = 1.0/2.0; // 2.0/3.0; 
   
   // To be used later...
   // double slope_x = ( GOAL_BOX_WIDTH/ 2 ) / FIELD_WIDTH; 
   double x_diff = minmax( MINX, globalBallEstimate.getX(), MAXX ) * slope_x;
   post.setX( post.getX() + x_diff );
 
  //  cout << "\n Target pose (" << post.getX() << "," << post.getY()
// 	<< ")" << endl;
   return post;
}



/**
 * Mohan::Function returns true if the robot is within a threshold
 * distance from a field border...
 */
bool WorldState::CloseToBorders() {
  //  if( SmoothPose.Confidence < TEAMMATE_POSE_CONF_THRESH ) {
  if( SmoothPose.Confidence < 0.65 ) {
    return false;
  }
  
  double buffer = 300;
  double dists[4];

  // Calculate distances from each of the four field borders...
  dists[0] = lbot.getDistanceToPoint( SmoothPose.Position );
  dists[1] = ltop.getDistanceToPoint( SmoothPose.Position );
  dists[2] = llft.getDistanceToPoint( SmoothPose.Position );
  dists[3] = lrgt.getDistanceToPoint( SmoothPose.Position );

  // Calculate the minimum distance from a field border...
  double mindist = 1000;
  for( int i = 0; i < 4; i ++ ) {
    if( dists[i] <= mindist ) {
      mindist = dists[i];
    }
  }

  if( mindist <= buffer ) {
    return true;
  }
  return false;
}




/**
 * Mohan::Function returns true iff the robot cannot get to the ball -
 * say stuck because of opponents, field walls etc...
 */
bool WorldState::CannotGetToBall() {
  static double BALL_DIST_UNCHANGED_THRESH = 100;
  // If the robot is not sure of the ball location it cannot make a
  // decision regarding whether it can get to the ball or not...
  if( BallConfidence < BALL_WELL_KNOWN_CONF ) {
    return false;
  }

  double minbd = NO_VISION_DIST, maxbd = 0;

  for( int i = 0; i < MAX_BALL_DISTANCES; i ++ ) {
    if( BallDistsArr[i] >= maxbd ) {
      maxbd = BallDistsArr[i];
    }
    if( BallDistsArr[i] <= minbd ) {
      minbd = BallDistsArr[i];
    }
  }

  if( BallDistsArr[BallDistID] >= 3 * BALL_DIST_UNCHANGED_THRESH && 
      fabs( maxbd - minbd ) <= BALL_DIST_UNCHANGED_THRESH ) {
    return true;
  }
  return false;
}



/**
 * Mohan::Function returns true iff the ball is withing the robot's
 * own goal box. Thr robot is assumed to be well localized and is
 * assumed to have a good idea of the ball...
 */
bool WorldState::BallInOwnGoalBox() {
  double x = globalBallEstimate.getX();
  double y = globalBallEstimate.getY();

  return( x >= GOAL_BOX_X_MIN &&
	  x <= GOAL_BOX_X_MAX &&
	  y <= OWN_GOAL_BOX_Y_MAX );
}




/**
 * Mohan::Function returns true iff the current path from the robot to
 * the ball takes the robot through its own goal box - the function
 * returns false if the robot has a low confidence of its pose and/or
 * the ball position...
 */
bool WorldState::PathThroughGoalBox() {
  if( SmoothPose.Confidence < TEAMMATE_POSE_CONF_THRESH ||
      BallConfidence < BALL_WELL_KNOWN_CONF ||
      BallInOwnGoalBox() ) {
    // cout << "\n Cannot make a decision here \n";
    return false;
  }
  // Calculate the line joining the robot's current position to the
  // ball's current position, in global frame of reference...
  Line2D RobotLineToBall = Line2D::makeLineFromTwoPoints( SmoothPose.Position, 
							  globalBallEstimate );
  
  NumGBInts = 0;
  Point2D Inter1, Inter2, Inter3;
  
  // Depending on the team (red or blue) determine the intersections
  // and the number of intersections (if any) that the the robot
  // proposed path to the ball would make with the goal box lines
  // (defined in the constructor)...
  Inter1 = RobotLineToBall.getMyIntersection( LineA );
  Inter2 = RobotLineToBall.getMyIntersection( LineB );
  Inter3 = RobotLineToBall.getMyIntersection( LineC );
  
  if( Inter1.isBetweenY( LineAP1, LineAP2 ) &&
      Inter1.PointIsBetweenY( SmoothPose.Position, globalBallEstimate ) ) {
    GBIntersects[ NumGBInts ] = Inter1;
    ++NumGBInts;
  }
  if( Inter2.isBetweenX( LineBP1, LineBP2 ) &&
      Inter2.PointIsBetweenY( SmoothPose.Position, globalBallEstimate ) ) {
    GBIntersects[ NumGBInts ] = Inter2;
    ++NumGBInts;
  }
  if( Inter3.isBetweenY( LineCP1, LineCP2 ) &&
      Inter3.PointIsBetweenY( SmoothPose.Position, globalBallEstimate ) ) {
    GBIntersects[ NumGBInts ] = Inter3;
    ++NumGBInts;
  }

  // If there are one (or more - though > 2 is not possible)
  // intersections with the goal box lines, it means that the robot is
  // going to walk through the box if it follows its intended path...
  if( NumGBInts >= 1 ) {
//     for( int i = 0; i < NumGBInts; i ++ ) {
//       cout << "\n Intersection is (" << GBIntersects[i].getX() << ","
// 	   << GBIntersects[i].getY() << ")"; 
//     }
    // Next, we determine the intersection point that is closest to
    // the robot - first intialize the variable to correspond to the
    // first intersection point in the list...
    ClosestInt = GBIntersects[0];
    double cdist = SmoothPose.Position.getDistanceTo( GBIntersects[0] );

    double tempdist;
    // Determine the closest intersection point...
    for( int i = 0; i < NumGBInts; i ++ ) {
      tempdist = SmoothPose.Position.getDistanceTo( GBIntersects[i] );
      if( tempdist <= cdist ) {
	cdist = tempdist;
	ClosestInt = GBIntersects[i];
      }
    }
    // Finally we need to determine the new target location around the
    // goal box that the robot has to now approach to avoid entering
    // into the goal box (to avoid the 30 second illegal defender
    // penalty). But this shall happen in another function (below) -
    // return true now to emphasize this situation...
    return true;
  }
  else {
    return false;
  }
}



/**
 * WARNING::FUNCTION MAKES SENSE IFF THE FUNCTION ABOVE RETURNS TRUE.
 * Mohan::Given that the current path of the robot to the ball does
 * involve walking right through its own penalty zone, this function
 * determines the "safe" point to walk to around the goal box...
 */
Point2D WorldState::NewGBTargetLocation() {
  Point2D TargetLoc( -1, -1 );
//   if( !PathThroughGoalBox() ) {
//     return TargetLoc;
//   }
  // The simplest way to accomplish this objective is to determine the
  // goal box corner (one of the two corners that are approx. 600mm
  // from the horizontal edges of the field) closest to the nearest
  // intersection point and move there, taking care to this point is
  // on the path to the ball...
  double compdist = 1e7, tmpdist;

  if( LineAP2.PointIsBetweenX( SmoothPose.Position, globalBallEstimate ) ) {
    tmpdist = ClosestInt.getDistanceTo( LineAP2 );
    if( tmpdist <= compdist ) {
      compdist = tmpdist;
      TargetLoc = LineAP2 + Point2D( -200, 200 );
    }
  }
  
  if( LineBP2.PointIsBetweenX( SmoothPose.Position, globalBallEstimate ) ) {
    tmpdist = ClosestInt.getDistanceTo( LineBP2 );
    if( tmpdist <= compdist ) {
      compdist = tmpdist;
      TargetLoc = LineBP2 + Point2D( 200, 200 );
    }
  }

//   cout << "\n (Pose,Ball)=(" << SmoothPose.Position.getX() << "," 
//        << SmoothPose.Position.getY() << "," << globalBallEstimate.getX() << ","
//        << globalBallEstimate.getY() << ")";

//   cout << "\n Target Location (" << TargetLoc.getX() << ","
//        << TargetLoc.getY() << ")" << endl;

  return TargetLoc;
}



/**
 * Mohan::Function takes in the (x,y,z) point in robot-body-center
 * coordinate frame and returns an overall tilt and pan in the
 * robot-neck-centric coordinate frame...
 */
void WorldState::GetAngles2Point3D( double relx, double rely, double relz, 
				    double &tilt, double &pan ) {
  Matrix InPoint(4,1);
  InPoint.M[0][0] = relx;   InPoint.M[1][0] = rely;    
  InPoint.M[2][0] = Body_Height + relz;

  Matrix InvertedBody = Body.Inverse();
  Matrix OutPoint = Body2Neck * InvertedBody * InPoint;

  double planedist = sqrt( pow( OutPoint.M[0][0], 2) + pow( OutPoint.M[1][0], 2) );

  tilt = atanDeg( OutPoint.M[2][0] / planedist );
  pan = -atan2Deg( OutPoint.M[0][0], OutPoint.M[1][0] );
}



/**
 * Mohan::Function takes the current velocities into account to decide
 * the new point to reach so as to avoid collisions...
 */
Point2D WorldState::AntiCollisionPoint() {
  // Default base position...
  Point2D newpost = SmoothPose.Position;
  Point2D offset = Point2D( 0, 0 );
  
  double buffer_dist = 300;

  // FInd the direction the robot needs to move to avoid colliding -
  // for now, this shall be along the line joining the current
  // position to the center of the field...
  Point2D field_center = Point2D( FIELD_CENTER_X, FIELD_CENTER_Y ); 
  AngDeg offset_dir = ( field_center - SmoothPose.Position ).getDirection();

  offset = Point2D( buffer_dist, offset_dir, POLAR );

  cout << "\n Have to get to: (" << (newpost+offset).getX()
       << "," << (newpost+offset).getY() << ") \n";

  return( newpost+offset );
}



/**
 * Mohan::Function that takes the accelrometer statistics as inputs to
 * decide if the robot is colliding or not - returns true iff the
 * robot is colliding...
 */
bool WorldState::CollidingAmI() {
  double varx = V_x;
  //  double vary = V_y;
  double varz = V_z;

  double rhoxy = R_xy;
  //  double rhoyz = R_yz;
  double rhozx = R_zx;

  if( varx <= 0.424334 ) { 
    return false; 
  }
  else { // if( varx > 0.424334 ) {
    if( varx <= 6.493036 ) {
      if( varz <= 0.829458 ) {
	if( rhoxy <= 0.509448 ) {
	  if( rhozx <= -0.243248 ) {
	    return false;
	  }
	  else { //if( rhozx > -0.243248 ) {
	    if( varx <= 1.922249 ) {
	      if( rhozx <= 0.450594 ) {
		return true;
	      }
	      else { //if( rhozx > 0.450594 ) {
		return false;
	      }
	    }
	    else { //if( varx > 1.922249 ) {
	      return false;
	    }
	  }
	}
	else { // if( rhoxy > 0.509448 ) {
	  return false;
	}
      }
      else { //if( varz > 0.829458 ) {
	return false;
      }
    }
    else { //if( varx > 6.493036 ) {
      return true;
    }
  }
}



/**
 * Mohan::Function stores values of Accelerometers over a few frames
 * and then computes the sample to test for collision...
 */
void WorldState::GenerateTestSample() {
  // For the first N frames, the robot only stores the accelerometer
  // values along the three axes - noise filtering by averaging
  // later...
  if( NumAccVals < COLL_FRAMES ) {
    AccelValsStore( 1e-6 * sensors[ACCELEROMETER_X], 
		    1e-6 * sensors[ACCELEROMETER_Y], 
		    1e-6 * sensors[ACCELEROMETER_Z] );
    NewCollVals = false;
  }
  // After the first N frames of value storage, it is time to generate
  // a single test sample from these...
  else {
    // Generate test sample by the same function used to generate the
    // training sample...
    GenerateStats();
    
    // Initialize array values to start averaging again...
    NumAccVals = 0;
    accsumx = 0; accsumy = 0; accsumz = 0;
  }  
}



//##################################################################################3
/**
 * Mohan::Function that stores the values of the accelerometer valus
 * into the array to be used for averaging and generating training
 * samples later...
 */
void WorldState::AccelValsStore( double accelx, double accely, double accelz ) {
  // Sanity check...
  if( NumAccVals >= 40 ) {
    NumAccVals = 0;
    cout << "\n DUMBO - you are exceeding array bounds... \n";
  }
  
  XYZAccelStore[NumAccVals][0] = accelx;
  accsumx += accelx;

  XYZAccelStore[NumAccVals][1] = accely;
  accsumy += accely;

  XYZAccelStore[NumAccVals][2] = accelz;
  accsumz += accelz;

  ++NumAccVals;
}



/**
 * Mohan::Function that calculates the statistics for the current
 * sample based on the values stored in the last 1 second - generates
 * one training sample...
 */
void WorldState::GenerateStats() {
  // Sanity Check...
  if( NumAccVals < 10 ) {
    cout << "\n Less than 10 values in a second - something wrong \n";
  }

  double meanx, meany, meanz;
  double varx = 0, vary = 0, varz = 0;
  double covxy = 0, covyz = 0, covzx = 0;

  meanx = accsumx/NumAccVals;
  meany = accsumy/NumAccVals;
  meanz = accsumz/NumAccVals;
  
  for( int i = 0; i < NumAccVals; i ++ ) {
    // Calculate variance sums...
    varx += pow( ( XYZAccelStore[i][0] - meanx ), 2 );
    vary += pow( ( XYZAccelStore[i][1] - meany ), 2 );
    varz += pow( ( XYZAccelStore[i][2] - meanz ), 2 );

    // Cslculate covariance sums...
    covxy += ( ( XYZAccelStore[i][0] - meanx ) * ( XYZAccelStore[i][1] - meany ) );
    covyz += ( ( XYZAccelStore[i][1] - meany ) * ( XYZAccelStore[i][2] - meanz ) );
    covzx += ( ( XYZAccelStore[i][2] - meanz ) * ( XYZAccelStore[i][0] - meanx ) );
  }

//   OSYSPRINT(("\n Means=(%f,%f,%f), Vars=(%f,%f,%f) \n", 
// 	     meanx, meany, meanz, varx, vary, varz));

//   OSYSPRINT(("\n Covs=(%f,%f,%f) Class=%f \n", covxy, covyz, covzx, CollClass));

  if( varx >= 1e6 || vary >= 1e6 || varz >= 1e6 ) {
    cout << "\n Values too BIG \n";
    return;
  }
  else if( vary <= 1e-6 || vary <= 1e-6 || varz <= 1e-6 ||
	   fabs(covxy) <= 1e-6 || fabs(covyz) <= 1e-6 || 
	   fabs(covzx) <= 1e-6 ) {
    cout << "\n Values too SMALL \n";
  }
  else {
    // Calculate correlation coefficients - "NumAccVals" cancel out
    // between numerator and denominator...
    double rhoxy = covxy / sqrt( varx * vary );
    double rhoyz = covyz / sqrt( vary * varz );
    double rhozx = covzx / sqrt( varz * varx );

    // But actual unbiased variance estimates are necessary...
    varx /= ( NumAccVals - 1 );
    vary /= ( NumAccVals - 1 );
    varz /= ( NumAccVals - 1 );

    // If only testing is required right now, return with just the
    // calculated values...
    if( COLLISION_DETECT ) {
      V_x = varx;      V_y = vary;      V_z = varz;
      R_xy = rhoxy;    R_yz = rhoyz;    R_zx = rhozx;
      NewCollVals = true;
      return;
    }

    CalculatedValues[ NumCalVals ][0] = varx;
    CalculatedValues[ NumCalVals ][1] = vary;
    CalculatedValues[ NumCalVals ][2] = varz;

    CalculatedValues[ NumCalVals ][3] = rhoxy;
    CalculatedValues[ NumCalVals ][4] = rhoyz;
    CalculatedValues[ NumCalVals ][5] = rhozx;
    CalculatedValues[ NumCalVals ][6] = CollClass;
   
    ++NumCalVals;
  }
}




/**
 * Mohan::Function that calls the functions to first store the
 * accelerometer valus for a second and then calculate the statistics
 * to provide the training sample...
 */
void WorldState::GenerateATSamples() {
  // For the first 1 second, the robot only stores the accelerometer
  // values along the three axes - noise filtering by averaging
  // later...
  if( fabs(time - tslavs) <= TIME_WINDOW ) {
    AccelValsStore( 1e-6 * sensors[ACCELEROMETER_X], 
		    1e-6 * sensors[ACCELEROMETER_Y], 
		    1e-6 * sensors[ACCELEROMETER_Z] );
  }
  // After the first 1 second of value storage, it is time to generate

  // a single training sample from these...
  else {
    // If the limits of the array to store the training samples have
    // been exceeded, output a message and do nothing...
    if( NumCalVals >= TRAINING_SAMPLE_SIZE ) {
      cout << "\n Cannot Save Values - need to write stuff into file \n";
    }
    // If the array of training samples still has some slots left...
    else {
      // Generate training sample which also increments the count of
      // samples if the training sample is actually stored...
      GenerateStats();

      // Initialize array values to start averaging again...
      NumAccVals = 0;
      accsumx = 0; accsumy = 0; accsumz = 0;

      // Reset time variable...
      tslavs = time;
    }
  }
}


/**
 * Mohan::Reset accelerometer readings array index and some other
 * values if the computation is requested to be stopped...
 */
void WorldState::StopResetVals() {
  tslavs = time;
  NumAccVals = 0;
  accsumx = 0; accsumy = 0; accsumz = 0;
}

//##################################################################################3



/**
 * Mohan::Function used only for distance to landmark calibration -
 * NOT USED DURING GAMETIME...
 */
void WorldState::AverageLandMarkDist( double dist, double angle ) {
  static int Nummergest = 0;

  // If not already collecting data and appropriate sensor (head/UTassist) was
  // not tapped recently, the robot returns without collecting data
  // for beacon distance calibration...
  if( !CollectData ) {
    return;
  }

  BeacDists[ Nummergest ] = dist;
  ++Nummergest;

  cout << " Nummergest = " << Nummergest << endl;

  double sumdist = 0;
  if( Nummergest >= 250 ) {
    CollectData = false;

    for( int i = 0; i < Nummergest; i ++ ) {
      sumdist += BeacDists[i];
    }
    MeasuredDistHist[ Measdistcount ] = sumdist / Nummergest;

    cout << "\n ESTIMATE AT POINT NUM " << Measdistcount << " IS AVERAGED AS "
	 << MeasuredDistHist[ Measdistcount ] << endl;

    ++Measdistcount;
    Nummergest = 0;
  }
}



/**
 * Function used for the stability test for the ICRA04 paper -
 * definitely NOT USED DURING GAMETIME...
 */
void WorldState::StabilityCalculations() {
  static int Fserial = 0;
  
  // For the first 10 seconds, the robot just confirms its pose
  // estimate...
  if( ( time - PrevTime ) <= 10.0 ) {
    cout << "\n Warming Up Now...\n";
  }
  // After that, store values for the next 20 seconds...
  else if( ( time - PrevTime ) <= 30.0 ) {
    if( NumArrVals >= 600 ) {
      cout << "\n Array overflow \n";
    }
    else {
      PoseArr[NumArrVals][0] = bodyPos.getX();
      PoseArr[NumArrVals][1] = bodyPos.getY();
      PoseArr[NumArrVals][2] = bodyAng;
      ++NumArrVals;
    }
  }
  // After 10 seconds of stability and 20 seconds of storing values in
  // array, put these values in a file...
  else {
    char Fname[128];
    sprintf(Fname, "/MS/OPEN-R/MW/DATA/P/FILE%04d.TXT", Fserial);
    cout << "Writing " << Fname << endl;;
    
    FILE* F1 = fopen(Fname, "w");
    if (F1 == 0) {
      cout << "\n Could not Open file to write to...\n";
      return;
    }
    for( int i = 0; i < NumArrVals; i ++ ) {
      fprintf(F1, "%f\t%f\t%f\n", PoseArr[i][0], PoseArr[i][1], PoseArr[i][2]);
    }
    fclose(F1);
    ++Fserial;
    
    ++NumTrials;
    NumArrVals = 0;
    PrevTime = time;
    HistCleaned = false;
    cout << "\n Completed writing values in the file...\n";

    if( NumTrials >= 10 ) {
      StartStabilityTest = false;
    }
  }
}



