#include "Connection.h"
#include "Logger.h"

extern Logger Log;
using namespace std;

Connection::Connection( )
{
  m_iMaxMsgSize = MAX_MSG;
}

Connection::Connection( const char *hostname, int port, int iMaxSize)
{
  m_iMaxMsgSize = iMaxSize;
  if ( connect( hostname, port ) )
    Log.log( 1, "(Connection:connection) Socket connection made with %s:%d",
	     hostname, port );
  else
    Log.log( 1, "(Connection:Connection) Could not create connection with %s:%d",
	     hostname, port );
}

Connection::~Connection()
{
  disconnect();
}

bool Connection::connect( const char *host, int port )
{
  struct hostent *host_ent;
  struct in_addr *addr_ptr;
  struct sockaddr_in cli_addr ;
  int sockfd ;

  m_sock.socketfd = -1 ;

  // Get hostname.
  if ( ( host_ent = (struct hostent*) gethostbyname( host ) ) == NULL ) {
    if ( inet_addr( host ) == INADDR_NONE ) {
      cerr << "(Connection::connect) Cannot find host " << host << endl;
      return false ;
    }
  }
  else {
    addr_ptr = (struct in_addr *) *host_ent->h_addr_list;
    host = inet_ntoa( *addr_ptr );
  }

  //  Open UDP socket.
  if ( ( sockfd = socket( AF_INET, SOCK_DGRAM, 0 ) ) < 0 ) {
    cerr << "(Connection::connect) Cannot create socket " << host << endl;
    return false ;
  }

  // Insert the information of the client
  cli_addr.sin_family = AF_INET ;
  cli_addr.sin_addr.s_addr = htonl(INADDR_ANY) ;
  cli_addr.sin_port = htons(0) ;

  // Bind the client to the server socket
  if ( bind( sockfd, (struct sockaddr *) &cli_addr, sizeof(cli_addr) ) < 0 ) {
    cerr << "(Connection::connect) Cannot bind local address " << host << endl;
    return false ;
  }

  //  Fill in the structure with the address of the server.
  m_sock.socketfd = sockfd ;

  m_sock.serv_addr.sin_family = AF_INET ;
  m_sock.serv_addr.sin_addr.s_addr  = inet_addr(host);
  m_sock.serv_addr.sin_port = htons(port) ;

  return true;
}

void Connection::disconnect()
{
  if ( isConnected() ) {
    close( m_sock.socketfd ) ;
    m_sock.socketfd = -1;
  }
}

bool Connection::isConnected() const
{
  return( m_sock.socketfd != -1 );
}

int Connection::receiveMessage( char *msg, int maxsize )
{
  socklen_t servlen ;
  int n;
  struct sockaddr_in serv_addr ;

  servlen = sizeof(serv_addr) ;

  // receive message from server
  n = recvfrom( m_sock.socketfd, msg, maxsize, 0,
		(struct sockaddr *) &serv_addr, &servlen );

  if ( n < 0 )
  {
    if ( n == -1 && errno == EWOULDBLOCK )
    {
      msg[0] = '\0' ;
      return 0 ;
    }
    else
      return -1;
  }
  else {
    m_sock.serv_addr.sin_port = serv_addr.sin_port ;
    msg[n] = '\0' ;

    return ( n == 0 ) ? 0 : 1 ;
  }
}

bool Connection::sendMessage( const char *msg )
{
  int n = strlen( msg );
  if ( sendto( m_sock.socketfd, msg, n, 0,
	       (struct sockaddr *) &m_sock.serv_addr, 
	       sizeof(m_sock.serv_addr) ) != n )
    return false ;
  Log.log( 1, "(Connection): Sent message: %s", msg );
  return true ;
}


