#include "Logger.h"
#include <stdio.h>
#include <stdarg.h>
#include <string>

#include "Player.h"

// Logger instantiation used by all classes.
Logger Log; 

Logger::Logger( ostream& o, int iMin, int iMax )
{
  strcpy( m_strHeader, "" );
  m_timing.restartTime();
  m_os = &o;
  P = NULL;
}

bool Logger::setPlayer( Player *p )
{
  return ( ( P = p ) != NULL );
}

bool Logger::log( int iLevel, string str )
{
  if ( P == NULL )
	return false;

  if( isInLogLevel( iLevel ) ) { 
    *m_os << P->getCurrentTime() << "|" << iLevel << "||" << str << endl;
    return true;
  }
  return false;
}

bool Logger::log( int iLevel, char *str, ... )
{
  if ( P == NULL )
    return false;
  if ( isInLogLevel( iLevel ) ) {
    va_list ap;
    va_start( ap, str );
    vsprintf( m_buf, str, ap );
    va_end(ap);
    *m_os << P->getCurrentTime() << "|" << iLevel << "||" << m_buf << endl;
    
    return true;
  }
  return false;
}

bool Logger::logWithTime( int iLevel, char *str, ... )
{
  if ( P == NULL )
    return false;
  if ( isInLogLevel( iLevel ) ) {
    va_list ap;
    va_start( ap, str );
    vsprintf( m_buf, str, ap );
    va_end(ap);

    string s = m_strHeader;
    s.append( m_buf );
    s.copy( m_buf, string::npos );
    m_buf[s.length()] = '\0';
    *m_os << P->getCurrentTime() << "|" << iLevel << "||";
    m_timing.printTimeDiffWithText( *m_os, m_buf );
   
    return true;
  }
  return false;
}

void Logger::restartTimer()
{
  return m_timing.restartTime();
}

Timing Logger::getTiming( )
{
  return m_timing;
}

bool Logger::isInLogLevel( int iLevel )
{
  return m_setLogLevels.find( iLevel ) != m_setLogLevels.end() ;
}

bool Logger::addLogLevel( int iLevel )
{
  m_setLogLevels.insert( iLevel ); 
  return true;
}

bool Logger::addLogRange( int iMin, int iMax )
{
  bool bReturn = true;
  for( int i = iMin ; i <= iMax;  i++ )
    bReturn &= addLogLevel( i );
  return bReturn;
}

char* Logger::getHeader( )
{
  return m_strHeader;
}

bool Logger::setHeader( char *str )
{
  strcpy( m_strHeader, str );
  return true;
}

bool Logger::setHeader( int i1, int i2 )
{
  sprintf( m_strHeader, "(%d, %d) ", i1, i2 );
  return true;
}

bool Logger::setOutputStream( ostream& o )
{
  m_os = &o;
  return true;
}

ostream& Logger::getOutputStream(  )
{
  return *m_os;
}

void Logger::showLogLevels( ostream &os )
{
  set<int>::iterator itr;
  for (itr = m_setLogLevels.begin() ; itr != m_setLogLevels.end() ; itr++)
    os << *itr << " " ;
  os << endl;
}

/**********/
/* TIMING */
/**********/

double Timing::getTimeDifference( struct timeval tv1, struct timeval tv2 )
{
  return  ((double)tv1.tv_sec + (double)tv1.tv_usec/1000000 ) -
          ((double)tv2.tv_sec + (double)tv2.tv_usec/1000000 ) ;
}

void Timing::printTimeDiffWithText( ostream &os, char *str, int iFactor )
{
  // set the with to 6 and fill remaining places with '0'.
  os <<setw(6)<< setfill('0')<< getElapsedTime()*iFactor << ":" << str << endl;
}

double Timing::getElapsedTime( )
{
  struct timeval time2;
  gettimeofday( &time2, NULL );
  return getTimeDifference( time2, time1 );
}

void Timing::restartTime( )
{
  gettimeofday( &time1, NULL );
}
