#include <stdlib.h>
#include <string.h>

#include "Connection.h"
#include "Logger.h"
#include "Parse.h"
#include "Player.h"

#define DEFAULT_SERVER_PORT 8675

extern Logger Log;

void printOptions();

int main( int argc, char * argv[] )
{
  int iPort = DEFAULT_SERVER_PORT;
  int iMinLogLevel = 0;
  int iMaxLogLevel = 0;
  char strHost[128] = "localhost";
  bool bSuppliedLogFile = false;
  bool bSingleEstimate = false;
  bool bNoReseed = false;
  ofstream os;

  // read in all the command options and change the associated variables
  // assume every two values supplied at prompt, form a duo
  char *str;
  for ( int i = 1 ; i < argc ; i = i + 2  ) {
    // help is only option that does not have to have an argument
//      if ( i + 1 >= argc && strncmp( argv[i], "-help", 3 ) != 0 ) {
//        cout << "Need argument for option: " << argv[i] << endl;
//        exit( 0 );
//      }
    // read a command option
    if ( argv[i][0] == '-' && strlen( argv[i] ) > 1 ) {
      switch( argv[i][1] ) {
      case '?':                                   // print help
	printOptions( );
	exit(0);
	break;
      case 'h':                                   // host server or help
	if( strlen( argv [i]) > 2 && argv[i][2] == 'e' ) {
	  printOptions( );
	  exit(0);
	}
	else
	  strcpy( strHost, argv[i+1] );
	break;
      case 'l':                                   // loglevel int[..int]
	str = &argv[i+1][0];
	iMinLogLevel = Parse::parseFirstInt( &str );
	while ( iMinLogLevel != 0 ) {
	  if ( *str == '.' || *str == '-' ) {
	    str++;
	    iMaxLogLevel = Parse::parseFirstInt( &str );            
	    if ( iMaxLogLevel == 0 ) 
	      iMaxLogLevel = iMinLogLevel;
	    Log.addLogRange( iMinLogLevel, iMaxLogLevel );
	  }
	  else
	    Log.addLogLevel( iMinLogLevel );
	  iMinLogLevel = Parse::parseFirstInt( &str );                          
	}
	break;
      case 'o':                                   // output file log info
	os.open( argv[i+1] );
	bSuppliedLogFile = true;
	break;
      case 'p':                                   // port
	str = &argv[i+1][0];
	iPort = Parse::parseFirstInt( &str );
	break;
      case 's':
	bSingleEstimate = true;
	i--;
	break;
      case 'n':
	bNoReseed = true;
	i--;
	break;
      default:
	cerr << "(main) Unknown command option: " << argv[i] << endl;
      }
    }
  }

  if( bSuppliedLogFile == true )
    Log.setOutputStream( os );
  else
    Log.setOutputStream( cout );  
  Log.restartTimer();

  WorldState ws( PLAYER_B );
  Connection c( strHost, iPort, MAX_MSG );
  PFLocalization l( &ws, bSingleEstimate, bNoReseed );
  Player player( &ws, &c, &l );
  Log.setPlayer( &player );
  player.mainLoop();

  c.disconnect();
  os.close();
}

void printOptions( )
{
  cout << "Command options:"                                         << endl <<
    " he(lp)                - print this information"                << endl <<
    " h(ost) hostname       - host to connect with"                  << endl <<
    " l(oglevel) int[..int] - level of debug info"                   << endl <<
    " n(o reseeding)        - turn off use of reseed values"         << endl <<
    " o(utput) file         - write log info to (screen is default)" << endl <<
    " p(ort)                - port number to connect with"           << endl <<
    " s(ingle estimate)     - just use single estimate instead of pf" << endl;
}
